// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef __CONSTRAINTS_H
#define __CONSTRAINTS_H 1

// this module contains constraints on a curve or surface F

#include <iostream>
#include <geotyp.h>

#include "param.h"

//-----------------------------------------------------------------------//
//                          ControlPointConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   P[index] = point
//
// where P[i] is the i-th controlpoint of the curve/surface
//

class ControlPointConstraint {
  public:
    int index;      // index of controlpoint
    double3 point;

  public:
    ControlPointConstraint(): point(0,0,0) {}
    // default constructor

    ControlPointConstraint(int i0, const double3& p0): index(i0), point(p0) {}
    // constructor

    ControlPointConstraint(const ControlPointConstraint& cons)
      { index = cons.index; point = cons.point; }
    // copy constructor

    ~ControlPointConstraint() {}
    // destructor

    void Show() {
      std::cout << "Control point P[" << index << "] must be equal to " << point << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const ControlPointConstraint&);
    friend std::istream& operator>>(std::istream&, ControlPointConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const ControlPointConstraint& cons)
{
  return os << "f " << cons.index << " " << cons.point << std::endl;
}
 
inline std::istream& operator>>(std::istream& is, ControlPointConstraint& cons)
{
  return is >> cons.index >> cons.point;
}

//-----------------------------------------------------------------------//
//                          PointToPointConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   F(t) = point
//

class PointToPointConstraint {
  public:
    double3 point;
    Parameter t;

  public:
    PointToPointConstraint(): point(0,0,0), t(0) {}
    // default constructor

    PointToPointConstraint(const Parameter& t0, const double3& p): point(p), t(t0) {}
    // constructor

    PointToPointConstraint(const PointToPointConstraint& cons)
      { point = cons.point; t = cons.t; }
    // copy constructor

    ~PointToPointConstraint() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " must be equal to " << point << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToPointConstraint&);
    friend std::istream& operator>>(std::istream&, PointToPointConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPointConstraint& cons)
{
  return os << "p " << cons.t << " " << cons.point << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPointConstraint& cons)
{
  return is >> cons.t >> cons.point;
}

//-----------------------------------------------------------------------//
//                          PointToLineConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   F(t) on line
//

class PointToLineConstraint {
  public:
    Line3D line;
    Parameter t;

  public:
    PointToLineConstraint(): line(double3(0,0,0), double3(1,0,0)) {}
    // default constructor

    PointToLineConstraint(Parameter t0, const Line3D& l): line(l), t(t0) {}
    // constructor

    PointToLineConstraint(const PointToLineConstraint& cons)
      { line = cons.line; t = cons.t; }
    // copy constructor

    ~PointToLineConstraint() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " must be on the line " << line << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToLineConstraint&);
    friend std::istream& operator>>(std::istream&, PointToLineConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToLineConstraint& cons)
{
  return os << "l " << cons.t << " " << cons.line << std::endl;
}
 
inline std::istream& operator>>(std::istream& is, PointToLineConstraint& cons)
{
  return is >> cons.t >> cons.line;
}

//-----------------------------------------------------------------------//
//                          PointToPlaneConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   F(t) = point
//

class PointToPlaneConstraint {
  public:
    Plane3D plane;
    Parameter t;

  public:
    PointToPlaneConstraint(): plane(double3(1,0,0), 0), t(0) {}
    // default constructor

    PointToPlaneConstraint(const Parameter& t0, const Plane3D& p): plane(p), t(t0) {}
    // constructor

    PointToPlaneConstraint(const PointToPlaneConstraint& cons)
      { plane = cons.plane; t = cons.t; }
    // copy constructor

    ~PointToPlaneConstraint() {}
    // destructor

    friend std::ostream& operator<<(std::ostream&, const PointToPlaneConstraint&);
    friend std::istream& operator>>(std::istream&, PointToPlaneConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPlaneConstraint& cons)
{
  return os << "v " << cons.t << " " << cons.plane << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPlaneConstraint& cons)
{
  return is >> cons.t >> cons.plane;
}

//-----------------------------------------------------------------------//
//                          NormalToVectorConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   N(t) is parallel to normal
//
// where N(t) is the normal vector in F(t).
//
// N.B. this constraint only applies to plane curves and surfaces!
//

class NormalToVectorConstraint {
  public:
    double3 normal;
    Parameter t;

  public:
    NormalToVectorConstraint(): normal(1,0,0), t(0) {}
    // default constructor

    NormalToVectorConstraint(const Parameter& t0, const double3& n): normal(n), t(t0) {}
    // constructor

    NormalToVectorConstraint(const NormalToVectorConstraint& cons)
      { normal = cons.normal; t = cons.t; }
    // copy constructor

    ~NormalToVectorConstraint() {}
    // destructor

    void Show() {
      std::cout << "Normal in point with parameter " << t << " must be parallel to " << normal << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const NormalToVectorConstraint&);
    friend std::istream& operator>>(std::istream&, NormalToVectorConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const NormalToVectorConstraint& cons)
{
  return os << "n " << cons.t << " " << cons.normal << std::endl;
}

inline std::istream& operator>>(std::istream& is, NormalToVectorConstraint& cons)
{
  return is >> cons.t >> cons.normal;
}

//-----------------------------------------------------------------------//
//                          TangentToVectorConstraint
//-----------------------------------------------------------------------//
// This class represents the constraint
//
//                   F'(t) is parallel to tangent
//
// where F'(t) is the tangent vector in F(t).
//
// N.B. this constraint only applies to space curves!
//

class TangentToVectorConstraint {
  public:
    double3 tangent;
    Parameter t;

  public:
    TangentToVectorConstraint(): tangent(1,0,0), t(0) {}
    // default constructor

    TangentToVectorConstraint(const Parameter& t0, const double3& tan): tangent(tan), t(t0) {}
    // constructor

    TangentToVectorConstraint(const TangentToVectorConstraint& cons)
      { tangent = cons.tangent; t = cons.t; }
    // copy constructor

    ~TangentToVectorConstraint() {}
    // destructor

    void Show() {
      std::cout << "Tangent in point with parameter " << t << " must be parallel to " << tangent << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const TangentToVectorConstraint&);
    friend std::istream& operator>>(std::istream&, TangentToVectorConstraint&);
};

inline std::ostream& operator<<(std::ostream& os, const TangentToVectorConstraint& cons)
{
  return os << "t " << cons.t << " " << cons.tangent << std::endl;
}

inline std::istream& operator>>(std::istream& is, TangentToVectorConstraint& cons)
{
  return is >> cons.t >> cons.tangent;
}

#endif
