// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef __DISTFUNC_H
#define __DISTFUNC_H 1

#include <iostream>
#include <function.h>
#include <geotyp.h>
#include "bspline.h"

//-----------------------------------------------------------------------//
//                          PointDistFunc
//-----------------------------------------------------------------------//
// This function represents the distance between a B-spline curve x and a
// point P:
//                    f(t) = d(x(t),P)^2
//

class PointDistFunc: public Function1D {
  protected:
    const BSpline& x;
    double3 P;

  public:
    PointDistFunc(const BSpline& curve, const double3& point);
    // constructor

    PointDistFunc(const PointDistFunc&);
    // copy constructor

    ~PointDistFunc();
    // destructor

    double operator()(double t) const;
    // returns f(t)

    double Deriv(double t) const;
    // returns f'(t)

    double DDeriv(double t) const;
    // returns f''(t)

    int MinimizeNewton(double& t, double eps = 1e-3, int maxiter = 10,
                       double a = 0, double b = 1);

    friend std::ostream& operator<<(std::ostream&, const PointDistFunc&);
    friend std::istream& operator>>(std::istream&, PointDistFunc&);
};

//-----------------------------------------------------------------------//
//                          LineDistFunc
//-----------------------------------------------------------------------//
// This function represents the distance between a B-spline curve x and a
// Line3D l:
//                    f(t) = d(x(t),l)^2
//
// A and b are such that d(y,l) = ||Ay + b||
//

class LineDistFunc: public Function1D {
  protected:
    const BSpline& x;
    double3x3 A;
    double3 b;

  public:
    LineDistFunc(const BSpline& curve, const Line3D& line);
    // constructor

    LineDistFunc(const LineDistFunc&);
    // copy constructor

    ~LineDistFunc();
    // destructor

    double operator()(double t) const;
    // returns f(t)

    double Deriv(double t) const;
    // returns f'(t)

    double DDeriv(double t) const;
    // returns f''(t)

    int MinimizeNewton(double& t, double eps = 1e-3, int maxiter = 10,
                       double a = 0, double b = 1);

    friend std::ostream& operator<<(std::ostream&, const LineDistFunc&);
    friend std::istream& operator>>(std::istream&, LineDistFunc&);
};

#endif

