// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 2-5-95

// this module contains a datatype for handling with fixed points

#ifndef __FIXPOINT_H
#define __FIXPOINT_H

#include <listtmpl.h>
#include <ivector.h>
#include <dvector.h>
#include <double3.h>

//-----------------------------------------------------------------------//
//                          TFixedPointHandler                           //
//-----------------------------------------------------------------------//
// The class TFixedPointHandler represents a vector X that is the
// concatenation of n 3D points P[i] = (x_i, y_i, z_i), (i = 0..n-1).
// Some of these points have a fixed value, while others are free.
//
class TFixedPointHandler {
  protected:
    IntVector fixed; // fixed[i] == 0  ->  Point i is free
                     // fixed[i] == 1  ->  Point i is fixed

    IntVector index; // fixed[i] == 0  ->  Point i is the index[i]-th free point
                     // fixed[i] == 1  ->  Point i is the index[i]-th fixed point

    List<double3> fixedpoints; // contains the values of the fixed points
    int numpoints;

  public:
    TFixedPointHandler();
    // default constructor

    TFixedPointHandler(int np);
    // constructor (initializes numpoints)

    int IsPointFixed(int i) { return fixed[i]; }
    // returns 1 if point i is fixed, else 0

    int GetIndex(int i) { return index[i]; }

    DoubleVector GetFreePointsVector(DoubleVector& X);
    // given a concatenation vector X
    //
    //            X  =  [ P[0] .. P[numpoints-1] ]
    //
    // this function returns the concatenation vector Y
    //
    //            Y  =  [ P[i(0)] .. P[i(NumFreePoints()-1)] ]
    //
    // where i(p) is the index of the p-th free point.

    DoubleVector GetAllPointsVector(DoubleVector& Y);
    // given a concatenation vector Y
    //
    //            Y  =  [ P[i(0)] .. P[i(NumFreePoints()-1)] ]
    //
    // this function returns the concatenation vector X
    //
    //            X  =  [ P[0] .. P[numpoints-1] ]
    //
    // where i(p) is the index of the p-th free point.

    double3& GetFixedPoint(int i) { return fixedpoints[i]; }
    // returns the value of the i-th fixed point

    int NumPoints() { return numpoints; }
    // returns the number of points

    int NumFixedPoints() { return fixedpoints.Count; }
    // returns the number of fixed points

    int NumFreePoints() { return numpoints - NumFixedPoints(); }
    // returns the number of free points

    void FixPoint(int i, const double3& point);
    // makes the i-th point fixed

    void FreePoint(int i);
    // makes the i-th point free

    void SetSize(int np);
    // changes the number of points to np; the current contents are lost!

    void Show();
    // shows the contents of the class
};

#endif


