// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef __POTENTIALS_H
#define __POTENTIALS_H 1

// this module contains potentials on a curve or surface F

#include <iostream>
#include <string.h>
#include <geotyp.h>
#include <double3.h>

#include "param.h"

//-----------------------------------------------------------------------//
//                          StretchEnergy
//-----------------------------------------------------------------------//
// This class represents the expression
//
//       -                     T
//      |  weight * DF(t) DF(t)  dt
//     -
//   region
//
// where DF(t) is the Jacobian of F in the point with parameter t.

class StretchEnergy {
  public:
    Interval region;
    double weight;

  public:
    StretchEnergy(): region(0,1), weight(1) {}
    // default constructor

    StretchEnergy(Interval r, double w): region(r), weight(w) {}
    // constructor

    StretchEnergy(const StretchEnergy& oper)
      { region = oper.region; weight = oper.weight; }
    // copy constructor

    ~StretchEnergy() {}
    // destructor

    void Show() {
      std::cout << "  Stretch Energy on region " << region << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const StretchEnergy&);
    friend std::istream& operator>>(std::istream&, StretchEnergy&);
};

inline std::ostream& operator<<(std::ostream& os, const StretchEnergy& energy)
{
  return os << "S " << energy.region << " " << energy.weight << std::endl;
}

inline std::istream& operator>>(std::istream& is, StretchEnergy& energy)
{
  return is >> energy.region >> energy.weight;
}

//-----------------------------------------------------------------------//
//                          BendEnergy
//-----------------------------------------------------------------------//
// This class represents the expression
//
//       -                        2
//      |  weight * trace( H(F(t)) ) dt
//     -
//   region
//
// where H(F(t)) is the Hessian of F in the point with parameter t.

class BendEnergy {
  public:
    Interval region;
    double weight;

  public:
    BendEnergy(): region(0,1), weight(1) {}
    // default constructor

    BendEnergy(const BendEnergy& oper)
      { region = oper.region; weight = oper.weight; }   
    // copy constructor

    BendEnergy(Interval r, double w): region(r), weight(w) {}
    // constructor

    ~BendEnergy() {}
    // destructor

    void Show() {
      std::cout << "  Bend Energy on region " << region << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const BendEnergy&);
    friend std::istream& operator>>(std::istream&, BendEnergy&);
};

inline std::ostream& operator<<(std::ostream& os, const BendEnergy& energy)
{
  return os << "B " << energy.region << " " << energy.weight << std::endl;
}

inline std::istream& operator>>(std::istream& is, BendEnergy& energy)
{
  return is >> energy.region >> energy.weight;
}

//-----------------------------------------------------------------------//
//                          PointToPointAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * d(F(t),point)^2
//
// where d(P,point) is the distance between the points P and point

class PointToPointAttractor {
  public:
    Parameter t;
    double weight;
    double3 point;
    
  public:
    PointToPointAttractor(): t(0), weight(1), point(0,0,0) {}
    // default constructor

    PointToPointAttractor(const PointToPointAttractor& attr)
      { t = attr.t; weight = attr.weight; point = attr.point; }
    // copy constructor

    PointToPointAttractor(const Parameter& t0, double w, const double3& p): t(t0), weight(w), point(p) {}
    // constructor    

    ~PointToPointAttractor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is attracted to " << point << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToPointAttractor&);
    friend std::istream& operator>>(std::istream&, PointToPointAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPointAttractor& attr)
{
  return os << "P " << attr.t << " " << attr.weight << " " << attr.point << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPointAttractor& attr)
{
  return is >> attr.t >> attr.weight >> attr.point;
}

//-----------------------------------------------------------------------//
//                          PointToLineAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * d(F(t),line)^2
//
// where d(P,line) is the distance between the point P and the Line3D line

class PointToLineAttractor {
  public:
    Parameter t;
    double weight;
    Line3D line;

  public:
    PointToLineAttractor(): t(0), weight(1), line(double3(0,0,0), double3(1,0,0)) {}
    // default constructor

    PointToLineAttractor(const PointToLineAttractor& attr)
      { t = attr.t; weight = attr.weight; line = attr.line; }
    // copy constructor

    PointToLineAttractor(const Parameter& t0, double w, const Line3D& l):
      t(t0), weight(w), line(l) {}
    // constructor

    ~PointToLineAttractor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is attracted to the line" << line << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToLineAttractor&);
    friend std::istream& operator>>(std::istream&, PointToLineAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToLineAttractor& attr)
{
  return os << "L " << attr.t << " " << attr.weight << std::endl << attr.line << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToLineAttractor& attr)
{
  return is >> attr.t >> attr.weight >> attr.line;
}

//-----------------------------------------------------------------------//
//                          PointToPlaneAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * d(F(t),plane)^2
//
// where d(P,plane) is the distance between the point P and the Plane3D plane

class PointToPlaneAttractor {
  public:
    Parameter t;
    double weight;
    Plane3D plane;

  public:
    PointToPlaneAttractor(): t(0), weight(1), plane(double3(1,0,0), 0) {}
    // default constructor

    PointToPlaneAttractor(const PointToPlaneAttractor& attr)
      { t = attr.t; weight = attr.weight; plane = attr.plane; }
    // copy constructor

    PointToPlaneAttractor(const Parameter& t0, double w, const Plane3D& p):
      t(t0), weight(w), plane(p) {}
    // constructor

    ~PointToPlaneAttractor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is attracted to the plane" << plane << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToPlaneAttractor&);
    friend std::istream& operator>>(std::istream&, PointToPlaneAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPlaneAttractor& attr)
{
  return os << "V " << attr.t << " " << attr.weight << std::endl << attr.plane << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPlaneAttractor& attr)
{
  return is >> attr.t >> attr.weight >> attr.plane;
}

//-----------------------------------------------------------------------//
//                          RegionToPointAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),point)^2  dt
//   -
// region
//
// where d(P,point) is the distance between the points P and point

class RegionToPointAttractor {
  public:
    Interval region;
    double weight;
    double3 point;

  public:
    RegionToPointAttractor(): region(0,1), weight(1), point(0,0,0) {}
    // default constructor

    RegionToPointAttractor(Interval r, double w, const double3& p): region(r), weight(w), point(p) {}
    // constructor

    RegionToPointAttractor(const RegionToPointAttractor& attr)
      { region = attr.region; weight = attr.weight; point = attr.point; }
    // copy constructor

    ~RegionToPointAttractor() {}    
    // destructor

    void Show() {
      std::cout << "  RegionToPointAttractor: " << std::endl;
      std::cout << "     Interval " << region << " is attracted to " << point << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToPointAttractor&);
    friend std::istream& operator>>(std::istream&, RegionToPointAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToPointAttractor& attr)
{
  return os << "Q " << attr.region << " " << attr.weight << " " << attr.point << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToPointAttractor& attr)
{
  return is >> attr.region >> attr.weight >> attr.point;
}

//-----------------------------------------------------------------------//
//                          RegionToLineAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),line)^2  dt
//   -
// region
//
// where d(P,line) is the distance between the point P and the Line3D line


class RegionToLineAttractor {
  public:
    Interval region;
    double weight;
    Line3D line;

  public:
    RegionToLineAttractor(): region(0,1), weight(1), line(double3(0,0,0), double3(1,0,0)) {}
    // default constructor

    RegionToLineAttractor(Interval r, double w, const Line3D& l):
      region(r), weight(w), line(l) {}
    // constructor

    RegionToLineAttractor(const RegionToLineAttractor& attr)
      { region = attr.region; weight = attr.weight; line = attr.line; }
    // copy constructor

    ~RegionToLineAttractor() {}
    // destructor

    void Show() {
      std::cout << "  Line Attractor on region " << region << std::endl;
      std::cout << "     region is attracted to line " << line << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToLineAttractor&);
    friend std::istream& operator>>(std::istream&, RegionToLineAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToLineAttractor& attr)
{
  return os << "M " << attr.region << " " << attr.weight << std::endl << attr.line << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToLineAttractor& attr)
{
  return is >> attr.region >> attr.weight >> attr.line;
}

//-----------------------------------------------------------------------//
//                          RegionToPlaneAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),plane)^2  dt
//   -
// region
//
// where d(P,plane) is the distance between the point P and the Plane3D plane

class RegionToPlaneAttractor {
  public:
    Interval region;
    double weight;
    Plane3D plane;

  public:
    RegionToPlaneAttractor(): region(0,1), weight(1), plane(double3(1,0,0), 0) {}
    // default constructor

    RegionToPlaneAttractor(Interval r, double w, const Plane3D& p):
      region(r), weight(w), plane(p) {}
    // constructor

    RegionToPlaneAttractor(const RegionToPlaneAttractor& attr)
      { region = attr.region; weight = attr.weight; plane = attr.plane; }
    // copy constructor

    ~RegionToPlaneAttractor() {}
    // destructor

    void Show() {
      std::cout << "  Plane Attractor on region " << region << std::endl;
      std::cout << "    attractor plane: " << plane << std::endl;
      std::cout << "     weight factor: weight = " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToPlaneAttractor&);
    friend std::istream& operator>>(std::istream&, RegionToPlaneAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToPlaneAttractor& attr)
{
  return os << "W " << attr.region << " " << attr.weight << std::endl << attr.plane << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToPlaneAttractor& attr)
{
  return is >> attr.region >> attr.weight >> attr.plane;
}

//-----------------------------------------------------------------------//
//                          TangentToVectorAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * (F'(t),tangent)^2
//
// where F'(t) is the tangent vector in F(t) and (P,Q) is the inner product
// between P and Q
//
// N.B. this potential only applies to space curves!

class TangentToVectorAttractor {
  public:
    Parameter t;
    double weight;
    double3 tangent;

  public:
    TangentToVectorAttractor(): t(0), weight(1), tangent(1,0,0) {}
    // default constructor

    TangentToVectorAttractor(const Parameter& t0, double w, const double3& tan):
      t(t0), weight(w), tangent(tan) {}
    // constructor

    TangentToVectorAttractor(const TangentToVectorAttractor& attr)
      { t = attr.t; weight = attr.weight; tangent = attr.tangent; }
    // copy constructor

    ~TangentToVectorAttractor() {}
    // destructor

    void Show() {
      std::cout << "Tangent in point with parameter " << t << " is attracted to the vector" << tangent << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const TangentToVectorAttractor&);
    friend std::istream& operator>>(std::istream&, TangentToVectorAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const TangentToVectorAttractor& attr)
{
  return os << "T " << attr.t << " " << attr.weight << std::endl << attr.tangent << std::endl;
}

inline std::istream& operator>>(std::istream& is, TangentToVectorAttractor& attr)
{
  return is >> attr.t >> attr.weight >> attr.tangent;
}

//-----------------------------------------------------------------------//
//                          NormalToVectorAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * [ (Fu(t),normal)^2 + (Fv(t),normal)^2 ]
//
// where Fu(t) and Fv(t) are the partial derivatives of F in t and (P,Q) is
// the inner product between P and Q
//
// N.B. this potential only applies to surfaces!

class NormalToVectorAttractor {
  public:
    Parameter t;
    double weight;
    double3 normal;

  public:
    NormalToVectorAttractor(): t(0), weight(1), normal(1,0,0) {}
    // default constructor

    NormalToVectorAttractor(const Parameter& t0, double w, const double3& n):
      t(t0), weight(w), normal(n) {}
    // constructor

    NormalToVectorAttractor(const NormalToVectorAttractor& attr)
      { t = attr.t; weight = attr.weight; normal = attr.normal; }
    // copy constructor

    ~NormalToVectorAttractor() {}
    // destructor

    void Show() {
      std::cout << "Normal in point with parameter " << t << " is attracted to the vector" << normal << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const NormalToVectorAttractor&);
    friend std::istream& operator>>(std::istream&, NormalToVectorAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const NormalToVectorAttractor& attr)
{
  return os << "N " << attr.t << " " << attr.weight << std::endl << attr.normal << std::endl;
}

inline std::istream& operator>>(std::istream& is, NormalToVectorAttractor& attr)
{
  return is >> attr.t >> attr.weight >> attr.normal;
}

//-----------------------------------------------------------------------//
//                          TangentRegionToVectorAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//     -
//    |   weight * (F'(t),tangent)^2  dt
//   -
// region
//
// where F'(t) is the tangent vector in F(t) and (P,Q) is the inner product
// between P and Q
//
// N.B. this potential only applies to space curves!

class TangentRegionToVectorAttractor {
  public:
    Interval region;
    double weight;
    double3 tangent;

  public:
    TangentRegionToVectorAttractor(): region(0,1), weight(1), tangent(1,0,0) {}   
    // default constructor

    TangentRegionToVectorAttractor(const Interval& r, double w, const double3& t):
      region(r), weight(w), tangent(t) {}
    // constructor

    TangentRegionToVectorAttractor(const TangentRegionToVectorAttractor& attr)
      { region = attr.region; weight = attr.weight; tangent = attr.tangent; }
    // copy constructor

    ~TangentRegionToVectorAttractor() {}
    // destructor

    void Show() {
      std::cout << "Tangent in region " << region << " is attracted to the vector" << tangent << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const TangentRegionToVectorAttractor&);
    friend std::istream& operator>>(std::istream&, TangentRegionToVectorAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const TangentRegionToVectorAttractor& attr)
{
  return os << "U " << attr.region << " " << attr.weight << std::endl << attr.tangent << std::endl;
}

inline std::istream& operator>>(std::istream& is, TangentRegionToVectorAttractor& attr)
{
  return is >> attr.region >> attr.weight >> attr.tangent;
}

//-----------------------------------------------------------------------//
//                          NormalRegionToVectorAttractor
//-----------------------------------------------------------------------//
// This class represents the expression
//     -
//    |   weight * [ (Fu(t),normal)^2 + (Fv(t),normal)^2 ] dt
//   -
// region
//
// where Fu(t) and Fv(t) are the partial derivatives of F in t and (P,Q) is
// the inner product between P and Q
//
// N.B. this potential only applies to surfaces!

class NormalRegionToVectorAttractor {
  public:
    Interval region;
    double weight;
    double3 normal;

  public:
    NormalRegionToVectorAttractor(): region(0,1), weight(1), normal(1,0,0) {}
    // default constructor

    NormalRegionToVectorAttractor(const Interval& r, double w, const double3& n):
      region(r), weight(w), normal(n) {}
    // constructor

    NormalRegionToVectorAttractor(const NormalRegionToVectorAttractor& attr)
      { region = attr.region; weight = attr.weight; normal = attr.normal; }
    // copy constructor

    ~NormalRegionToVectorAttractor() {}
    // destructor

    void Show() {
      std::cout << "Normal in region " << region << " is attracted to the vector" << normal << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const NormalRegionToVectorAttractor&);
    friend std::istream& operator>>(std::istream&, NormalRegionToVectorAttractor&);
};

inline std::ostream& operator<<(std::ostream& os, const NormalRegionToVectorAttractor& attr)
{
  return os << "N " << attr.region << " " << attr.weight << std::endl << attr.normal << std::endl;
}

inline std::istream& operator>>(std::istream& is, NormalRegionToVectorAttractor& attr)
{
  return is >> attr.region >> attr.weight >> attr.normal;
}

//-----------------------------------------------------------------------//
//                          PointToPointRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight / ( d(F(t),point)^2 )
//
// where d(P,point) is the distance between the points P and point

class PointToPointRepellor {
  public:
    Parameter t;
    double weight;
    double3 point;

  public:
    PointToPointRepellor(): t(0), weight(1), point(0,0,0) {}
    // default constructor

    PointToPointRepellor(const Parameter& t0, double w, const double3& p): t(t0), weight(w), point(p) {}
    // constructor

    PointToPointRepellor(const PointToPointRepellor& rep)
      { t = rep.t; weight = rep.weight; point = rep.point; }
    // copy constructor

    ~PointToPointRepellor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is repelled from " << point << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToPointRepellor&);
    friend std::istream& operator>>(std::istream&, PointToPointRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPointRepellor& rep)
{
  return os << "R " << rep.t << " " << rep.weight << " " << rep.point << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPointRepellor& rep)
{
  return is >> rep.t >> rep.weight >> rep.point;
}

//-----------------------------------------------------------------------//
//                          PointToLineRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * d(F(t),line)^2
//
// where d(P,line) is the distance between the point P and the Line3D line

class PointToLineRepellor {
  public:
    Parameter t;
    double weight;
    Line3D line;

  public:
    PointToLineRepellor(): t(0), weight(1), line(double3(0,0,0), double3(1,0,0)) {}
    // default constructor

    PointToLineRepellor(const PointToLineRepellor& rep)
      { t = rep.t; weight = rep.weight; line = rep.line; }
    // copy constructor

    PointToLineRepellor(const Parameter& t0, double w, const Line3D& l):
      t(t0), weight(w), line(l) {}
    // constructor

    ~PointToLineRepellor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is repelled from the line" << line << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToLineRepellor&);
    friend std::istream& operator>>(std::istream&, PointToLineRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToLineRepellor& rep)
{
  return os << "0 " << rep.t << " " << rep.weight << std::endl << rep.line << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToLineRepellor& rep)
{
  return is >> rep.t >> rep.weight >> rep.line;
}

//-----------------------------------------------------------------------//
//                          PointToPlaneRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * d(F(t),plane)^2
//
// where d(P,plane) is the distance between the point P and the Plane3D plane

class PointToPlaneRepellor {
  public:
    Parameter t;
    double weight;
    Plane3D plane;

  public:
    PointToPlaneRepellor(): t(0), weight(1), plane(double3(1,0,0), 0) {}
    // default constructor

    PointToPlaneRepellor(const PointToPlaneRepellor& rep)
      { t = rep.t; weight = rep.weight; plane = rep.plane; }
    // copy constructor

    PointToPlaneRepellor(const Parameter& t0, double w, const Plane3D& p):
      t(t0), weight(w), plane(p) {}
    // constructor

    ~PointToPlaneRepellor() {}
    // destructor

    void Show() {
      std::cout << "Point with parameter " << t << " is repelled from the plane" << plane << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const PointToPlaneRepellor&);
    friend std::istream& operator>>(std::istream&, PointToPlaneRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const PointToPlaneRepellor& rep)
{
  return os << "1 " << rep.t << " " << rep.weight << std::endl << rep.plane << std::endl;
}

inline std::istream& operator>>(std::istream& is, PointToPlaneRepellor& rep)
{
  return is >> rep.t >> rep.weight >> rep.plane;
}

//-----------------------------------------------------------------------//
//                          RegionToPointRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),point)^2  dt
//   -
// region
//
// where d(P,point) is the distance between the points P and point

class RegionToPointRepellor {
  public:
    Interval region;
    double weight;
    double3 point;

  public:
    RegionToPointRepellor(): region(0,1), weight(1), point(0,0,0) {}
    // default constructor

    RegionToPointRepellor(Interval r, double w, const double3& p): region(r), weight(w), point(p) {}
    // constructor

    RegionToPointRepellor(const RegionToPointRepellor& rep)
      { region = rep.region; weight = rep.weight; point = rep.point; }
    // copy constructor

    ~RegionToPointRepellor() {}    
    // destructor

    void Show() {
      std::cout << "  RegionToPointRepellor: " << std::endl;
      std::cout << "     Interval " << region << " is repelled from " << point << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToPointRepellor&);
    friend std::istream& operator>>(std::istream&, RegionToPointRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToPointRepellor& rep)
{
  return os << "2 " << rep.region << " " << rep.weight << " " << rep.point << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToPointRepellor& rep)
{
  return is >> rep.region >> rep.weight >> rep.point;
}

//-----------------------------------------------------------------------//
//                          RegionToLineRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),line)^2  dt
//   -
// region
//
// where d(P,line) is the distance between the point P and the Line3D line


class RegionToLineRepellor {
  public:
    Interval region;
    double weight;
    Line3D line;

  public:
    RegionToLineRepellor(): region(0,1), weight(1), line(double3(0,0,0), double3(1,0,0)) {}
    // default constructor

    RegionToLineRepellor(Interval r, double w, const Line3D& l):
      region(r), weight(w), line(l) {}
    // constructor

    RegionToLineRepellor(const RegionToLineRepellor& rep)
      { region = rep.region; weight = rep.weight; line = rep.line; }
    // copy constructor

    ~RegionToLineRepellor() {}
    // destructor

    void Show() {
      std::cout << "  Line Repellor on region " << region << std::endl;
      std::cout << "     region is repelled from line " << line << std::endl;
      std::cout << "     weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToLineRepellor&);
    friend std::istream& operator>>(std::istream&, RegionToLineRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToLineRepellor& rep)
{
  return os << "3 " << rep.region << " " << rep.weight << std::endl << rep.line << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToLineRepellor& rep)
{
  return is >> rep.region >> rep.weight >> rep.line;
}

//-----------------------------------------------------------------------//
//                          RegionToPlaneRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//     -
//    |   weight d(F(t),plane)^2  dt
//   -
// region
//
// where d(P,plane) is the distance between the point P and the Plane3D plane

class RegionToPlaneRepellor {
  public:
    Interval region;
    double weight;
    Plane3D plane;

  public:
    RegionToPlaneRepellor(): region(0,1), weight(1), plane(double3(1,0,0), 0) {}
    // default constructor

    RegionToPlaneRepellor(Interval r, double w, const Plane3D& p):
      region(r), weight(w), plane(p) {}
    // constructor

    RegionToPlaneRepellor(const RegionToPlaneRepellor& rep)
      { region = rep.region; weight = rep.weight; plane = rep.plane; }
    // copy constructor

    ~RegionToPlaneRepellor() {}
    // destructor

    void Show() {
      std::cout << "  Plane Repellor on region " << region << std::endl;
      std::cout << "    repactor plane: " << plane << std::endl;
      std::cout << "     weight factor: weight = " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const RegionToPlaneRepellor&);
    friend std::istream& operator>>(std::istream&, RegionToPlaneRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const RegionToPlaneRepellor& rep)
{
  return os << "4 " << rep.region << " " << rep.weight << std::endl << rep.plane << std::endl;
}

inline std::istream& operator>>(std::istream& is, RegionToPlaneRepellor& rep)
{
  return is >> rep.region >> rep.weight >> rep.plane;
}

//-----------------------------------------------------------------------//
//                          TangentToVectorRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//
//                 weight * (F'(t),tangent)^2
//
// where F'(t) is the tangent vector in F(t) and (P,Q) is the inner product
// between P and Q
//
// N.B. this potential only applies to space curves!

class TangentToVectorRepellor {
  public:
    Parameter t;
    double weight;
    double3 tangent;

  public:
    TangentToVectorRepellor(): t(0), weight(1), tangent(1,0,0) {}
    // default constructor

    TangentToVectorRepellor(const Parameter& t0, double w, const double3& tan):
      t(t0), weight(w), tangent(tan) {}
    // constructor

    TangentToVectorRepellor(const TangentToVectorRepellor& rep)
      { t = rep.t; weight = rep.weight; tangent = rep.tangent; }
    // copy constructor

    ~TangentToVectorRepellor() {}
    // destructor

    void Show() {
      std::cout << "Tangent in point with parameter " << t << " is repelled from the vector" << tangent << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const TangentToVectorRepellor&);
    friend std::istream& operator>>(std::istream&, TangentToVectorRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const TangentToVectorRepellor& rep)
{
  return os << "5 " << rep.t << " " << rep.weight << std::endl << rep.tangent << std::endl;
}

inline std::istream& operator>>(std::istream& is, TangentToVectorRepellor& rep)
{
  return is >> rep.t >> rep.weight >> rep.tangent;
}

//-----------------------------------------------------------------------//
//                          TangentRegionToVectorRepellor
//-----------------------------------------------------------------------//
// This class represents the expression
//     -
//    |   weight * (F'(t),tangent)^2  dt
//   -
// region
//
// where F'(t) is the tangent vector in F(t) and (P,Q) is the inner product
// between P and Q
//
// N.B. this potential only applies to space curves!

class TangentRegionToVectorRepellor {
  public:
    Interval region;
    double weight;
    double3 tangent;

  public:
    TangentRegionToVectorRepellor(): region(0,1), weight(1), tangent(1,0,0) {}   
    // default constructor

    TangentRegionToVectorRepellor(const Interval& r, double w, const double3& t):
      region(r), weight(w), tangent(t) {}
    // constructor

    TangentRegionToVectorRepellor(const TangentRegionToVectorRepellor& rep)
      { region = rep.region; weight = rep.weight; tangent = rep.tangent; }
    // copy constructor

    ~TangentRegionToVectorRepellor() {}
    // destructor

    void Show() {
      std::cout << "Tangent in region " << region << " is repelled from the vector" << tangent << std::endl;
      std::cout << "  weight factor: " << weight << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const TangentRegionToVectorRepellor&);
    friend std::istream& operator>>(std::istream&, TangentRegionToVectorRepellor&);
};

inline std::ostream& operator<<(std::ostream& os, const TangentRegionToVectorRepellor& rep)
{
  return os << "6 " << rep.region << " " << rep.weight << std::endl << rep.tangent << std::endl;
}

inline std::istream& operator>>(std::istream& is, TangentRegionToVectorRepellor& rep)
{
  return is >> rep.region >> rep.weight >> rep.tangent;
}

//-----------------------------------------------------------------------//
//                          Profiler
//-----------------------------------------------------------------------//
// This class represents the expression
//
//              -
//             |  weight * d(PP(F(t)), G(t))^2 dt
//            -
//          region
//
//  where PP is the parallel projection in the Plane3D plane along the
//  direction dir, G is a given curve and d(P,Q) is the distance between the
//  points P and Q
//
// N.B. this potential only applies to space curves!

class Profiler {
  public:
    Interval region;
    double weight;
    Plane3D plane;
    double3 dir;
    char FileName[120]; // file in which the curve G is located

  public:
    Profiler(): region(0,1), weight(0), plane(double3(1,0,0), 0), dir(0,1,0) {}
    // default constructor

    Profiler(Interval r, double w, const Plane3D& p, const double3& d, char* FName):
      region(r), weight(w), plane(p), dir(d) {
      strcpy(FileName, FName);
    }
    // constructor

    Profiler(const Profiler& prof)
      { weight = prof.weight; plane = prof.plane; dir = prof.dir; strcpy(FileName, prof.FileName); }
    // copy constructor

    ~Profiler() {}
    // destructor

    void Show() {
      std::cout << "  Profiler on region          " << region << std::endl;
      std::cout << "     weight factor:           " << weight << std::endl;
      std::cout << "     projection plane:        " << plane << std::endl;
      std::cout << "     direction:               " << dir << std::endl;
      std::cout << "     curve is located in file " << FileName << std::endl;
    }

    friend std::ostream& operator<<(std::ostream&, const Profiler&);
    friend std::istream& operator>>(std::istream&, Profiler&);
};

inline std::ostream& operator<<(std::ostream& os, const Profiler& prof)
{
  return os << "F " << prof.region << " " << prof.weight << std::endl
            << prof.FileName << std::endl
            << prof.plane << " " << prof.dir << std::endl;
}

inline std::istream& operator>>(std::istream& is, Profiler& prof)
{
  return is >> prof.region >> prof.weight >> prof.FileName >> prof.plane >> prof.dir;
}

#endif
