// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sat 06-22-1996

#ifndef __TENSORBSPLINE_H
#define __TENSORBSPLINE_H

#include <consurf.h>
#include <bspline.h>
#include <polygon.h>

const int MAXROWS = 20;
const int MAXCOLS = 20;
const int MAXPOINTS = MAXROWS*MAXCOLS;
const int MAXKNOTS  = 100;

//-----------------------------------------------------------------------//
//                          TensorBSpline                                //
//-----------------------------------------------------------------------//
class TensorBSpline: public ControlPointSurface
// This class represents a uniform cubic tensor product B-spline surface x with
// knot vectors [0 1 2 3 ... ] and [0 1 2 3 ... ] in u- and v-direction.
//
//               m-1 n-1 
//      x(u,v) = Sum Sum N[i,3](u) N[j,3](v) P[i,j]
//               i=0 j=0
//
// with P[i,j] the controlpoints and N[i,n] the i-th basis function of degree n
//
{
  protected:
    double3 V[MAXPOINTS];
    double uknots[MAXKNOTS];
    double vknots[MAXKNOTS];
    int numRows, numCols;      // number of control points in u- and v-direction

  public:
    TensorBSpline();
    // default constructor

    TensorBSpline(int nrows, int ncols);
    // constructor

    TensorBSpline(char* FileName);
    // constructor (reads data from file)

    void operator= (TensorBSpline& G);
    // assignment

    ///////////////////////////////////////////////////////////////////////
    //                       parametrization + knots
    ///////////////////////////////////////////////////////////////////////

    int NumUKnots() const { return numRows+2; }
    // returns the number of knots in the u-direction

    int NumVKnots() const { return numCols+2; }
    // returns the number of knots in the v-direction

    int NumUSegments() const { return numRows-3; }
    // returns the number of segments in the u-direction

    int NumVSegments() const { return numCols-3; }
    // returns the number of segments in the v-direction

    double MinUPar() const { return uknots[2]; }
    // returns the minimum u-parameter value of the surface

    double MinVPar() const { return vknots[2]; }
    // returns the minimum v-parameter value of the surface

    double MaxUPar() const { return uknots[NumUKnots()-3]; }
    // returns maximum u-parameter value of the surface

    double MaxVPar() const { return vknots[NumVKnots()-3]; }
    // returns maximum v-parameter value of the surface

    int UIndex(double u) const;
    // returns the u-index i of u (i.e. uknots[i] <= u < uknots[i+1])

    int VIndex(double v) const;
    // returns the v-index i of v (i.e. vknots[i] <= v < vknots[i+1])

    int USegment(double u) const { return UIndex(u)-2; }
    // returns the number of the segment that contains u
    // the segments are numbered from 0 .. NumUSegments()-1

    int VSegment(double v) const { return VIndex(v)-2; }
    // returns the number of the segment that contains v
    // the segments are numbered from 0 .. NumVSegments()-1

    double& UKnot(int i) { return uknots[i]; }
    // returns the i-th u-knot of the surface (0 <= i < NumUKnots())

    double& VKnot(int i) { return vknots[i]; }
    // returns the i-th v-knot of the surface (0 <= i < NumVKnots())

    double* UKnotVector() { return uknots; }
    // returns a pointer to the uknot vector

    double* VKnotVector() { return vknots; }
    // returns a pointer to the vknot vector

    ///////////////////////////////////////////////////////////////////////
    //                       control points
    ///////////////////////////////////////////////////////////////////////

    int NumControlPoints() const { return numRows*numCols; }
    // returns the number of control points of the surface

    int NumRows() const { return numRows; }
    // returns the number of control points in a row of the control net

    int NumCols() const { return numCols; }
    // returns the number of control points in a column of the control net

    int ControlPointIndex(int i, int j) const { return numCols*i + j; }
    // returns the index of ControlPoint(i,j) in the array V

    const double3& GetControlPoint(int i) const { return V[i]; }
    // returns the i-th control point of the surface (0 <= i < NumControlPoints())

    const double3& GetControlPoint(int i, int j) const { return V[i*numCols + j]; }
    // returns the j-th control point in the i-th row
    // (0 <= i < numRows) && (0 <= j < numCols)

    void SetControlPoint(int i, const double3& point) { V[i] = point; }
    // sets the value of the i-th control point (0 <= i < NumControlPoints())

    void SetControlPoint(int i, int j, const double3& point) { V[i*numCols + j] = point; }
    // sets the value of thej-th control point in the i-th row
    // (0 <= i < numRows) && (0 <= j < numCols)

    void GetControlPointMesh(PolygonMesh& mesh) const;
    // generates a mesh of the control polyhedron

    ///////////////////////////////////////////////////////////////////////
    //                       geometric features
    ///////////////////////////////////////////////////////////////////////

    int NumPatches() const { return NumUSegments()*NumVSegments(); }
    // returns the number of patches of the surface

    Rectangle2D GetDomain() const;
    // returns the domain of the surface

    void GetBoundaryCurves(BSpline& umincurve, BSpline& umaxcurve, BSpline& vmincurve, BSpline& vmaxcurve) const;
    // returns the curves that represent the boundary of the surface
    
    double3 GetDerivative(const double2& u, int k, int l) const;
    // returns the partial derivative
    //
    //              d^(k+l) F(u,v)
    //              --------------
    //                du^k dv^l

    double3 GetDirectionalDerivative(const double2& u, const double2& d, int k) const;
    // returns the directional derivative
    //
    //               (k)
    //              g   (t)             with g(t) = F(u+td)

    IntVector GetPatchIndices(int p, int q) const;
    // returns the indices of the control points
    //
    //          ControlPoint(i+p,j+q),      0 <= i,j <= 3
    //
    // that belong to the patch with domain [p+2,p+3] x [q+2,q+3]

    double GetSizeMeasure() const;
    // returns a rough approximation of the size of the surface
    
    void GetSurfaceMesh(int numrows, int numcols, SurfaceMesh& mesh) const;
    // generates a mesh of the surface by splitting the domain in numcols*numrows rectangles

    void GetSurfaceMesh(const Polygon2D& gamma, SurfaceMesh& mesh, int resolution) const;
    // generates a mesh of the surface on the domain gamma

    SurfSample GetSurfSample(const double2& u) const;
    // computes a sample of F at the point with parameter u

    ///////////////////////////////////////////////////////////////////////
    //                       input/output         
    ///////////////////////////////////////////////////////////////////////
    
    int LoadFile(char*);
    // loads control points from a file

    int SaveFile(char*) const;
    // saves control points to a file

    void Show() const;
    // writes the contents to standard output

    ///////////////////////////////////////////////////////////////////////
    //                       other                
    ///////////////////////////////////////////////////////////////////////
    
    void SetCylinder(int NumRows, int NumCols, double h, double R, double Angle);
    // arranges the control points of the surface such that they lie
    // on the cylinder:
    //
    //   f(u,v) = (R cos(u), R sin(u), h)
    //
    // The surface is an approximation of the cylinder on the domain
    // { 0 <= u <= M_PI*Angle/180  &&  0 <= v <= h }

    void SetPlane(int NumRows, int NumCols, double Width, double Height);
    // arranges the control points of the surface such that they lie
    // on the plane:
    //
    //   f(u,v) = (u, v, 0)
    //
    // The surface is an (exact) approximation of the plane on the domain
    // { 0 <= u <= Width  &&  0 <= v <= Height }

    void SetTorus(int NumRows, int NumCols, double R, double RAngle, double r, double rAngle);
    // arranges the control points of the surface such that they lie
    // on the torus:
    //
    //   f(u,v) = ( (R + r cos(v)) cos(u), (R + r cos(v) sin(u)), r sin(v) )
    //
    // The surface is an approximation of the torus on the domain
    // { 0 <= u <= M_PI*Angle/180  &&  0 <= v <= Height }
};

#endif

