// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 29-3-95

#ifndef __TEQUATION_H
#define __TEQUATION_H

// This module contains datatypes for (systems of) equations.

#include <iostream>
#include <cmath>
#include <ivector.h>
#include <double3.h>
#include <dvector.h>
#include <dmatrix.h>
#include <listtmpl.h>

//-----------------------------------------------------------------------//
//                          TDouble                                      //
//-----------------------------------------------------------------------//
// This class represents a double. To make it suitable for numerical
// computation an attribute iszero is added to it to indicate if it has
// the value zero.

class TDouble {
  protected:
    double val;
    int iszero;

  public:
    TDouble(): iszero(1) {}
    TDouble(double v): val(v), iszero(0) {}
    double Value() { if (iszero) return 0; else return val; }
    int IsZero() { return iszero; }

    int IsAbsBigger(TDouble& x) { return (!iszero && (fabs(val) > fabs(x.Value()))); }
    // returns TRUE if |val| > x

    void MakeZero() { iszero = 1; }

    void operator+= (double v);
    void operator+= (TDouble& x);
    void operator*= (double t);

    friend std::ostream& operator<<(std::ostream& s, TDouble& x);
};

//-----------------------------------------------------------------------//
//                          Equation                                     //
//-----------------------------------------------------------------------//
// This class represents the linear equation a^T x = b, where b is a 3D point.

class Equation {
  protected:
    int done; // This attribute is used by the class EquationSystem to indicate
              // if the equation was already used for clearing a column.
              // It is absolutely useless for the class Equation itself.
    TDouble* a;
    double3 b;
    int n;    // size of a

  public:
    Equation(): done(0), n(0) {}
    Equation(int n0): done(0), b(0), n(n0) { a = new TDouble[n0]; }
    Equation(int n0, TDouble* a0, const double3& b0): done(0), a(a0), b(b0), n(n0) {}
    ~Equation() { if (n>0) delete[] a; }

    double3& Const() { return b; }
    int GetSize() { return n; }
    TDouble& operator[](int i) { return a[i]; }
    int IsDone() { return done; }
    void SetConst(const double3& b0) { b = b0; }
    void SetDone(int d=1) { done = d; }
    void SetSize(int n0);

    void Swap(int i, int j) { TDouble temp = a[i]; a[i] = a[j]; a[j] = temp; }
    // swaps a[i] and a[j]

    void Add(Equation& Equation, double t);
    // adds t*Equation to (*this)

    void operator+= (Equation& Equation);
    void operator*= (double t);
    friend std::ostream& operator<<(std::ostream& s, Equation& Equation);
};

//-----------------------------------------------------------------------//
//                          EquationSystem                               //
//-----------------------------------------------------------------------//
// This class represents a linear system of equations.

class EquationSystem {
  protected:
    List<Equation> rows;
    IntVector sigma; // info about permutations of columns
                     // sigma[i] == j -> the current column j was initially column i
                     // is valid after a call to the method Clear()

  public:
    EquationSystem(): rows(10,10) {}
    ~EquationSystem() {}

    int NumRows() { return rows.Count; }

    int NumCols();

    Equation& operator[](int i) { return rows[i]; }

    void AddEquation(Equation* eq) { rows.Insert(eq); }
    // adds an equation to the system

    void GetVector(DoubleVector&);
    // computes the righthandside of the equations

    void GetMatrix(DoubleMatrix&);
    // returns the system matrix

    void GetRightMatrix(DoubleMatrix&);
    // returns the matrix that consists of the last (num rows - num columns) columns

    void GetPermutationVector(IntVector& s) { s = sigma; }
    // returns sigma

    void SwapColumns(int i, int j);
    // swaps the columns i and j; sigma is updated

    void GetPivotElement(int& row, int& col);
    // searches for an element in the matrix that is in a row that is
    // not done yet with maximum absolute value
    // if no nonzero element is found, the values of row and col are unchanged

    void ClearColumn(int row, int col);
    // by elementary row operations the column col is transformed into the
    // unitvector e[row]

    void Clear();
    // By elementary row operations and permutations of columns, the matrix
    // is transformed into a matrix of the form [ I(m) F ], where I(m) is the
    // m-dimensional unit matrix, and m is the number of columns minus the
    // number of rows.
    // precondition:
    //  -> none of the rows is labeled done
    //  -> the number of columns is at least as big as the number of rows

    void GetSimpleForm(DoubleMatrix& C, DoubleVector& D);
    // Computes a matrix C and a vector D, such that the original system
    //
    //                         A X = B
    // is equivalent to
    //                           X = C Y + D

    friend std::ostream& operator<<(std::ostream& s, EquationSystem& System);
};

//--- INLINE FUNCTIONS ------------------------------------------------//

inline void TDouble::operator+= (double v)
{
  if (iszero) {
    iszero = 0;
    val = v;
  }
  else
    val += v;
}

inline void TDouble::operator+= (TDouble& x)
{
  if (!x.IsZero()) (*this) += x.Value();
}

inline void TDouble::operator*= (double t)
{
  if (!iszero) val *= t;
}

inline std::ostream& operator<<(std::ostream& s, TDouble& x)
{
  if (x.IsZero())
    s << '.';
  else
    s << x.Value();
  return s;
}

inline void Equation::SetSize(int n0)
{
  if (n > 0)
    delete[] a;
  n = n0;
  a = new TDouble[n0];
}

inline int EquationSystem::NumCols()
{
  int numcols;
  if (rows.Count>0)
    numcols = rows[0].GetSize();
  else
    numcols = 0;
  return numcols;
}

#endif
