// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-09-1996

#ifndef __BIPOLYNOMIAL_H
#define __BIPOLYNOMIAL_H

#include <iostream>
#include <poly.h>
#include <geotyp.h>
#include <trapezd.h>

//-----------------------------------------------------------------------//
//                          BiPolynomial                                 //
//-----------------------------------------------------------------------//
// The class BiPolynomial represents the bivariate polynomial f
//
//           f(x,y) = sum     a(i,j) x^i y^j t^n
//                    i+j<=n
//
// The coefficients a(i,j) are stored in the array coefs in the
// following order: 
//
//     a(0,0), a(0,1), a(1,0), a(0,2), a(1,1), a(2,0), a(0,3), ...
//
// Hence,
//           a(i,j)   = coefs[ (i+j)*(i+j+1)/2 + i ]
//
// and
//
//    coefs[i] = a(Binominal(p,2)+i+p, Binominal(p,2)-i-1),
//
// where p is the smallest positive integer such that i < Binominal(p,2)
//
//      =>     p <= (sqrt(8i+1) - 1)/2 < p + 1
//      =>     p = ceil( (sqrt(8i+1) - 1)/2 ) + 1           mag dit???

class BiPolynomial
{
  protected:
    double* coefs;
    int n;      // degree
    int maxn;   // maximum degree that the array coefs can hold

    void allocate(int degree, int CopyContents=0);
    // maxn is set to max(maxn,d);
    // if necessary new memory is allocated for coefs;
    //   if (CopyContents) is true, then the current contents of coefs are copied
    
    void raise_degree(int degree);
    // n is set to max(degree,n) and the coefficients
    // a[i,j], i+j > n are set to 0
    // if necessary, new memory is allocated

    int numcoefs(int degree) const { return (degree+1)*(degree+2)/2; }
    double a(int i, int j) const { return coefs[(i+j)*(i+j+1)/2 + i]; }
    double& a(int i, int j) { return coefs[(i+j)*(i+j+1)/2 + i]; }
    
  public:
    BiPolynomial(): n(0), maxn(0) { coefs = new double[numcoefs(maxn)]; coefs[0]=1; }
    // default constructor; creates the polynomial f(x,y) = 1

    BiPolynomial(int d);
    // constructor; creates the degree d polynomial f(x,y) = 0

    BiPolynomial(const Polynomial& g, const Polynomial& h);
    // constructor; creates the polynomial f(x,y) = g(x) h(y)

    BiPolynomial(double* c, int degree): coefs(c), n(degree), maxn(degree) {}
    // constructor

    BiPolynomial(const BiPolynomial&);
    // copy constructor

    ~BiPolynomial() { delete[] coefs; }
    // destructor

    void RaiseDegree(int degree) { raise_degree(degree); }
    // the degree is set to max(degree,n) and the coefficients
    // a[i,j], i+j > n are set to 0
    
    void SetDegree(int d);
    // sets the degree of f to d; the current contents are lost

    int Degree() const { return n; }
    // returns the degree of the polynomial

    double& Coefficient(int i, int j) { return a(i,j); }
    // returns the i-th coefficient (0 <= i <= n)

    double Coefficient(int i, int j) const { return a(i,j); }
    // returns the i-th coefficient (0 <= i <= n)

    void Show() const;
    // writes the polynomial to standard out

    double operator()(double x, double y);
    // returns f(x,y)

    BiPolynomial Primitive(int k, int l) const;
    // returns a polynomial F, such that
    //
    //             d^(k+l) F
    //             ---------   =   f
    //             dx^k dy^l
    
    BiPolynomial Derivative(int k, int l) const;
    // returns a polynomial F, such that
    //
    //             d^(k+l) f
    //             ---------   =   F
    //             dx^k dy^l
    
    double Derivative (double x, double y, int k, int l) const;
    // returns the value
    //
    //             d^(k+l) f
    //             --------- (x,y)
    //             dx^k dy^l

    double Integrate(double x0, double x1, double a0, double a1, double b0, double b1) const;
    // returns the following integral:
    //
    //      x1   b0 + b1 x
    //      -       -
    //     |       |        f(x,t) dt  dx
    //    -       -
    //    x0   a0 + a1 x

    double Integrate(const VerticalTrapezoid2D& t) const
      { return Integrate(t.x0, t.x1, t.a0, t.a1, t.b0, t.b1); }

    void operator  = (double c) { n=0; coefs[0]=c; }
    void operator  = (const BiPolynomial&);
    void operator += (const BiPolynomial&);
    void operator -= (const BiPolynomial&);
    void operator *= (double);
    void operator /= (double);

    friend BiPolynomial operator - (const BiPolynomial&);
    friend BiPolynomial operator + (const BiPolynomial&, const BiPolynomial&);
    friend BiPolynomial operator - (const BiPolynomial&, const BiPolynomial&);
    friend BiPolynomial operator * (const BiPolynomial&, const BiPolynomial&);
    friend BiPolynomial operator * (double, const BiPolynomial&);
    friend BiPolynomial operator * (const BiPolynomial&, double);
    friend BiPolynomial operator / (const BiPolynomial&, double);

    friend std::ostream& operator<<(std::ostream&, const BiPolynomial&);
    friend std::istream& operator>>(std::istream&, BiPolynomial&);
    // format: degree coefs[0] coefs[1] ...
};

#endif
