// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 22-6-95

#ifndef __DOUBLE2_H
#define __DOUBLE2_H

#include <cmath>
#include <iostream>
#include <int2.h>
#include <mymanip.h>

#ifndef EPSILON
#define EPSILON 1e-10
#endif

//-----------------------------------------------------------------------//
//                          double2                                       //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 2 doubles

struct double2
{
  protected:
    double X[2];

  public:
    // constructors
    double2();
    double2(double, double);
    double2(double fill) { X[0]=X[1]=fill; }
    double2(const double2&);
    double2(const int2&);

    int Dim() { return 2; }

    double Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..2)

    double& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..2)

    double operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..1)

    double& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..1)

    double& x() { return X[0]; }
    double& y() { return X[1]; }
    double x() const { return X[0]; }
    double y() const { return X[1]; }

    double Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double getNorm() const;
    // returns the length of the vector

    double Dist(const double2& p) const;
    // returns the Euclidian distance from the point X to p

    double2 Unit() const;
    // returns the vector X / Length(X)

    double2 getNormalized() const;
    // returns the vector X / Length(X)

    void MakeUnit();
    // makes the vector equal to (X) / Length(X)

    double normalize();
    // makes the vector equal to (X) / Length(X) and returns Length(X)

    void MakeZero();
    // makes the vector equal to (0,0)

    void Negate();
    // replaces the vector with -X

    void    operator = (const double2&);
    double2  operator - () const;
    double2& operator += (const double2&);
    double2& operator -= (const double2&);
    double2& operator *= (double);
    double2& operator /= (double);

    friend int operator== (const double2&, const double2&);
    friend double2 operator + (const double2&, const double2&);
    friend double2 operator - (const double2&, const double2&);
    friend double2 operator * (const double2&, double);
    friend double2 operator * (double, const double2&);
    friend double2 operator / (const double2&, double);
    friend double operator * (const double2&, const double2&); // inner product
    friend double operator , (const double2&, const double2&); // inner product
    friend double operator | (const double2&, const double2&); // inner product
    friend double operator ^ (const double2&, const double2&); // cross product
    friend double Dist(const double2&, const double2&);
    friend std::ostream& operator<<(std::ostream& s, const double2& p);
    friend std::istream& operator>>(std::istream& s, double2& p);
};

inline double2::double2 ()
{ X[0]=X[1]=0; }

inline double2::double2 (double x0, double x1)
{ X[0]=x0; X[1]=x1; }

inline double2::double2 (const double2& p)
{ X[0]=p[0]; X[1]=p[1]; }

inline double2::double2 (const int2& p)
{ X[0]=p[0]; X[1]=p[1]; }

inline void double2::operator = (const double2& p)
{ X[0]=p[0]; X[1]=p[1]; }

inline double2 double2::operator - () const
{ return double2(-X[0], -X[1]); }

inline double2& double2::operator += (const double2& p)
{ X[0]+=p[0]; X[1]+=p[1]; return *this; }

inline double2& double2::operator -= (const double2& p)
{ X[0]-=p[0]; X[1]-=p[1]; return *this; }

inline double2& double2::operator *= (double t)
{ X[0]*=t; X[1]*=t; return *this; }

inline double2& double2::operator /= (double t)
{ X[0]/=t; X[1]/=t; return *this; }

inline double double2::Length2() const
{ return X[0]*X[0] + X[1]*X[1]; }

inline double double2::Length () const
{ return sqrt( (*this).Length2() ); }

inline double double2::getNorm() const
{ return Length(); }

inline int operator== (const double2& a, const double2& b)
{ return (a[0]==b[0] && a[1]==b[1]); }

inline double2 operator + (const double2& a, const double2& b)
{ return double2(a[0]+b[0], a[1]+b[1]); }

inline double2 operator - (const double2& a, const double2& b)
{ return double2(a[0]-b[0], a[1]-b[1]); }

inline double double2::Dist(const double2& p) const
{ return (*this - p).Length(); }

inline double2 operator * (const double2& a, double t)
{ return double2(t*a[0], t*a[1]); }

inline double2 operator * (double t, const double2& p)
{ return p * t; }

inline double2 operator / (const double2& a, double t)
{ return double2(a[0]/t, a[1]/t); }

inline double operator * (const double2& a, const double2& b)
{ return a[0]*b[0] + a[1]*b[1]; }

inline double operator , (const double2& a, const double2& b)
{ return a*b; }

inline double operator | (const double2& a, const double2& b)
{ return a*b; }

inline double operator ^ (const double2& a, const double2& b)
{ return a[0]*b[1] - a[1]*b[0]; }

inline double Dist(const double2& a, const double2& b)
{ return a.Dist(b); }

inline double2 double2::Unit() const
{ double l = Length();
  if (l > 0.0) return (*this/l); else return double2(0,0);
}

inline double2 double2::getNormalized() const
{ return Unit(); }

inline void double2::MakeUnit()
{ double l = Length(); if (l > 0.0) (*this)/=l; }

inline double double2::normalize()
{ double l = Length(); if (l > 0.0) (*this)/=l; return l; }

inline void double2::MakeZero()
{ X[0] = X[1] = 0; }

inline void double2::Negate ()
{ X[0] = -X[0]; X[1] = -X[1]; }

inline std::ostream& operator<<(std::ostream& s, const double2& a)
{ return s << '(' << a[0] << ',' << a[1] << ')'; }

inline std::istream& operator>>(std::istream& s, double2& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipbracket;
}

#endif

