// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 30-05-95

#ifndef __DOUBLE2X2_H
#define __DOUBLE2X2_H

#include <double2.h>

//-----------------------------------------------------------------------//
//                          double2x2                                     //
//-----------------------------------------------------------------------//
class double2x2
{
  protected:
    double A[2][2]; 

  public:
    double2x2() {}
    // constructor

    double2x2(const double2x2& B) { *this = B;}
    // copy constructor

    double2x2(double a00, double a01, double a10, double a11);
    // constructor

    double2x2(const double2& row0, const double2& row1); 
    // constructor (initializes rows)
    
    ~double2x2() {}
    // destructor

    double& Elem(int i, int j) { return A[i][j]; }
    // returns the (i,j)-th component; 1 <= i <= 2     1 <= j <= 2
    
    double Elem(int i, int j) const { return A[i][j]; }
    // returns the (i,j)-th component; 1 <= i <= 2     1 <= j <= 2
    
    double2 Row(int i) const { return double2(A[i][0], A[i][1]); }
    // returns the i-th row of the matrix (i = 0..1)

    double2 Col(int i) const { return double2(A[0][i], A[1][i]); }
    // returns the i-th column of the matrix (i = 0..1)

    const double* operator[](int i) const { return A[i]; }
    // returns a pointer to the i-th row of A
    // this makes it possible to access a variable B of type double2x2 via B[i][j]

    double* operator[](int i) { return A[i]; }
    // returns a pointer to the i-th row of A
    // this makes it possible to access a variable B of type double2x2 via B[i][j]

    double2x2 Inverse() const;
    // returns the inverse of the matrix
   
    double2x2 Transpose() const;
    // returns the transpose of the matrix
   
    double FrobeniusNorm() const;
    // returns the Frobenius norm of the matrix
    // this norm is defined as
    //
    //                  2   2            2
    //          Sqrt[  Sum Sum  |A[i][j]|   ]
    //                 i=0 j=0
               
    double Det() const;
    // returns the determinant of the matrix

    double Trace() const;
    // returns the trace of the matrix

    int GetEigenvalues(double& l1, double& l2);
    // computes the eigenvalues of the matrix
    // If these eigenvalues are complex, false is returned.

    void Fill(double fill);
    // makes all elements of the matrix equal to fill

    void FillDiag(int fill) { A[0][0] = A[1][1] = fill; }
    // makes all elements (i,i) of the matrix equal to fill

    void MakeZero() { Fill(0); }
    // makes all elements of the matrix equal to 0

    void MakeUnit() { Fill(0); FillDiag(1); }
    // makes all (i,i) elements equal to 1 and the rest equal to 0

    void MakeSymmetric();
    // if the matrix is square every element A(i,j) is replaced with (A(i,j) + A(j,i))/2

    void MultCol(int i, double t);
    // multiplies the i-th column of the matrix by t  (i,j = 0..1)

    void MultRow(int i, double t);
    // multiplies the i-th row of the matrix by t (i,j = 0..1)

    void operator  = (const double2x2&);
    void operator += (const double2x2&);
    void operator -= (const double2x2&);
    void operator *= (double t);
    void operator /= (double t);

    friend double2x2 operator - (const double2x2&);
    friend double2x2 operator + (const double2x2& A, const double2x2& B);
    friend double2x2 operator - (const double2x2& A, const double2x2& B);
    friend double2x2 operator * (double t, const double2x2& A);
    friend double2x2 operator * (const double2x2& A, double t);
    friend double2x2 operator / (const double2x2& A, double t);
    friend double2x2 operator * (const double2x2& A, const double2x2& B);
    friend double2 operator * (const double2x2& A, const double2& x);
    friend double2 operator * (const double2& x, const double2x2& A);
    // returns the vector x^T A as a column vector

    friend std::ostream& operator<<(std::ostream& s, const double2x2& A);
    friend std::istream& operator>>(std::istream& s, double2x2& A);
};

double2x2 DiagonalMatrix(const double2& x);
// returns the diagonal matrix with x(i) as (i,i)-th element

double2x2 DyadeMatrix(const double2& x, const double2& y);
// returns the matrix        T
//                       x  y

inline double2x2::double2x2(double a00, double a01, double a10, double a11)
{
  A[0][0] = a00;
  A[0][1] = a01;
  A[1][0] = a10;
  A[1][1] = a11;
}

inline double2x2 operator * (const double2x2& A, double t) { return t*A; }

inline double2 operator * (const double2x2& A, const double2& x)
{
  return double2( (A.Row(0),x), (A.Row(1),x) );
}

inline double2 operator * (const double2& x, const double2x2& A)
{
  return double2( (A.Col(0),x), (A.Col(1),x) );
}

#endif
