// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 30-05-95

#ifndef __DOUBLE3X3_H
#define __DOUBLE3X3_H

#include <double3.h>

//-----------------------------------------------------------------------//
//                          double3x3                                     //
//-----------------------------------------------------------------------//
class double3x3
{
  protected:
    double A[3][3]; 

  public:
    double3x3() {}
    // constructor

    double3x3(const double3x3& B) { *this = B;}
    // copy constructor

	 double3x3(double a00, double a01, double a02,
				  double a10, double a11, double a12,
				  double a20, double a21, double a22);

	 double3x3(const double3& row0, const double3& row1, const double3& row2);
	 // constructor (initializes rows)

	 ~double3x3() {}
	 // destructor

	 double& Elem(int i, int j) { return A[i][j]; }
	 // returns the (i,j)-th component; 1 <= i <= 3     1 <= j <= 3

	 double Elem(int i, int j) const { return A[i][j]; }
	 // returns the (i,j)-th component; 1 <= i <= 3     1 <= j <= 3

	 double3 Row(int i) const { return double3(A[i][0], A[i][1], A[i][2]); }
	 // returns the i-th row of the matrix (i = 0..2)

	 double3 Col(int i) const { return double3(A[0][i], A[1][i], A[2][i]); }
	 // returns the i-th column of the matrix (i = 0..2)

	 const double* operator[](int i) const { return A[i]; }
	 // returns a pointer to the i-th row of A
	 // this makes it possible to access a variable B of type double3x3 via B[i][j]

	 double* operator[](int i) { return A[i]; }
	 // returns a pointer to the i-th row of A
	 // this makes it possible to access a variable B of type double3x3 via B[i][j]

	 double3x3 Transpose() const;
	 // returns the transpose of the matrix

	 double FrobeniusNorm() const;
    // returns the Frobenius norm of the matrix
    // this norm is defined as
    //
    //                  2   2            2
    //          Sqrt[  Sum Sum  |A[i][j]|   ]
    //                 i=0 j=0
               
    double Det() const;
    // returns the determinant of the matrix

    void Fill(double fill);
    // makes all elements of the matrix equal to fill

    void FillDiag(int fill) { A[0][0] = A[1][1] = A[2][2] = fill; }
    // makes all elements (i,i) of the matrix equal to fill

    void MakeZero() { Fill(0); }
    // makes all elements of the matrix equal to 0

    void MakeUnit() { Fill(0); FillDiag(1); }
    // makes all (i,i) elements equal to 1 and the rest equal to 0

    void MakeSymmetric();
    // if the matrix is square every element A(i,j) is replaced with (A(i,j) + A(j,i))/2

    void SwapCols(int i, int j);
    // swaps the i-th and j-th column (i,j = 0..2)

    void SwapRows(int i, int j);
    // swaps the i-th and j-th row (i,j = 0..2)

    void MultCol(int i, double t);
    // multiplies the i-th column of the matrix by t  (i,j = 0..2)

    void MultRow(int i, double t);
    // multiplies the i-th row of the matrix by t (i,j = 0..2)

    // axial rotation matrix
    double3x3& AxRot(const double3&, double);

    // rotation matrix over angle phi round X|Y|Z-axis
    double3x3& Xrot(double);
    double3x3& Yrot(double);
    double3x3& Zrot(double);

    void operator  = (const double3x3&);
    void operator += (const double3x3&);
    void operator -= (const double3x3&);
    void operator *= (double t);
    void operator /= (double t);

    friend double3x3 operator - (const double3x3&);
    friend double3x3 operator + (const double3x3& A, const double3x3& B);
    friend double3x3 operator - (const double3x3& A, const double3x3& B);
    friend double3x3 operator * (double t, const double3x3& A);
    friend double3x3 operator * (const double3x3& A, double t);
    friend double3x3 operator / (const double3x3& A, double t);
    friend double3x3 operator * (const double3x3& A, const double3x3& B);
    friend double3 operator * (const double3x3& A, const double3& x);
    friend double3 operator * (const double3& x, const double3x3& A);
    // returns the vector x^T A as a column vector

    friend std::ostream& operator<<(std::ostream& s, const double3x3& A);
    friend std::istream& operator>>(std::istream& s, double3x3& A);
};

double3x3 DiagonalMatrix(const double3& x);
// returns the diagonal matrix with x(i) as (i,i)-th element

double3x3 DyadeMatrix(const double3& x, const double3& y);
// returns the matrix        T
//                       x  y

inline double3x3 operator * (const double3x3& A, double t) { return t*A; }

inline double3 operator * (const double3x3& A, const double3& x)
{
  return double3( (A.Row(0),x), (A.Row(1),x), (A.Row(2),x) );
}

inline double3 operator * (const double3& x, const double3x3& A)
{
  return double3( (A.Col(0),x), (A.Col(1),x), (A.Col(2),x) );
}


inline double3x3& double3x3::Xrot(double phi)
{ double  cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]= 1.0; A[0][1]= 0.0; A[0][2]= 0.0;
  A[1][0]= 0.0; A[1][1]=  cp; A[1][2]=  sp;
  A[2][0]= 0.0; A[2][1]= -sp; A[2][2]=  cp;

  return(*this);
}


inline double3x3& double3x3::Yrot(double phi)
{ 
  double  cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]=  cp; A[0][1]= 0.0; A[0][2]=  sp;
  A[1][0]= 0.0; A[1][1]= 1.0; A[1][2]= 0.0;
  A[2][0]= -sp; A[2][1]= 0.0; A[2][2]=  cp;

  return(*this);
}


inline double3x3& double3x3::Zrot(double phi)
{ 
  double cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]=  cp; A[0][1]=  sp; A[0][2]= 0.0;
  A[1][0]= -sp; A[1][1]=  cp; A[1][2]= 0.0;
  A[2][0]= 0.0; A[2][1]= 0.0; A[2][2]= 1.0;

  return(*this);
}

#endif
