// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 13-11-95

#ifndef __DYNALIST_H
#define __DYNALIST_H

#include <iostream>

//
//----------------------------------------------------------------------
//                       DynaList<T>
//----------------------------------------------------------------------
// this class implements a dynamic list of objects of type T.
// The list contains actual copies of the objects added to it, obtained
// by the '=' operator defined for T.
//
// actual extending the list takes place with blocks of size EXTSIZE.
// the sizes are never reduced, untill destruction of the object.
// EXTSIZE can (if desired) be set by construction, default is 25.
//

template <class T>
class DynaList {
  protected:
    T        *elements;
    int      contens,EXTSIZE,SIZE;
    void MakeRoomFor(int i);
  public:
    DynaList();
    DynaList(const DynaList<T> &dl);
    DynaList(int V_EXTSIZE);
    ~DynaList();

    void cat(T element);
    void cat(DynaList<T> dl);
    void insertAt(T element,int i);
    void insertAt(DynaList<T> dl,int i);
    void del(int i);
    void delFrom(int i,int j);
    
    int Size() const;
    DynaList<T> Take(int i);
    DynaList<T> Drop(int i);
    
    T at(int i) const;
      // alternate notation for []-operator
    T operator[](int i) const;
    T &operator[](int i);
    DynaList<T> &operator=(const DynaList<T> &dl);
    //friend std::ostream &operator<<(std::ostream &s, DynaList<T> &dl);
};

template <class T>
DynaList<T>::DynaList()
{ elements=0;
  contens=0;
  EXTSIZE=25;
  SIZE=0;
}

template <class T>
DynaList<T>::DynaList(const DynaList<T> &dl)
{ if (dl.SIZE==0) elements=0;
  else elements=new T[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
}

template <class T>
DynaList<T>::DynaList(int V_EXTSIZE)
{ elements=0;
  contens=0;
  EXTSIZE=V_EXTSIZE;
  SIZE=0;
}

template <class T>
DynaList<T>::~DynaList()
{ delete[] elements;
}

template <class T>
void DynaList<T>::MakeRoomFor(int i)
{ if ((contens+i)>SIZE) {
    SIZE=((contens+i)/EXTSIZE)*EXTSIZE;
    if ((contens+i)>SIZE) SIZE+=EXTSIZE;
    T *h=new T[SIZE];
    for(int j=0;j!=contens;j++) h[j]=elements[j];
    if (elements!=0) delete[] elements;
    elements=h;
  }
}

template <class T>
void DynaList<T>::cat(T element)
{ MakeRoomFor(1);
  elements[contens]=element;
  contens++;
}

template<class T>
void DynaList<T>::cat(DynaList<T> dl)
{ MakeRoomFor(dl.contens);
  for(int i=0;i!=dl.contens;i++) {
    elements[contens]=dl.elements[i];
    contens++;
  }
}

template <class T>
void DynaList<T>::insertAt(T element,int i)
{ MakeRoomFor(1);
  for(int j=contens;j!=i;j--) elements[j]=elements[j-1];
  elements[i]=element;
  contens++;
}

template <class T>
void DynaList<T>::insertAt(DynaList<T> dl,int i)
{ MakeRoomFor(dl.contens);
  for(int j=contens;j!=i;j--) elements[dl.contens+j-1]=elements[j-1];
  for(int j=0;j!=dl.contens;j++) elements[i+j]=dl.elements[j];
  contens+=dl.contens;
}

template <class T>
void DynaList<T>::del(int i)
{ for(int k=i+1;k!=contens;k++) elements[k-1]=elements[k];
  contens--;
}

template <class T>
void DynaList<T>::delFrom(int i,int j)
{ for(int k=i+j;k!=contens;k++) elements[k-j]=elements[k];
  contens-=j;
} 

template <class T>
int DynaList<T>::Size() const
{ return contens;
}

template <class T>
DynaList<T> DynaList<T>::Take(int i)
{ DynaList<T> dl(EXTSIZE);
  if (i>contens) i=contens; 
  else if (i<=0) return dl;
   else {
    dl.MakeRoomFor(i);
    for(int k=0;k!=i;k++) dl.elements[k]=elements[k];
    dl.contens=i;
    return dl;
  }
}

template <class T>
DynaList<T> DynaList<T>::Drop(int i)
{ DynaList<T> dl;
  if (i<0) i=0;
  else if (i>=contens) return dl;
  else {
    dl.MakeRoomFor(contens-i);
    for(int k=i;k!=contens;k++) dl.elements[k-i]=elements[k];
    dl.contens=contens-i;
    return dl;
  }
}

template <class T>
T DynaList<T>::at(int i) const
{ return elements[i];
}

template <class T>
T DynaList<T>::operator[](int i) const
{ return elements[i];
}

template <class T>
T &DynaList<T>::operator[](int i)
{ return elements[i];
}

template <class T>
DynaList<T> &DynaList<T>::operator=(const DynaList<T> &dl)
{ if (elements!=0) delete[] elements;
  if (dl.SIZE==0) elements=0;
  else elements=new T[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
  
  return *this;
}

template <class T>
std::ostream &operator<<(std::ostream &os,DynaList<T> &dl)
{ if (dl.Size()==0) os<<"empty";
  else { 
    os<<"["<<dl[0];
    for(int k=1;k!=dl.Size();k++) os<<","<<dl[k];
    os<<"]";
  }
  return os;
}

//
//----------------------------------------------------------------------
//                       DynaSList<T>
//----------------------------------------------------------------------
// this class implements a dynamic sorted list of objects of type T.
// The list contains actual copies of the objects added to it, obtained
// by the '=' operator defined for T. The order is determined by the 
// operators '<','<=' and '==' that must be defined on T.
//
// actual extending the list takes place with blocks of size EXTSIZE.
// the sizes are never reduced, untill destruction of the object.
// EXTSIZE can (if desired) be set by construction, default is 25.
//

template <class T>
class DynaSList {
  protected:
    T        *elements;
    int      contens,EXTSIZE,SIZE;
    void MakeRoomFor(int i);
  public:
    DynaSList();
    DynaSList(const DynaSList<T> &dl);
    DynaSList(int V_EXTSIZE);
    ~DynaSList();

    void insert(T element);
    void insert(DynaSList<T> dl);
    void del(int i);
    void delFrom(int i,int j);
    
    int Size() const;
    int Find(T &el) const;
    DynaSList<T> Take(int i);
    DynaSList<T> Drop(int i);
    
    T at(int i) const;
      // alternate notation for []-operator
    T operator[](int i) const;
    T &operator[](int i);
    DynaSList<T> &operator=(const DynaSList<T> &dl);
    //friend std::ostream &operator<<(std::ostream &s, DynaSList<T> &dl);
};

template <class T>
DynaSList<T>::DynaSList()
{ elements=0;
  contens=0;
  EXTSIZE=25;
  SIZE=0;
}

template <class T>
DynaSList<T>::DynaSList(const DynaSList<T> &dl)
{ if (dl.SIZE==0) elements=0;
  else elements=new T[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
}

template <class T>
DynaSList<T>::DynaSList(int V_EXTSIZE)
{ elements=0;
  contens=0;
  EXTSIZE=V_EXTSIZE;
  SIZE=0;
}

template <class T>
DynaSList<T>::~DynaSList()
{ delete[] elements;
}

template <class T>
void DynaSList<T>::MakeRoomFor(int i)
{ if ((contens+i)>SIZE) {
    SIZE=((contens+i)/EXTSIZE)*EXTSIZE;
    if ((contens+i)>SIZE) SIZE+=EXTSIZE;
    T *h=new T[SIZE];
    for(int j=0;j!=contens;j++) h[j]=elements[j];
    if (elements!=0) delete[] elements;
    elements=h;
  }
}

template <class T>
void DynaSList<T>::insert(T element)
{ int a=0,b=contens-1;
  if (b>=0) {
    if (element<=elements[a]) b=a;
    else if (elements[b]<=element) { a=b;b++; }
    else {
      while (b-a>1) {
         int h=(a+b)/2;
         if (elements[h]<=element) a=h;
         else b=h;
      }
    }
    if (elements[a]==element) return;
  } else b=0;
  MakeRoomFor(1);
  for(int h=contens;h!=b;h--) elements[h]=elements[h-1];
  elements[b]=element;
  contens++;
}

template <class T>
void DynaSList<T>::insert(DynaSList<T> dl)
{ T *h=elements;
  int c=contens;
  elements=0;SIZE=0;contens=0;
  MakeRoomFor(c+dl.contens);
  int i=0,j=0;
  while ((i!=c)&&(j!=dl.contens)) {
    if (h[i]<dl.elements[j]) { elements[contens]=h[i]; i++; }
    else if (dl.elements[j]<h[i]) { elements[contens]=dl.elements[j]; j++; }
    else { elements[contens]=h[i]; i++; j++; }
    contens++;
  }
  while (i!=c) { elements[contens]=h[i]; i++; contens++; }
  while (j!=dl.contens) { elements[contens]=dl.elements[j]; j++; contens++; }
  delete[] h;
}

template <class T>
void DynaSList<T>::del(int i)
{ for(int k=i+1;k!=contens;k++) elements[k-1]=elements[k];
  contens--;
}

template <class T>
void DynaSList<T>::delFrom(int i,int j)
{ for(int k=i+j;k!=contens;k++) elements[k-j]=elements[k];
  contens-=j;
} 

template <class T>
int DynaSList<T>::Size() const
{ return contens;
}

template <class T>
int DynaSList<T>::Find(T &el) const
{ int a=0,b=contens-1;
  if (b>=0) {
    if (el<=elements[a]) b=a;
    else if (elements[b]<=el) { a=b;b++; }
    else {
      while (b-a>1) {
         int h=(a+b)/2;
         if (elements[h]<=el) a=h;
         else b=h;
      }
    }
    if (elements[a]==el) return a;
  };
  return -1;
}

template <class T>
DynaSList<T> DynaSList<T>::Take(int i)
{ DynaSList<T> dl(EXTSIZE);
  if (i>contens) i=contens; 
  else if (i<=0) return dl;
   else {
    dl.MakeRoomFor(i);
    for(int k=0;k!=i;k++) dl.elements[k]=elements[k];
    dl.contens=i;
    return dl;
  }
}

template <class T>
DynaSList<T> DynaSList<T>::Drop(int i)
{ DynaSList<T> dl;
  if (i<0) i=0;
  else if (i>=contens) return dl;
  else {
    dl.MakeRoomFor(contens-i);
    for(int k=i;k!=contens;k++) dl.elements[k-i]=elements[k];
    dl.contens=contens-i;
    return dl;
  }
}

template <class T>
T DynaSList<T>::at(int i) const
{ return elements[i];
}

template <class T>
T DynaSList<T>::operator[](int i) const
{ return elements[i];
}

template <class T>
T &DynaSList<T>::operator[](int i)
{ return elements[i];
}

template <class T>
DynaSList<T> &DynaSList<T>::operator=(const DynaSList<T> &dl)
{ if (elements!=0) delete[] elements;
  if (dl.SIZE==0) elements=0;
  else elements=new T[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
  
  return *this;
}

template <class T>
std::ostream &operator<<(std::ostream &os,DynaSList<T> &dl)
{ if (dl.Size()==0) os<<"empty";
  else { 
    os<<"["<<dl[0];
    for(int k=1;k!=dl.Size();k++) os<<","<<dl[k];
    os<<"]";
  }
  return os;
}

//
//----------------------------------------------------------------------
//                       DynaPList<T>
//----------------------------------------------------------------------
// this class implements a dynamic list of pointers to objects of type T.
//
// actual extending the list takes place with blocks of size EXTSIZE.
// the sizes are never reduced, untill destruction of the object.
// EXTSIZE can (if desired) be set by construction, default is 25.
//
// while destructing an object of this class only the pointers are destroyed.
// NOT the elements they point to.

template<class T>
class DynaPList { // <T> moet weg! WW
  protected:
    T        **elements;
    int      contens,EXTSIZE,SIZE;
    void MakeRoomFor(int i);
  public:
    DynaPList();
    DynaPList(const DynaPList<T> &dl);
    DynaPList(int V_EXTSIZE);
    ~DynaPList();

    void cat(T *element);
    void cat(DynaPList<T> dl);
    void insertAt(T *element,int i);
    void insertAt(DynaPList<T> dl,int i);
    void del(int i);
    void delFrom(int i,int j);
    
    int Size() const;
    DynaPList<T> Take(int i);
    DynaPList<T> Drop(int i);
    
    T *at(int i) const;
    T *operator[](int i) const;
    DynaPList<T> &operator=(const DynaPList<T> &dl);
    //friend std::ostream &operator<<(std::ostream &s, DynaPList<T> &dl);
};

template<class T>
DynaPList<T>::DynaPList()
{ elements=0;
  contens=0;
  EXTSIZE=25;
  SIZE=0;
}

template<class T>
DynaPList<T>::DynaPList(const DynaPList<T> &dl)
{ if (dl.SIZE==0) elements=0;
  else elements=new T *[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
}

template<class T>
DynaPList<T>::DynaPList(int V_EXTSIZE)
{ elements=0;
  contens=0;
  EXTSIZE=V_EXTSIZE;
  SIZE=0;
}

template<class T>
DynaPList<T>::~DynaPList()
{ delete[] elements;
}

template<class T>
void DynaPList<T>::MakeRoomFor(int i)
{ if ((contens+i)>SIZE) {
    SIZE=((contens+i)/EXTSIZE)*EXTSIZE;
    if ((contens+i)>SIZE) SIZE+=EXTSIZE;
    T **h=new T *[SIZE];
    for(int j=0;j!=contens;j++) h[j]=elements[j];
    if (elements!=0) delete[] elements;
    elements=h;
  }
}

template<class T>
void DynaPList<T>::cat(T *element)
{ MakeRoomFor(1);
  elements[contens]=element;
  contens++;
}

template<class T>
void DynaPList<T>::cat(DynaPList<T> dl)
{ MakeRoomFor(dl.contens);
  for(int i=0;i!=dl.contens;i++) {
    elements[contens]=dl.elements[i];
    contens++;
  }
}

template<class T>
void DynaPList<T>::insertAt(T *element,int i)
{ MakeRoomFor(1);
  for(int j=contens;j!=i;j--) elements[j]=elements[j-1];
  elements[i]=element;
  contens++;
}

template<class T>
void DynaPList<T>::insertAt(DynaPList<T> dl,int i)
{ MakeRoomFor(dl.contens);
  for(int j=contens;j!=i;j--) elements[dl.contens+j-1]=elements[j-1];
  for(int j=0;j!=dl.contens;j++) elements[i+j]=dl.elements[j];
  contens+=dl.contens;
}

template<class T>
void DynaPList<T>::del(int i)
{ for(int k=i+1;k!=contens;k++) elements[k-1]=elements[k];
  contens--;
}

template<class T>
void DynaPList<T>::delFrom(int i,int j)
{ for(int k=i+j;k!=contens;k++) elements[k-j]=elements[k];
  contens-=j;
} 

template<class T>
int DynaPList<T>::Size() const
{ return contens;
}

template<class T>
DynaPList<T> DynaPList<T>::Take(int i)
{ DynaPList<T> dl(EXTSIZE);
  if (i>contens) i=contens; 
  else if (i<=0) return dl;
   else {
    dl.MakeRoomFor(i);
    for(int k=0;k!=i;k++) dl.elements[k]=elements[k];
    dl.contens=i;
    return dl;
  }
}

template<class T>
DynaPList<T> DynaPList<T>::Drop(int i)
{ DynaPList<T> dl;
  if (i<0) i=0;
  else if (i>=contens) return dl;
  else {
    dl.MakeRoomFor(contens-i);
    for(int k=i;k!=contens;k++) dl.elements[k-i]=elements[k];
    dl.contens=contens-i;
    return dl;
  }
}

template<class T>
T *DynaPList<T>::at(int i) const
{ if ((0<=i)&&(i<contens)) return elements[i];
  return 0;
}

template<class T>
T *DynaPList<T>::operator[](int i) const
{ if ((0<=i)&&(i<contens)) return elements[i];
  return 0;
}

template<class T>
DynaPList<T> &DynaPList<T>::operator=(const DynaPList<T> &dl)
{ if (elements!=0) delete[] elements;
  if (dl.SIZE==0) elements=0;
  else elements=new T *[dl.SIZE];
  for(int i=0;i!=dl.contens;i++) elements[i]=dl.elements[i];
  contens=dl.contens;
  EXTSIZE=dl.EXTSIZE;
  SIZE=dl.SIZE;
  
  return *this;
}

template<class T>
std::ostream &operator<<(std::ostream &os,DynaPList<T> &dl)
{ if (dl.Size()==0) os<<"empty";
  else { 
    os<<"["<<dl[0];
    for(int k=1;k!=dl.Size();k++) os<<","<<*dl[k];
    os<<"]";
  }
  return os;
}

#endif

