// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#ifndef __INT2_H
#define __INT2_H

#include <cmath>
#include <iostream>
#include <mymanip.h>

//-----------------------------------------------------------------------//
//                          int2                                         //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 2 integers

struct int2
{
  protected:
    int X[2];

  public:
    // constructors
    int2();
    int2(int, int);
    int2(const int2&);

    int Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..2)

    int& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..2)

    int operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..1)

    int& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..1)

    int& x() { return X[0]; }
    int& y() { return X[1]; }

    int Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double Dist(const int2& p) const;
    // returns the Euclidian distance from the point X to p

    int2 Unit() const;
    // returns the vector X / Length(X)

    void MakeZero();
    // makes the vector equal to (0,0)

    bool IsZero() const;
    // tests if the vector is equal to (0,0)

    void Negate();
    // replaces the vector with -X

    void  operator = (const int2&);
    int2  operator - () const;
    int2& operator += (const int2&);
    int2& operator -= (const int2&);
    int2& operator *= (int);

    friend int2 operator - (const int2&, const int2&);
    friend int2 operator + (const int2&, const int2&);
    friend int2 operator * (const int2&, int);
    friend int2 operator * (int, const int2&);
    friend int2 operator / (const int2&, int);
    friend int operator * (const int2&, const int2&); // inner product
    friend int operator , (const int2&, const int2&); // inner product
    friend int operator == (const int2&, const int2&);
    friend int operator != (const int2&, const int2&);
    friend double Dist(const int2&, const int2&);
    friend std::ostream& operator<<(std::ostream& s, const int2& p);
    friend std::istream& operator>>(std::istream& s, int2& p);
};

inline int2::int2 ()
{ X[0]=X[1]=0; }

inline int2::int2 (int x0, int x1)
{ X[0]=x0; X[1]=x1; }

inline int2::int2 (const int2& p)
{ X[0]=p[0]; X[1]=p[1]; }

inline void int2::operator = (const int2& p)
{ X[0]=p[0]; X[1]=p[1]; }

inline int2 int2::operator - () const
{ return int2(-X[0], -X[1]); }

inline int2& int2::operator += (const int2& p)
{ X[0]+=p[0]; X[1]+=p[1]; return *this; }

inline int2& int2::operator -= (const int2& p)
{ X[0]-=p[0]; X[1]-=p[1]; return *this; }

inline int2& int2::operator *= (int t)
{ X[0]*=t; X[1]*=t; return *this; }

inline int int2::Length2() const
{ return X[0]*X[0] + X[1]*X[1]; }

inline double int2::Length () const
{ return sqrt( (*this).Length2() ); }

inline int2 operator - (const int2& a, const int2& b)
{ return int2(a[0]-b[0], a[1]-b[1]); }

inline int2 operator + (const int2& a, const int2& b)
{ return int2(a[0]+b[0], a[1]+b[1]); }

inline double int2::Dist(const int2& p) const
{ return ((*this) - p).Length(); }

inline int2 operator * (const int2& a, int t)
{ return int2(t*a[0], t*a[1]); }

inline int2 operator * (int t, const int2& p)
{ return p * t; }

inline int operator * (const int2& a, const int2& b)
{ return a[0]*b[0] + a[1]*b[1]; }

inline int operator , (const int2& a, const int2& b)
{ return a*b; }

inline int operator == (const int2& a, const int2& b)
{ return (a[0] == b[0]) && (a[1] == b[1]); }

inline int operator != (const int2& a, const int2& b)
{ return !(a == b); }

inline double Dist(const int2& a, const int2& b)
{ return a.Dist(b); }

inline void int2::MakeZero()
{ X[0] = X[1] = 0; }

inline bool int2::IsZero() const
{ return X[0] == 0 && X[1] == 0; }

inline void int2::Negate ()
{ X[0] = -X[0]; X[1] = -X[1]; }

inline std::ostream& operator<<(std::ostream& s, const int2& a)
{ return s << '(' << a[0] << ',' << a[1] << ')'; }

inline std::istream& operator>>(std::istream& s, int2& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipbracket;
}

#endif

