// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef __LISTTMPL_H
#define __LISTTMPL_H

#include <iostream>

#define MaxListSize 10000

template <class X> class List
{
  private:
    X** Items;
    int Limit;    // currently allocated size (in elements) of the list
    int Delta;
    bool isowner; // if true, the destructor is called when deleting an element

  public:
    int Count; // current number of items in the list

    List();
    List(int ALimit, int ADelta, int owner=false);
    List(const List<X>& x);
    ~List();

    X* At(int Index) const;
    // returns a pointer to the Index-th element
    
    X& operator [] (int Index);
    // returns a reference to the Index-th element
    
    const X& operator [] (int Index) const;
    // returns a reference to the Index-th element
    
    X* AtDelete(int Index);
    // deletes the Index-th element from the list and
    // returns a pointer to the element

    void AtInsert(int Index, X* Item);
    // inserts X at position Index
    
    void AtPut(int Index, X* Item);
    // replaces the Index-th element with X
    
    void Delete(X* Item);
    // deletes X from the list; N.B. X must be in the list!!
    
    void DeleteAll();
    // deletes all elements from the list
    
    int IndexOf(X* Item) const;
    // returns the index of X; if X is not in the list, -1 is returned
    
    void Insert(X* Item);
    // inserts X at the end of the list
    
    void SetLimit(int ALimit);
    // sets Limit to min(Count, ALimit)
    
    void Show() const;
    // prints info on the number of elements etc.

    void operator+=(const List<X>& x);
    // adds the elements of x to this list

    void operator+=(X* x);
    // adds the element x to this list

  private:
    inline int InRange(int Index) const;

    void CopyList(const List<X>& x);
    // This method is called in the copy constructor, because otherwise the
    // Sun compiler doesn't compile it.
};

template <class X>
inline List<X>::List()
{
  isowner = false;
  Limit = 10;
  Delta = 10;
  Count = 0;
  Items = new X*[Limit];
}

template <class X>
inline List<X>::List(int ALimit, int ADelta, int owner)
{
  isowner = owner;
  if (ALimit > MaxListSize) ALimit = MaxListSize;
  Limit = ALimit;
  Delta = ADelta;
  Count = 0;
  Items = new X*[ALimit];
}

template <class X>
inline void List<X>::CopyList(const List<X>& x)
{
  for (int i=0; i<Count; i++) Items[i] = x.Items[i];
}

template <class X>
inline List<X>::List(const List<X>& x)
{
  Limit = x.Limit;
  Delta = x.Delta;
  Count = x.Count;
  Items = new X*[Limit];
  CopyList(x);
}

template <class X>
inline List<X>::~List()
{
  Limit = 0;
  Count = 0;
  delete[] Items;
}

template <class X>
inline X* List<X>::At(int Index) const
{
  if (InRange(Index)) return Items[Index];
  else return 0;
}

template <class X>
inline X& List<X>::operator [] (int Index)
{
  if (InRange(Index)) return *(Items[Index]);
  else return *(Items[0]);  // ????????????????????????
}

template <class X>
inline const X& List<X>::operator [] (int Index) const
{
  if (InRange(Index)) return *(Items[Index]);
  else return *(Items[0]);  // ????????????????????????
}
 
template <class X>
inline X* List<X>::AtDelete(int Index)
{
  X* ReturnValue = 0;
  
  if (InRange(Index)) {
  	ReturnValue = Items[Index];
  	if (isowner)
  	  delete Items[Index];
    for (int i=Index; i<Count-1; i++) Items[i] = Items[i+1];
    Count--;
  }

  return ReturnValue;
}

template <class X>
inline void List<X>::AtInsert(int Index, X* Item)
{
  if (Count >= Limit) SetLimit(Limit + Delta);
  for (int i=Count; i>Index; i--) Items[i] = Items[i-1];
  Items[Index] = Item;
  Count++;
}

template <class X>
inline void List<X>::AtPut(int Index, X* Item)
{
  if (InRange(Index)) {
  	if (isowner) 
  	  delete Items[Index];
    Items[Index] = Item;
  }
}

template <class X>
inline void List<X>::Delete(X* Item)
{
  AtDelete(IndexOf(Item));
}

template <class X>
inline void List<X>::DeleteAll()
{
  if (isowner) {
    for (int i=0; i<Count; i++)
      delete Items[i];
  }
  Count = 0;
}

template <class X>
inline int List<X>::IndexOf(X* Item) const
{
  int i = 0;
  for (; i<Count; i++)
    if (Items[i] == Item) break;
  if (i < Count) return i;
  else return -1;
}

template <class X>
inline void List<X>::Insert(X* Item)
{
  AtInsert(Count, Item);
}

template <class X>
inline void List<X>::SetLimit(int ALimit)
{
  if (ALimit < Count) ALimit = Count;
  if (ALimit > MaxListSize) {
    std::cerr << "List full!" << std::endl;
    ALimit = MaxListSize;
  }

  Limit = ALimit;
  X** OldItems = Items;
  Items = new X*[ALimit];
  for (int i=0; i<Count; i++) Items[i] = OldItems[i];

  delete OldItems;
}

template <class X>
inline int List<X>::InRange(int Index) const
{
  if ((0 <= Index) && (Index < Count))
    return (0 == 0);
  else {
    std::cerr << "Range check error! " << std::endl;
    return (0 == 1);
  }
}

template <class X>
inline void List<X>::Show() const
{
  std::cout << "Count = " << Count << " Limit = " << Limit << " Delta = " << Delta << std::endl;
}

template <class X>
inline void List<X>::operator+=(const List<X>& x)
{
  for (int i=0; i<x.Count; i++)
    Insert(x.At(i));
}

template <class X>
inline void List<X>::operator+=(X* x)
{
  Insert(x);
}

template<class T>
class SimpleArray {
  protected:
    T* a;
    int size;
  public:
    SimpleArray()      { size=0; a=0; }
    SimpleArray(int n) { size=n; a=new T[n]; }
    ~SimpleArray()     { if (a) delete[] a; }
    int Size() { return size; }
    T& operator[] (int i) { return a[i]; }
    void SetSize(int n) {
      if (a) delete[] a;
      size = n;
      a = new T[n];
      for (int i=0; i<n; i++) a[i]=0;
    }
};

#endif

