// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-25-1996

#ifndef __PARAMETRICSURFACE_H
#define __PARAMETRICSURFACE_H

#include <geotyp.h>
#include <sample.h>
#include <surfmesh.h>

//-----------------------------------------------------------------------//
//                          ParametricSurface                            //
//-----------------------------------------------------------------------//
class ParametricSurface
// this class represents a parametric surface F: omega in R^2 -> R^3
{
  public:
    ParametricSurface() {}
    // default constructor

    virtual ~ParametricSurface() {}
    // destructor

    virtual double3 GetDerivative(const double2& u, int k, int l) const;
    // returns the partial derivative
    //
    //              d^(k+l) F(u,v)
    //              --------------
    //                du^k dv^l

    virtual double3 GetDirectionalDerivative(const double2& u, const double2& d, int k) const;
    // returns the directional derivative
    //
    //               (k)
    //              g   (t)             with g(t) = F(u+td)

    virtual void GetFuncSamples(const double2& u, FuncSample2D& F1, FuncSample2D& F2, FuncSample2D& F3) const;
    // computes samples of the coordinate functions F1, F2 and F3 at the point with parameter u
    
    virtual double3 GetNormal(const double2& u) const;
    // returns the normal of the surface at the point with parameter u

    virtual double3 GetOffsetPoint(const double2& u, double d) const;
    // returns F(u) + d*N(u), with N the unit normal

    virtual double3 GetPoint(const double2& u) const;
    // returns the point of the surface with parameter u

    virtual SurfSample GetSurfSample(const double2& u) const;
    // computes a sample of F at the point with parameter u

    void GetSurfSampleExt(const double2& u, SurfSampleExt& sample) const;
    // computes a sample of F at u with some extra info

    virtual Plane3D GetTangentPlane(const double2& u) const;
    // returns the tangent plane at the point with parameter u
    
    virtual double GetTraceExpr(const double2& u, const FuncSample2D& f) const;
    // returns the following expression:
    //              2   2 
    //     trace( (D  f)  ) * sqrt(g_F) (u)
    //              F
    
    double SimpleStretchEnergy(const double2& u) const;
    //  returns <Fu|Fu> + <Fv|Fv> (u)

    double DataStretchEnergy(const ParametricSurface& G, const double2& u) const;
    //                   -1
    //  returns  trace( I   I  ) sqrt(g ) (u)
    //                   G   F         G
    
    double ExactStretchEnergy(const double2& u) const;
    //           
    //  returns  sqrt(g ) (u)
    //                 F

    double SimpleBendEnergy(const double2& u) const;
    //  returns <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> (u)

    double DataBendEnergy(const ParametricSurface& G, const double2& u) const;
    //            3           2     2
    //  returns  sum trace( (D  F_i)  ) sqrt(g ) (u)
    //           i=1          G               G
    
    double ExactBendEnergy(const double2& u) const;
    //                                                
    //  returns  (kappa1)^2 + (kappa2)^2  sqrt(g ) (u)
    //                                          F       

    double PointDistance(const double2& u, const double3& point) const;
    // returns  d(F(u), point)^2

    double LineDistance(const double2& u, const Line3D& line) const;
    // returns  d(F(u), line)^2

    double PlaneDistance(const double2& u, const Plane3D& plane) const;
    // returns  d(F(u), plane)^2

    double NormalDistance(const double2& u, const double3& normal) const;
    // returns  (Fu(u),normal)^2 + (Fv(u),normal)^2
};

class SimpleStretchEnergyFunction {
  protected:
    const ParametricSurface& F;
  public:
    SimpleStretchEnergyFunction(const ParametricSurface& F0): F(F0) {}
    double operator() (const double2& u) { return F.SimpleStretchEnergy(u); }
};

class DataStretchEnergyFunction {
  protected:
    const ParametricSurface& F;
    const ParametricSurface& G;
  public:
    DataStretchEnergyFunction(const ParametricSurface& F0, const ParametricSurface& G0): F(F0), G(G0) {}
    double operator() (const double2& u) { return F.DataStretchEnergy(G,u); }
};

class ExactStretchEnergyFunction {
  protected:
    const ParametricSurface& F;
  public:
    ExactStretchEnergyFunction(const ParametricSurface& F0): F(F0) {}
    double operator() (const double2& u) { return F.ExactStretchEnergy(u); }
};

class SimpleBendEnergyFunction {
  protected:
    const ParametricSurface& F;
  public:
    SimpleBendEnergyFunction(const ParametricSurface& F0): F(F0) {}
    double operator() (const double2& u) { return F.SimpleBendEnergy(u); }
};

class DataBendEnergyFunction {
  protected:
    const ParametricSurface& F;
    const ParametricSurface& G;
  public:
    DataBendEnergyFunction(const ParametricSurface& F0, const ParametricSurface& G0): F(F0), G(G0) {}
    double operator() (const double2& u) { return F.DataBendEnergy(G,u); }
};

class ExactBendEnergyFunction {
  protected:
    const ParametricSurface& F;
  public:
    ExactBendEnergyFunction(const ParametricSurface& F0): F(F0) {}
    double operator() (const double2& u) { return F.ExactBendEnergy(u); }
};

class PointDistanceFunction {
  protected:
    const ParametricSurface& F;
    const double3& point;
  public:
    PointDistanceFunction(const ParametricSurface& F0, const double3& p): F(F0), point(p) {}
    double operator() (const double2& u) { return F.PointDistance(u,point); }
};

class LineDistanceFunction {
  protected:
    const ParametricSurface& F;
    const Line3D& line;
  public:
    LineDistanceFunction(const ParametricSurface& F0, const Line3D& l): F(F0), line(l) {}
    double operator() (const double2& u) { return F.LineDistance(u,line); }
};

class PlaneDistanceFunction {
  protected:
    const ParametricSurface& F;
    const Plane3D& plane;
  public:
    PlaneDistanceFunction(const ParametricSurface& F0, const Plane3D& p): F(F0), plane(p) {}
    double operator() (const double2& u) { return F.PlaneDistance(u,plane); }
};

class NormalDistanceFunction {
  protected:
    const ParametricSurface& F;
    const double3& normal;
  public:
    NormalDistanceFunction(const ParametricSurface& F0, const double3& n): F(F0), normal(n) {}
    double operator() (const double2& u) { return F.NormalDistance(u,normal); }
};

#endif

