// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 22-8-95

#ifndef __POLYNOMIAL_H
#define __POLYNOMIAL_H

#include <iostream>
#include "dynalist.h"
//
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
// this class implements a polynomial of arbitrary degree. 
// The polynomial may be seen in two ways:
//    (sum i:0<=i<=n:AiX^i) with Ai of type T and X of type S
//                       or with Ai of type T and X of type T
//

template <class T,class S>
class Polynomial { // <T,S> moet weg! WW
  protected:
    DynaList<T> coefs;
    int         degree;
  public:
    char varID;    // character ter representatie van X bij uitvoer naar
                   // std::ostream. Default is 'X'. Door de gebruiker zelf in
                   // te stellen.
    Polynomial();
    Polynomial(const Polynomial<T,S> &pl);
    Polynomial(T cfs[],int deg);
      // cfs is the array of coefficients Ai. The lenght must be
      // sufficient. Deg is the degree of the polynomial
    ~Polynomial();
    
    int Degree() ;
      // returns the degree of the polynomial
    T Val(S x) ;
      // evaluates the polynomial in x
    
    T Integral(S f,S t) ;
      // berekent de volgende integraal voor X van type S:
      //        t
      //        -
      //       |  (sum i:0<=i<=n:Ai X^i) dX
      //      -
      //      f
      //
    T SIntegral(T f,T t) ;
      // berekent de volgende integraal voor X van type T:
      //        t
      //        -
      //       |  (sum i:0<=i<=n:Ai X^i) dX
      //      -
      //      f
      //
    Polynomial Integral() ;
      // geeft symbolische representatie van
      //        t
      //        -
      //       |  (sum i:0<=i<=n:Ai X^i) dX
      //      -
      //      f
      //
      
    T Derivative(S x) ;
      // berekent de afgeleide van het polynoom in x
      
    Polynomial Derivative() ;
      // geeft symbolische representatie van de afgeleide van het polynoom
    
    Polynomial operator+(const Polynomial<T,S> &pl) ;
      // optellen van polynomen (dus diens coefficienten)
    Polynomial operator-(const Polynomial<T,S> &pl) ;
      // aftrekken van polynomen
    Polynomial operator-() ;
      // zelfde polynoom met omgekeerde coefficienten
    Polynomial operator*(const Polynomial<T,S> &pl) const;
      // vermenigvuldiging van polynomen
    Polynomial operator*(const T &c) ;
      // vermenigvuldiging met constante c van type T van alle coefficienten
    Polynomial &operator=(const Polynomial<T,S> &pl);
    T &operator[](int i) ;
      // geef referentie naar coefficient i van het polynoom
    friend Polynomial<T,S> operator*(const T &c,const Polynomial<T,S> &pl) ; // <T,S> toegevoegd WW
      // vermenigvuldiging met constante c van type T van alle coefficienten
    friend std::ostream &operator<<(std::ostream &os,const Polynomial<T,S> &pl);
      // geformatteerde uitvoer naar std::ostream.
};

template <class T,class S>
Polynomial<T,S>::Polynomial()
{ degree=-1;
  varID='X';
}

template <class T,class S>
Polynomial<T,S>::Polynomial(const Polynomial<T,S> &pl)
{ coefs=pl.coefs;
  degree=pl.degree;
  varID=pl.varID;
}

template <class T,class S>
Polynomial<T,S>::Polynomial(T cfs[],int deg)
{ degree=deg;
  for(int i=0;i<=deg;i++) coefs.cat(cfs[i]);
  varID='X';
}

template <class T,class S>
Polynomial<T,S>::~Polynomial()
{ }

template<class T,class S>
int Polynomial<T,S>::Degree() 
{ return degree;
}

template<class T,class S>
T Polynomial<T,S>::Val(S x) 
{ if (degree<0) return T();
  T result(coefs[0]);
  S factor=x;
  for(int i=1;i<=degree;i++) {
    result=result+coefs[i]*x;
    x=x*factor;
  }
  return result;
}

template<class T,class S>
T Polynomial<T,S>::Integral(S f,S t) 
  // returns Integral().Val(t)-Integral().Val(s)
{ if (degree<0) return T();
  S ff=f,ft=t;
  T result=coefs[0]*(ft-ff);
  for(int i=1;i<=degree;i++) {
    ff=ff*f;ft=ft*t;
    result=result+coefs[i]*(ft-ff)*1.0/(i+1);
  }
  return result;
}

template<class T,class S>
T Polynomial<T,S>::SIntegral(T f,T t) 
  // returns Integral().Val(t)-Integral().Val(s)
{ if (degree<0) return T();
  T ff=f,ft=t;
  T result=coefs[0]*(ft-ff);
  for(int i=1;i<=degree;i++) {
    ff=ff*f;ft=ft*t;
    result=result+(coefs[i]*(ft-ff))*(1.0/(i+1));
  }
  return result;
}

template<class T,class S>
Polynomial<T,S> Polynomial<T,S>::Integral() 
{ if (degree<0) return Polynomial<T,S>();
  Polynomial<T,S> result;
  result.degree=degree+1;
  result.coefs.cat(T());
  for(int i=0;i<=degree;i++) result.coefs.cat(coefs[i]*(1.0/(i+1)));
  return result;
}

template<class T,class S>
T Polynomial<T,S>::Derivative(S x) 
{ if (degree<1) return T();
  T result(coefs[1]);
  S factor=x;
  for(int i=2;i<=degree;i++) {
    result=result+coefs[i]*factor*i;
    factor=factor*x;
  }
  return result;
}

template<class T,class S>
Polynomial<T,S> Polynomial<T,S>::Derivative() 
{ if (degree<1) return Polynomial<T,S>();
  Polynomial<T,S> result;
  result.degree=degree-1;
  for(int i=1;i<=degree;i++) result.coefs.cat(coefs[i]*i);
  return result;
}

template <class T,class S>
Polynomial<T,S> Polynomial<T,S>::operator+(const Polynomial<T,S> &pl) 
{ int max=((degree<pl.degree) ? (pl.degree) : (degree)),
      min=((degree<pl.degree) ? (degree) : (pl.degree)); 
  int deg=max;
  T *cfs=new T[deg+1];
  int i=0;
  while (i<=min) {
    cfs[i]=coefs[i]+pl.coefs[i];
    i++;
  }
  if (degree<pl.degree) while(i<=deg) { cfs[i]=pl.coefs[i];i++; }
  else while(i<=deg) { cfs[i]=coefs[i];i++; }
  Polynomial<T,S> result(cfs,deg);
  result.varID=varID;
  delete[] cfs;
  return result;
}

template <class T,class S>
Polynomial<T,S> Polynomial<T,S>::operator-(const Polynomial<T,S> &pl) 
{ int max=(degree<pl.degree ? pl.degree : degree),
      min=(degree<pl.degree ? degree : pl.degree); 
  int deg=max;
  T *cfs=new T[deg+1];
  int i=0;
  while (i<=min) {
    cfs[i]=coefs[i]-pl.coefs[i];
    i++;
  }
  if (degree<pl.degree) while(i<=deg) cfs[i]=-pl.coefs[i];
  else while(i<=deg) cfs[i]=coefs[i];
  Polynomial<T,S> result(cfs,deg);
  result.varID=varID;
  delete[] cfs;
  return result;
}

template <class T,class S>
Polynomial<T,S> Polynomial<T,S>::operator-() 
{ Polynomial<T,S> result();
  result.varID=varID;
  result.degree=degree;
  for(int i=0;i<=degree;i++) result.coefs.cat(-coefs[i]);
  return result;
}

template <class T,class S>
Polynomial<T,S> Polynomial<T,S>::operator*(const Polynomial<T,S> &pl) const
{ int deg=degree+pl.degree;
  T *cfs=new T[deg+1];
  for(int i=0;i<=deg;i++) {
    // strange initialization of inner loop is necessary to avoid
    // tot=0 statement for a 0-polynomial. This would create a type-conflict
    int min=(i>pl.degree ? i-pl.degree : 0),
        max=(i>degree ? degree : i);
    T tot=coefs[min]*pl.coefs[i-min];
    for(int k=min+1;k<=max;k++) {
      tot=tot+coefs[k]*pl.coefs[i-k];
    }
    cfs[i]=tot;
  }
  Polynomial<T,S> result(cfs,deg);
  result.varID=varID;
  delete[] cfs;
  return result;
}

template <class T,class S>
Polynomial<T,S> Polynomial<T,S>::operator*(const T &c) 
{ Polynomial<T,S> p(*this);
  for (int i=0;i<=degree;i++) {
    p.coefs[i]=coefs[i]*c;
  }
  return p;
}

template <class T,class S>
Polynomial<T,S> &Polynomial<T,S>::operator=(const Polynomial<T,S> &pl)
{ coefs=pl.coefs;
  degree=pl.degree;
  varID=pl.varID;
  
  return *this;
}

template <class T,class S>
T &Polynomial<T,S>::operator[](int i) 
{ return coefs[i];
}

template <class T,class S>
Polynomial<T,S> operator*(const T &c,const Polynomial<T,S> &pl)
{ Polynomial<T,S> p(pl);
  for (int i=0;i<=pl.degree;i++) {
    p.coefs[i]=pl.coefs[i]*c;
  }
  return p;
}

template <class T,class S>
std::ostream &operator<<(std::ostream &os,const Polynomial<T,S> &pl)
{ if (pl.degree>=0) {
    os<<"("<<pl.coefs[0];
    for(int i=1;i<=pl.degree;i++) os<<"+"<<pl.coefs[i]<<pl.varID<<"^"<<i;
    os<<")";
  }
  return os;
}

#endif
