// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-04-1996

#ifndef __QUADRATICFUNCTION_H
#define __QUADRATICFUNCTION_H

#include <function.h>
#include <numtyp.h>
#include <tuple.h>
#include <dvector.h>
#include <dmatrix.h>

enum SolveMethod {
  BKP_method,
  CH_method,
  LDL_method,
  LU_method,
  QR_method,
  CG_method
};

//-----------------------------------------------------------------------//
//                          QuadraticFunction                            //
//-----------------------------------------------------------------------//
// The class QuadraticFunction represents the function
//
//       f(x) = x^T A x + B^T x + C
//
// A and B are stored as a pointers to a SparseMatrix/SparseVector.
// These pointers may be 0, to indicate a null matrix/null vector.

class QuadraticFunction: public Function
{
  protected:
    SparseMatrix* A;
    SparseVector* B;
    double C;
    int isowner;     // indicates if this class 'owns' A and B
    
  public:
    QuadraticFunction(): A(0), B(0), C(0), isowner(0) {}
    // default constructor

    QuadraticFunction(SparseMatrix* A0, SparseVector* B0, double C0, int owner=1):
       A(A0), B(B0), C(C0), isowner(owner) {}
    // constructor

    ~QuadraticFunction() { if (isowner) {delete A; delete B;} }
    // destructor

    SparseMatrix* GetA() { return A; }
    // returns the matrix A
    
    SparseVector* GetB() { return B; }
    // returns the vector B
    
    double GetC() const { return C; }
    // returns the constant C

    void GetMatrix(DoubleMatrix&) const;
    // returns the matrix A

    void GetVector(DoubleVector&) const;
    // returns the vector B

    double val(const DoubleVector& x) const;
    DoubleVector grad(const DoubleVector& x) const;
    DoubleMatrix hess(const DoubleVector& x) const;
    double hessleftright(const DoubleVector& x, const DoubleVector& d, const DoubleVector& e) const;

    int dimension() const;
    virtual void operator*= (double t) { if (A) (*A) *= t; if (B) (*B) *= t; C *= t; }

    void GetMinimum(DoubleVector& X, SolveMethod method=LU_method) const;
    // computes the minimum of the function f; returns 0 if no minimum is found
    // PRECONDITION: -> A is symmetric and positive definite

    void GetConstrainedMinimum(const DoubleMatrix& F, const DoubleVector& G, DoubleVector& X, SolveMethod method=LU_method) const;
    // computes the minimum of the function f under the following constraints:
    // min   f(X)  =  X^T A X  +  B^T X  +  C
    // s.t.  X = F Y + G

    void GetConstrainedMinimumLagrange(const DoubleMatrix& D, const DoubleVector& E,
      DoubleVector& X, DoubleVector& Lambda, SolveMethod method = LU_method) const;
    // computes the minimum of the function f under the following constraints:
    // min   f(X)  =  X^T A X  +  B^T X  +  C
    // s.t.  D X = E
    //
    // The vector Lambda contains the Lagrange multipliers.
    //
    // The first order necessary conditions are:
    //
    // [  A + A^T    D^T ] [  X   ]         [ -B ]
    // [                 ] [      ]    =    [    ]
    // [   D         0   ] [Lambda]         [  E ]
    //

  friend class QuadraticFunctionList;
};

//-----------------------------------------------------------------------//
//                          ListQuadraticFunctionVector                  //
//-----------------------------------------------------------------------//
// The class ListQuadraticFunctionVector is a SparseVector representation
// of the vector B of a list of quadratic functions f.

class ListQuadraticFunctionVector: public SparseVector
{
  protected:
    List<QuadraticFunction>& f;

  public:
    ListQuadraticFunctionVector(List<QuadraticFunction>& f0): f(f0) {}
    void AddToVector(DoubleVector& Y) const;
    void MultAddToVector(double t, DoubleVector&) const;
    double GetInnerProduct(const DoubleVector& Y) const;
    void operator*= (double t);
    int NumRows() const;
    void Show() const;
    void Read(std::ifstream& from) {}
    void Write(std::ofstream& to) const {}
};

//-----------------------------------------------------------------------//
//                          ListQuadraticFunctionMatrix                  //
//-----------------------------------------------------------------------//
// The class ListQuadraticFunctionVector is a SparseMatrix representation
// of the matrix A of a list of quadratic functions f.

class ListQuadraticFunctionMatrix: public SparseMatrix
{
  protected:
    List<QuadraticFunction>& f;

  public:
    ListQuadraticFunctionMatrix(List<QuadraticFunction>& f0): f(f0) {}
    void AddToMatrix(DoubleMatrix& B) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t);
    int NumRows() const;
    int NumCols() const;
};

//-----------------------------------------------------------------------//
//                          QuadraticFunctionList                        //
//-----------------------------------------------------------------------//
//
// This class represents a sum of sparse quadratic functions f[0] + f[1] + ...

class QuadraticFunctionList: public QuadraticFunction {
  protected:
    List<QuadraticFunction> f;
    
  public:
    QuadraticFunctionList() { A = new ListQuadraticFunctionMatrix(f); B = new ListQuadraticFunctionVector(f); C=0; }
    // constructor

    QuadraticFunctionList(QuadraticFunction* f0)
      { A = new ListQuadraticFunctionMatrix(f); B = new ListQuadraticFunctionVector(f); C=0; f += f0; }
    // constructor

    QuadraticFunctionList(List<QuadraticFunction>& list)
      { A = new ListQuadraticFunctionMatrix(f); B = new ListQuadraticFunctionVector(f); C=0; f = list; }
    // constructor

    QuadraticFunctionList(const QuadraticFunctionList& f0): f(f0.f)
      { A = new ListQuadraticFunctionMatrix(f);
        B = new ListQuadraticFunctionVector(f); C=0; }
    // copy constructor

    ~QuadraticFunctionList() {}
    // destructor

    double GetC() const;
    // returns the constant C

    double val(const DoubleVector& x) const;
    DoubleVector grad(const DoubleVector& x) const;
    DoubleMatrix hess(const DoubleVector& x) const;
    int dimension() const;
    // returns the dimension of f[0]; if the list f is empty, 0 is returned

    void operator+= (QuadraticFunction* g) { f += g; }
    void operator+= (Tuple<QuadraticFunction*>& tuple) { f += tuple[0]; f += tuple[1]; }
    void operator+= (List<QuadraticFunction>& list) { f += list; }
    void operator+= (QuadraticFunctionList& g) { f += g.f; }
    void operator*= (double t);

    QuadraticFunction& operator[] (int i) { return f[i]; }
    QuadraticFunction* At(int i) { return f.At(i); }
    int Size() { return f.Count; }
    void DeleteAll() { f.DeleteAll(); }
};

#endif

