// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-12-1996

#ifndef __TRAPEZD_H
#define __TRAPEZD_H

#include <geotyp.h>

class HorizontalTrapezoid2D;
class VerticalTrapezoid2D;

//-----------------------------------------------------------------------//
//                          VerticalTrapezoid2D                          //
//-----------------------------------------------------------------------//
// The class VerticalTrapezoid2D represents the trapezoid given by
//
//  { (x,y) in R^2 | (x0 <= x <= x1)  && (a0 + a1 x <= y <= b0 + b1 x) } 

class VerticalTrapezoid2D {
  public:
    double x0, x1;
    double a0, a1;
    double b0, b1;

  public:
    VerticalTrapezoid2D(): x0(0), x1(1), a0(0), a1(0), b0(1), b1(0) {}
    // default constructor

    VerticalTrapezoid2D(double x00, double x10, double a00, double a10, double b00, double b10):
       x0(x00), x1(x10), a0(a00), a1(a10), b0(b00), b1(b10) {}
    // constructor

    VerticalTrapezoid2D(const double2& left, double leftheight, const double2& right, double rightheight);
    // constructor
    
    VerticalTrapezoid2D(const VerticalTrapezoid2D& s)
      { x0=s.x0; x1=s.x1; a0=s.a0; a1=s.a1; b0=s.b0; b1=s.b1; }
    // copy constructor

    void operator= (const VerticalTrapezoid2D& s)
      { x0=s.x0; x1=s.x1; a0=s.a0; a1=s.a1; b0=s.b0; b1=s.b1; }
    // assignment operator

    void Translate(double2 d)
      { x0 += d.x(); x1 += d.x(); a0 += (d.y() - a1*d.x()); b0 += (d.y() - b1*d.x()); }
    // translates the trapezoid over the vector d

    void Scale(double t);
    // scales the trapezoid with factor t w.r.t. its center

    double2 Center() const
      { return double2( (x0+x1)/2, (a0+b0)/2 + (a1+b1)*(x0+x1)/4 ); }
    // returns the center of gravity of the trapezoid

    double2 Point(int i) const;
    // returns the i-th corner point (i=0..3)

    double LeftHeight() const { return (b0+b1*x0) - (a0+a1*x0); }
    // returns the height of the left side

    double RightHeight() const { return (b0+b1*x1) - (a0+a1*x1); }
    // returns the height of the left side

    double Area() const { return (x1-x0)*(LeftHeight() + RightHeight())/2; }
    // returns the area of the trapezoid

    void Show() const;
    // writes the contents to standard out
    
    int GetTop() const;
    // returns the index of the point with the highest y-value
    
    int GetBottom() const;
    // returns the index of the point with the highest y-value

    HorizontalTrapezoid2D GetHorizontalTrapezoid(double ymin, double ymax, const int2& left, const int2& right) const;
    // returns the trapezoid bounded by the lines
    //
    //               y=ymin
    //               y=ymax
    //
    // and the line segments
    //
    //      [ Point(left[0]), Point(left[1]) ]
    //      [ Point(right[0]), Point(right[1]) ]
    
    List<HorizontalTrapezoid2D> GetHorizontalTrapezoids(int SplitAtIntegers = 0, double precision = 1e-10) const;
    // splits the trapezoid in horizontal trapezoids by cutting it with
    // horizontal lines through the vertices
    //
    // SplitAtIntegers: if true, then the trapezoid is also cut at horizontal
    //                  lines y=yi, with yi an integer
    // 
    // precision: trapezoids with (height < precision) are neglected

    friend std::ostream& operator<<(std::ostream& os, const VerticalTrapezoid2D& s);
    friend std::istream& operator>>(std::istream& is, VerticalTrapezoid2D& s);
    
  private:
    double x(int i) const { return Point(i).x(); }
    double y(int i) const { return Point(i).y(); }
    int NumPoints() const { return 4; }
};

inline std::ostream& operator<<(std::ostream& os, const VerticalTrapezoid2D& s)
{
  return os << "VerticalTrapezoid" << std::endl;
}

inline std::istream& operator>>(std::istream& is, VerticalTrapezoid2D& s)
{
  return is;
}

//-----------------------------------------------------------------------//
//                          HorizontalTrapezoid2D                        //
//-----------------------------------------------------------------------//
// The class HorizontalTrapezoid2D represents the trapezoid given by
//
//  { (x,y) in R^2 | (y0 <= y <= y1)  && (a0 + a1 y <= x <= b0 + b1 y) } 

class HorizontalTrapezoid2D {
  public:
    double y0, y1;
    double a0, a1;
    double b0, b1;

  public:
    HorizontalTrapezoid2D(): y0(0), y1(1), a0(0), a1(0), b0(1), b1(0) {}
    // default constructor

    HorizontalTrapezoid2D(double y00, double y10, double a00, double a10, double b00, double b10):
       y0(y00), y1(y10), a0(a00), a1(a10), b0(b00), b1(b10) {}
    // constructor

    HorizontalTrapezoid2D(const double2& top, double topwidth, const double2& bottom, double bottomwidth);
    // constructor
    
    HorizontalTrapezoid2D(const HorizontalTrapezoid2D& s)
      { y0=s.y0; y1=s.y1; a0=s.a0; a1=s.a1; b0=s.b0; b1=s.b1; }
    // copy constructor

    void operator= (const HorizontalTrapezoid2D& s)
      { y0=s.y0; y1=s.y1; a0=s.a0; a1=s.a1; b0=s.b0; b1=s.b1; }
    // assignment operator

    void Translate(double2 d)
      { a0 = a0+d.x()-a1*d.y(); b0 = b0+d.x()-b1*d.y(); y0 += d.y(); y1 += d.y(); }
    // translates the trapezoid over the vector d

    void Scale(double t);
    // scales the trapezoid with factor t w.r.t. its center

    double2 Center() const
      { return double2( (y0+y1)/2, (a0+b0)/2 + (a1+b1)*(y0+y1)/4 ); }
    // returns the center of gravity of the trapezoid

    double2 Point(int i) const;
    // returns the i-th corner point (i=0..3)

    double TopWidth() const { return (b0+b1*y0) - (a0+a1*y0); }
    // returns the width of the top side

    double BottomWidth() const { return (b0+b1*y1) - (a0+a1*y1); }
    // returns the width of the top side

    double Area() const { return (y1-y0)*(TopWidth() + BottomWidth())/2; }
    // returns the area of the trapezoid

    void Show() const;
    // writes the contents to standard out

    int GetRight() const;
    // returns the index of the point with the highest x-value
    
    int GetLeft() const;
    // returns the index of the point with the lowest x-value

    VerticalTrapezoid2D GetVerticalTrapezoid(double xmin, double xmax, const int2& bottom, const int2& top) const;
    // returns the trapezoid bounded by the lines
    //
    //               x=xmin
    //               x=xmax
    //
    // and the line segments
    //
    //      [ Point(bottom[0]), Point(bottom[1]) ]
    //      [ Point(top[0]), Point(top[1]) ]
    
    List<VerticalTrapezoid2D> GetVerticalTrapezoids(int SplitAtIntegers = 0, double precision = 1e-10) const;
    // splits the trapezoid in vertical trapezoids by cutting it with
    // vertical lines through the vertices
    //
    // SplitAtIntegers: if true, then the trapezoid is also cut at vertical
    //                  lines x=xi, with xi an integer
    // 
    // precision: trapezoids with (width < precision) are neglected

    friend std::ostream& operator<<(std::ostream& os, const HorizontalTrapezoid2D& s);
    friend std::istream& operator>>(std::istream& is, HorizontalTrapezoid2D& s);
    
  private:
    double x(int i) const { return Point(i).x(); }
    double y(int i) const { return Point(i).y(); }
    int NumPoints() const { return 4; }
};

inline std::ostream& operator<<(std::ostream& os, const HorizontalTrapezoid2D& s)
{
  return os << "HorizontalTrapezoid" << std::endl;
}

inline std::istream& operator>>(std::istream& is, HorizontalTrapezoid2D& s)
{
  return is;
}

#endif

