// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-16-1996

#ifndef __VARIATIONALTENSORBSPLINE_H
#define __VARIATIONALTENSORBSPLINE_H

#include <tbspline.h>
#include <dmatrix.h>
#include <function.h>
#include <bspline.h>
#include <equation.h>
#include <quadfunc.h>
#include <geotyp.h>
#include <polygon.h>

//-----------------------------------------------------------------------//
//                          VariationalTensorBSpline                     //
//-----------------------------------------------------------------------//
class VariationalTensorBSpline: public TensorBSpline
// This class adds a number methods to a TensorBSpline for computing
// constraint and operator expressions
{
  public:
    VariationalTensorBSpline() {}
    // default constructor

    VariationalTensorBSpline(int nrows, int ncols): TensorBSpline(nrows, ncols) {}
    // constructor

    VariationalTensorBSpline(char* FileName): TensorBSpline(FileName) {}
    // constructor (reads data from file)

    void WriteInfo() const;
    // writes info of the surface to standard output
    
    // The following methods all return a quadratic function f of the form
    //
    //            n-1  n-1                       n-1                  
    //    f(P) =  sum  sum  A_ij < P_i | P_j > + sum < B_i | P_i > + C
    //            i=0  j=0                       i=0
    //
    // Notations:
    //
    //    F:    this surface
    //    G:    a given reference surface
    //    n:    the number of control points of F
    //  P_i:    the i-th control point of F
    //    P:    concatenation vector of the control points P_i
    //  I_F:    first fundamental form of F
    //  I_G:    first fundamental form of G
    //  g_G:    determinant of I_G
    
    QuadraticFunction* GetSimpleStretchEnergy() const;
    // computes a quadratic function f such that
    //
    //           -
    //  f(P) =  |  <Fu|Fu> + <Fv|Fv> du dv
    //         -
    
    QuadraticFunction* GetSimpleBendEnergy() const;
    // computes a quadratic function f such that
    //
    //           -
    //  f(P) =  |  <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
    //         -
    
    QuadraticFunctionList GetDataStretchEnergy(const VariationalTensorBSpline& G, int degree=3) const;
    // computes a list of quadratic functions f such that
    //
    //           -         -1
    //  f(P) =  |  trace( I   I  ) sqrt(g ) du dv
    //         -           G   F         G
    //
    // using numerical integration (valid degrees are 3,5,7 and 9)
    
    QuadraticFunctionList GetDataBendEnergy(const VariationalTensorBSpline& G, int degree=3) const;
    // computes a list of quadratic functions f such that
    //
    //           -  3           2     2
    //  f(P) =  |  sum trace( (D  F_i)  ) sqrt(g ) du dv
    //         -   i=1          G               G
    //
    // using numerical integration (valid degrees are 3,5,7 and 9)
    
    QuadraticFunction* GetPointToPointAttractor(const double2& u, const double3& point) const;
    // computes a quadratic function f such that
    //
    //                        2
    //  f(P) = d(F(u,v),point)

    QuadraticFunction* GetPointToLineAttractor(const double2& u, const Line3D& line) const;
    // computes a quadratic function f such that
    //
    //                       2
    //  f(P) = d(F(u,v),line)
    
    QuadraticFunction* GetPointToPlaneAttractor(const double2& u, const Plane3D& plane) const;
    // computes a quadratic function f such that
    //
    //                        2
    //  f(P) = d(F(u,v),plane)   

    QuadraticFunction* GetNormalToVectorAttractor(const double2& u, const double3& normal) const;
    // computes a quadratic function f such that
    //
    //                          2                      2
    //  f(P) = (Fu(u,v), normal)   +  (Fv(u,v), normal) 
    //

    QuadraticFunction* GetRegionToPointAttractor(const Polygon2D& gamma, const double3& point) const;
    // computes a quadratic function f such that
    //
    //             -                  2
    //  f(P) =    |    d(F(u,v),point)  du dv
    //           -
    //         gamma
    //
    // where gamma is a convex polygon

    QuadraticFunction* GetRegionToLineAttractor(const Polygon2D& gamma, const Line3D& line) const;
    // computes a quadratic function f such that
    //
    //             -                 2
    //  f(P) =    |    d(F(u,v),line)  du dv
    //           -
    //         gamma
    //
    // where gamma is a convex polygon

    QuadraticFunction* GetRegionToPlaneAttractor(const Polygon2D& gamma, const Plane3D& plane) const;
    // computes a quadratic function f such that
    //
    //             -                  2
    //  f(P) =    |    d(F(u,v),plane)  du dv
    //           -
    //         gamma
    //
    // where gamma is a convex polygon

    QuadraticFunction* GetNormalRegionToVectorAttractor(const Polygon2D& gamma, const double3& normal) const;
    // computes a quadratic function f such that
    //
    //             -                   2                      2  
    //  f(P) =    |   (Fu(u,v), normal)   +  (Fv(u,v), normal)   du dv
    //           -
    //         gamma

    DivFunction* GetPointToPointRepellor(const double2& u, const double3& point) const
      { return new DivFunction(GetPointToPointAttractor(u,point)); }
      
    DivFunction* GetPointToLineRepellor(const double2& u, const Line3D& line) const
      { return new DivFunction(GetPointToLineAttractor(u,line)); }
      
    DivFunction* GetPointToPlaneRepellor(const double2& u, const Plane3D& plane) const
      { return new DivFunction(GetPointToPlaneAttractor(u,plane)); }
      
    DivFunction* GetNormalToVectorRepellor(const double2& u, const double3& normal) const
      { return new DivFunction(GetNormalToVectorAttractor(u,normal)); }
      
    DivFunction* GetRegionToPointRepellor(const Polygon2D& gamma, const double3& point) const
      { return new DivFunction(GetRegionToPointAttractor(gamma,point)); }
      
    DivFunction* GetRegionToLineRepellor(const Polygon2D& gamma, const Line3D& line) const
      { return new DivFunction(GetRegionToLineAttractor(gamma,line)); }
      
    DivFunction* GetRegionToPlaneRepellor(const Polygon2D& gamma, const Plane3D& plane) const
      { return new DivFunction(GetRegionToPlaneAttractor(gamma,plane)); }
      
    DivFunction* GetNormalRegionToVectorRepellor(const Polygon2D& gamma, const double3& normal) const
      { return new DivFunction(GetNormalRegionToVectorAttractor(gamma,normal)); }
      
    LinearEquation* GetPointToPointConstraint(const double2& u, const double3& point) const;
    // returns the equation that represents the constraint
    //
    //                F(u,v) = point
    
    LinearEquation* GetPointToLineConstraint(const double2& u, const Line3D& line) const;
    // returns the equation that represents the constraint
    //
    //                F(u,v) on line
    
    LinearEquation* GetPointToPlaneConstraint(const double2& u, const Plane3D& plane) const;
    // returns the equation that represents the constraint
    //
    //                F(u,v) in plane

    LinearEquation* GetNormalToVectorConstraint(const double2& x, const double3& normal) const;
    // returns the equation that represents the constraint 
    // 
    //                N(u,v) parallel to normal
    
    LinearEquation* GetHorizontalCurveConstraint(int2 u, double v0, const BSpline& curve) const;
    // returns the equation that represents the constraint
    //
    //                F(u,v0) = curve(u+??)          u.x() <= u <= u.y()

    LinearEquation* GetVerticalCurveConstraint(double u0, int2 v, const BSpline& curve) const;
    // returns the equation that represents the constraint
    //
    //                F(u0,v) = curve(v+??)          v.x() <= v <= v.y()

    void GetBoundaryCurves(BSpline& umincurve, BSpline& umaxcurve, BSpline& vmincurve, BSpline& vmaxcurve) const;
    // returns the curves that represent the boundary of the surface   

    List<LinearEquation> GetBoundaryConstraints() const;
    // returns the equations that represent the boundary of the surface
    //
    //                F(umin,v) = ...
    //                F(umax,v) = ...
    //                F(u,vmin) = ...
    //                F(u,vmax) = ...
    
    double ComputeSimpleStretchEnergy() const;
    // computes the following integral
    //      -
    //     |   <Fu|Fu> + <Fv|Fv> du dv
    //    -

    double ComputeSimpleStretchEnergyNumeric(int degree=3) const;
    // computes the following integral numerically
    //      -
    //     |   <Fu|Fu> + <Fv|Fv> du dv
    //    -

    double ComputeDataStretchEnergy(const VariationalTensorBSpline& G, int degree=3) const;
    // computes the following integral
    //           -         -1
    //  f(P) =  |  trace( I   I  ) sqrt(g ) du dv
    //         -           G   F         G

    double ComputeStretchEnergy(int degree=3) const;
    // computes the following integral
    //      -
    //     |   1 dA
    //    -
    
    double ComputeSimpleBendEnergy() const;
    // computes the following integral
    //      -
    //     |   <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
    //    -

    double ComputeSimpleBendEnergyNumeric(int degree=3) const;
    // computes the following integral numerically
    //      - 
    //     |   <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
    //    -

    double ComputeDataBendEnergy(const VariationalTensorBSpline& G, int degree=3) const;
    // computes the following integral
    //           -  3           2     2
    //  f(P) =  |  sum trace( (D  F_i)  ) sqrt(g ) du dv
    //         -   i=1          G               G

    double ComputeBendEnergy(int degree=3) const;
    // computes the following integral
    //      -
    //     |   (kappa1)^2 + (kappa2)^2  dA
    //    -

    double ComputeRegionToLineAttractor(const Polygon2D& gamma, const Line3D& line) const;

    void TestSimpleStretchEnergy(double weight) const;
    void TestDataStretchEnergy(double weight, const VariationalTensorBSpline& G) const;
    void TestSimpleBendEnergy(double weight) const;
    void TestDataBendEnergy(double weight, const VariationalTensorBSpline& G) const;
    void TestPointToPointAttractor(double weight, const double2& u, const double3& P) const;
    void TestPointToLineAttractor(double weight, const double2& u, const Line3D& line) const;
    void TestPointToPlaneAttractor(double weight, const double2& u, const Plane3D& plane) const;
    void TestNormalToVectorConstraint(const double2& u, const double3& normal) const;
    void TestRegionToPointAttractor(const Polygon2D& gamma, const double3& P) const;
    void TestRegionToLineAttractor(const Polygon2D& gamma, const Line3D& line) const;
    void TestRegionToPlaneAttractor(const Polygon2D& gamma, const Plane3D& plane) const;
    void TestPointToPointConstraint(const double2& u, const double3& P) const;
    void TestPointToLineConstraint(const double2& u, const Line3D& line) const;
    void TestPointToPlaneConstraint(const double2& u, const Plane3D& plane) const;
    void TestHorizontalCurveConstraint(int2 u, double v0, const BSpline& curve) const;
    void TestVerticalCurveConstraint(double u0, int2 v, const BSpline& curve) const;
    void TestBoundaryConstraints() const;
};

#endif

/*
    void GetSimpleMeanIntegral(QuadraticFunction* f);
    // computes a quadratic function f such that
    //
    //           -
    //  f(P) =  |  <Fuu+Fvv|Fuu+Fvv> du dv
    //         -

    void GetDataMeanIntegral(VariationalTensorBSpline& G, QuadraticFunction* f);
    // computes a quadratic function f such that
    //
    //           -  3           2       2
    //  f(P) =  |  sum trace( (D  F_i) ) sqrt(g_G) du dv
    //         -   i=1          G

    void GetSimpleGaussIntegral(DoubleMatrix& A);
    // computes a quadratic function f such that
    //
    //           -
    //  f(P) =  |  <Fuu|Fvv> - <Fuv|Fuv> du dv
    //         -

    void GetDataGaussIntegral(VariationalTensorBSpline& G, QuadraticFunction* f);
    // computes a quadratic function f such that
    //
    //           -  3         2
    //  f(P) =  |  sum det( (D  F_i) ) sqrt(g_G) du dv
    //         -   i=1        G

    double ComputeMeanIntegral();
    // computes the following integral
    //      -
    //     |   (kappa1 + kappa2)^2  dA
    //    -

    double ComputeGaussIntegral();
    // computes the following integral
    //      -
    //     |   kappa1 kappa2  dA
    //    -

    double ComputeSimpleMeanIntegral();
    // computes the following integral
    //      -
    //     |   <Fuu+Fvv|Fuu+Fvv> du dv
    //    -

    double ComputeSimpleGaussIntegral();
    // computes the following integral
    //      -
    //     |   <Fuu|Fvv> - <Fuv|Fuv> du dv
    //    -

    double ComputeSimpleMeanIntegralNumeric();
    // computes the following integral numerically
    //      - 
    //     |   <Fuu+Fvv|Fuu+Fvv> du dv
    //    -

    double ComputeSimpleGaussIntegralNumeric();
    // computes the following integral numerically
    //      - 
    //     |   <Fuu|Fvv> - <Fuv|Fuv> du dv
    //    -
*/

