// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <dvector.h>
#include "tbspline.h"

//-----------------------------------------------------------------------//
//                          ComputeCylinder                              //
//-----------------------------------------------------------------------//
void ComputeCylinder(int NumRows, int NumCols, double h, double R, int Angle, std::string FileName)
// arranges the control points of a tensor product B-spline surface
// such that they lie on the cylinder:
//
//   f(u,v) = (R cos(u), R sin(u), h)
//
// The B-spline surface is an approximation of the cylinder on
// the domain { 0 <= u <= M_PI*Angle/180  &&  0 <= v <= h }
{
  std::ofstream to(FileName.c_str());
  if (!to) {
    std::cerr << "could not open file " << FileName << std::endl;
    return;
  }

  to << NumRows << " " << NumCols << std::endl;

  double du = (M_PI*Angle/180)/(NumRows-3);
  double dv = h/(NumCols-3);

  for (int i=0; i<NumRows; i++) {
    for (int j=0; j<NumCols; j++) {
      double u = (i-1)*du;
      double v = (j-1)*dv;
      to << double3(R*cos(u), R*sin(u), v) << " ";
    }
    to << std::endl;
  }
  to << std::endl;

  to.close();
}

//-----------------------------------------------------------------------//
//                          main                                         //
//-----------------------------------------------------------------------//
// usage: numrows, numcols, outputfile, [h, R, Angle]
int main(int argc, char* argv[])
{
  int NumRows, NumCols;
  double Height, Radius, Angle;
  char FileName[80];

  if (argc > 1) {
    NumRows = atoi(argv[1]);
  }
  else {
    std::cout << "NumRows: "; std::cin >> NumRows;
  }

  if (argc > 2) {
    NumCols = atoi(argv[2]);
  }
  else {
    std::cout << "NumCols: "; std::cin >> NumCols;
  }

  if (argc > 3) {
    strcpy(FileName, argv[3]);
  }
  else {
    std::cout << "output file: "; std::cin >> FileName;
  }

  if (argc > 5) {
    Height = atof(argv[4]);
    Radius = atof(argv[5]);
    Angle = atof(argv[6]);
  }
  else {
    Height = 10;
    Radius = 5;
    Angle = 180;
  }

  ComputeCylinder(NumRows, NumCols, Height, Radius, Angle, FileName);

  return 0;
}

