// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <iostream>
#include <sstream>
#include <fstream>
#include <bvalues.h>
#include <cstdlib>
#include <cmath>
#include <string>
#include <equation.h>
#include <dvector.h>
#include <dmatrix.h>
#include "vbspline.h"

//-----------------------------------------------------------------------//
//                          GetFileName                                  //
//-----------------------------------------------------------------------//
std::string GetFileName(std::string InFile, int i)
// creates the name InFile_i
{
  std::stringstream out;
  out << InFile << std::setw(3) << std::setfill('0') << i << ".dat" << std::ends;
  return out.str();
}

//-----------------------------------------------------------------------//
//                          WriteInfo                                    //
//-----------------------------------------------------------------------//

void WriteInfo(std::ostream& to, const BSpline& x)
{
  to << "  exact stretch energy   = " << x.ComputeStretchEnergy() << std::endl;
  to << "  simple stretch energy  = " << x.ComputeSimpleStretchEnergy() << std::endl;
  to << "  exact bend energy      = " << x.ComputeBendEnergy() << std::endl;
  to << "  simple bend energy     = " << x.ComputeSimpleBendEnergy() << std::endl;
}

//-----------------------------------------------------------------------//
//                          IterateCurveDirect                           //
//-----------------------------------------------------------------------//
void IterateCurveDirect(std::string BaseName, int NumIterations)
{
  std::string InFileName = BaseName + ".dat";
  std::string OutFileName;
  std::string ReportFileName = BaseName + ".rep";

  std::ofstream report(ReportFileName.c_str());
  if (!report)
    std::cerr << "Could not open file " << ReportFileName;

  // read the curve from a file
  VariationalBSpline x(InFileName.c_str());

  std::cout << "cps:" << x.NumControlPoints() << std::endl;

  LinearEquationSystem constraints;
  double tmin = x.MinPar();
  double tmax = x.MaxPar();
  double3 xmin = x.GetPoint(tmin);
  double3 xmax = x.GetPoint(tmax);
  double3 dmin = x.GetTangent(tmin);
  double3 dmax = x.GetTangent(tmax);
  constraints += x.GetPointToPointConstraint(tmin,xmin);
  constraints += x.GetPointToPointConstraint(tmax,xmax);
  constraints += x.GetTangentToVectorConstraint(tmin,dmin);
  constraints += x.GetTangentToVectorConstraint(tmax,dmax);

  DoubleMatrix D;
  DoubleVector E;
  constraints.GetMatrix(D);
  constraints.GetVector(E);
  DoubleVector lambda(constraints.NumRows(),0);

  DoubleMatrix F;
  DoubleVector G;
  IntVector index;
  constraints.GetEliminateForm(F,G,index);
  std::cout << "constraints computed!" << std::endl;
  DoubleVector X = x.GetControlVector();
 
  report << "------------------------------------------------------" << std::endl;
  report << "                Curve " << 0 << std::endl;
  report << "------------------------------------------------------" << std::endl;
  WriteInfo(report,x);
  OutFileName = GetFileName(BaseName, 0); x.SaveFile(OutFileName.c_str());
//  constraints.Check(X);

  for (int i=0; i<NumIterations; i++) {
    std::cout << "iteration " << i << std::endl;
    report << "------------------------------------------------------" << std::endl;
    report << "                Curve " << i+1 << std::endl;
    report << "------------------------------------------------------" << std::endl;

    BSpline y = x;

    QuadraticFunctionList f_total = x.GetDataBendEnergy(y);

    report << "voor: f(X) = " << f_total(X) << std::endl;
    f_total.GetConstrainedMinimumLagrange(D, E, X, lambda, BKP_method);
    report << "na  : f(X) = " << f_total(X) << std::endl;

    x.SetControlVector(X);
    WriteInfo(report,x);
    OutFileName = GetFileName(BaseName, i+1); x.SaveFile(OutFileName.c_str());
//    constraints.Check(X);

    f_total.DeleteAll();
  }
}

int main(int argc, char* argv[])
// usage: FileName, maxiter
{
  int maxiter;
  char BaseName[80];

  std::cout << "Iterating curve" << std::endl;

  if (argc > 1) {
    strcpy(BaseName, argv[1]);
  }
  else {
    std::cout << "input file: "; std::cin >> BaseName;
  }

  if (argc > 2) {
    maxiter = atoi(argv[2]);
  }
  else {
    std::cout << "number of iterations: "; std::cin >> maxiter;
  }

  if (maxiter <= 0) {
    std::cerr << "Number of iterations must be >= 0!" << std::endl;
    exit(0);
  }

  IterateCurveDirect(BaseName, maxiter);

  return 0;
}














/*
void IterateStretch(char* BaseName, int maxiter)
{
  int i;

  char FileName[80];
  strcpy(FileName, BaseName);
  strcat(FileName, ".dat");  
  VariationalBSpline x(FileName);
  
  double tmin = x.MinPar();
  double tmax = x.MaxPar();
  
  double3 xmin = x.GetPoint(tmin);
  double3 xmax = x.GetPoint(tmax);
  
  List<SparseConstraint> constraints;
  constraints.Insert(x.GetPointToPointConstraint(PointToPointConstraint(tmin,xmin)));
  constraints.Insert(x.GetPointToPointConstraint(PointToPointConstraint(tmax,xmax)));

  DoubleEquationSystem eqn;
  for (i=0; i<constraints.Count; i++)
    constraints[i].AddToEquationSystem(eqn);   
  DoubleMatrix F;
  DoubleVector G;
  IntVector index;
std::cout << "eliminating constraints..." << flush;
  eqn.GetSimpleForm(F, G, index);
std::cout << "done!" << std::endl;


  for (i=0; i<maxiter; i++) {
    BSpline y = x;
    BandMatrix* Astretch = x.GetStretchEnergy(1,y);
    SparseQuadraticFunction* fstretch = new SparseQuadraticFunction(Astretch, 0, 0);

    QuadraticFunction f(3*x.NumControlPoints());
    fstretch->AddToQuadraticFunction(f);

    DoubleVector X;
    X = x.GetControlVector();

    std::cout << "-------------- iteratie " << i << " -------------------" << std::endl;
    std::cout << "   x.stretch(x) = " << x.ComputeStretchEnergyRef(x) << std::endl;
    std::cout << "   x.stretch(y) = " << x.ComputeStretchEnergyRef(y) << std::endl;
    std::cout << "   x.stretch(x) = " << x.ComputeStretchEnergy() << std::endl;
    std::cout << "   XtAX         = " << (X, (*Astretch)*X) << std::endl;

    f.GetConstrainedMinimum(F, G, X);
    x.SetControlVector(X);

    std::cout << "-------------- na: ---------------------------------" << std::endl;
    std::cout << "   x.stretch(x) = " << x.ComputeStretchEnergyRef(x) << std::endl;
    std::cout << "   x.stretch(y) = " << x.ComputeStretchEnergyRef(y) << std::endl;
    std::cout << "   x.stretch(x) = " << x.ComputeStretchEnergy() << std::endl;
    std::cout << "   XtAX         = " << (X, (*Astretch)*X) << std::endl;
    std::cout << "----------------------------------------------------" << std::endl;

    char OutFileName[80];
    GetFileName(BaseName, i+1, OutFileName);
    x.SaveFile(OutFileName);
  }
}

void IterateBend(char* BaseName, int maxiter)
{
  int i;

  char FileName[80];
  strcpy(FileName, BaseName);
  strcat(FileName, ".dat");
  VariationalBSpline x(FileName);

  double tmin = x.MinPar();
  double tmax = x.MaxPar();

  double3 xmin = x.GetPoint(tmin);
  double3 xmax = x.GetPoint(tmax);

  double3 dmin = x.GetTangent(tmin);
  double3 dmax = x.GetTangent(tmax);

  List<SparseConstraint> constraints;
  constraints.Insert(x.GetPointToPointConstraint(PointToPointConstraint(tmin,xmin)));
  constraints.Insert(x.GetPointToPointConstraint(PointToPointConstraint(tmax,xmax)));
  Tuple<SparseConstraint*> S = x.GetTangentToVectorConstraint(TangentToVectorConstraint(tmin,dmin));
  constraints.Insert(S[0]);
  constraints.Insert(S[1]);
  S = x.GetTangentToVectorConstraint(TangentToVectorConstraint(tmax,dmax));
  constraints.Insert(S[0]);
  constraints.Insert(S[1]);

  DoubleEquationSystem eqn;
  for (i=0; i<constraints.Count; i++)
    constraints[i].AddToEquationSystem(eqn);
  DoubleMatrix F;
  DoubleVector G;
  IntVector index;
std::cout << "eliminating constraints..." << flush;
  eqn.GetSimpleForm(F, G, index);
std::cout << "done!" << std::endl;

  for (i=0; i<maxiter; i++) {
    BSpline y = x;
    BandMatrix* Abend = x.GetBendEnergy(1,y);
    SparseQuadraticFunction* fbend = new SparseQuadraticFunction(Abend, 0, 0);

    QuadraticFunction f(3*x.NumControlPoints());
    fbend->AddToQuadraticFunction(f);

    DoubleVector X;
    X = x.GetControlVector();

    std::cout << "-------------- iteratie " << i << " -------------------" << std::endl;
    std::cout << "   x.bend(x) = " << x.ComputeBendEnergyRef(x) << std::endl;
    std::cout << "   x.bend(y) = " << x.ComputeBendEnergyRef(y) << std::endl;
    std::cout << "   x.bend()  = " << x.ComputeBendEnergy() << std::endl;
    std::cout << "   XtAX      = " << (X, (*Abend)*X) << std::endl;
    std::cout << "   f(X)      = " << f(X) << std::endl;
    std::cout << "   fbend(X)  = " << (*fbend)(X) << std::endl;

    f.GetConstrainedMinimum(F, G, X);
    x.SetControlVector(X);

    std::cout << "-------------- na: ---------------------------------" << std::endl;
    std::cout << "   x.bend(x) = " << x.ComputeBendEnergyRef(x) << std::endl;
    std::cout << "   x.bend(y) = " << x.ComputeBendEnergyRef(y) << std::endl;
    std::cout << "   x.bend()  = " << x.ComputeBendEnergy() << std::endl;
    std::cout << "   XtAX      = " << (X, (*Abend)*X) << std::endl;
    std::cout << "   f(X)      = " << f(X) << std::endl;
    std::cout << "   fbend(X)  = " << (*fbend)(X) << std::endl;
    std::cout << "----------------------------------------------------" << std::endl;

    char OutFileName[80];
    GetFileName(BaseName, i+1, OutFileName);
    x.SaveFile(OutFileName);
  }
}

*/
/*
    DoubleMatrix A;
    f_total.GetMatrix(A);
    DoubleMatrix A0(A);
    A0.MakeSymmetric();
    std::cout << "|A-A0| = " << (A-A0).FrobeniusNorm() << std::endl;
 
{
  DoubleVector Z = X*(*f_total.GetA()) + (*f_total.GetA())*X + lambda*D;
  std::cout << "|Z| = " << Z.Length() << std::endl;

  DoubleVector X0(X);
  f_total.GetConstrainedMinimum(F, G, X0, LU_method);
  std::cout << "Test..." << std::endl;
  std::cout << "   |X0-X| = " << (X0-X).Length() << std::endl;
  std::cout << "   fX  = " << f_total(X) << std::endl;
  std::cout << "   fX0 = " << f_total(X0) << std::endl;

  DoubleMatrix A;
  DoubleVector B;
  double C;
  f_total.GetMatrix(A);
  f_total.GetVector(B);
  C = f_total.GetC();

  DoubleMatrix P = F.Transpose()*A*F;
  DoubleVector Q = (G*A + A*G + B)*F;
  double R = (G,A*G) + (B,G);
  DoubleVector Y = (P+P.Transpose()).SolveLU(-Q);
  std::ifstream from("quadfunc.mtx");
  from >> P >> Q >> R >> Y;
  DoubleVector X1 = F*Y+G;

  std::cout << "|DF|     = " << (D*F).FrobeniusNorm() << std::endl;
  std::cout << "|DG - E| = " << (D*G-E).Length() << std::endl;
  std::cout << "|D*X1-E| = " << (D*X1-E).Length() << std::endl;
  std::cout << "|D*X0-E| = " << (D*X0-E).Length() << std::endl;
  std::cout << "|D*X-E|  = " << (D*X-E).Length() << std::endl;
  std::cout << "fX1 = " << (X1,A*X1) + (B,X1) + C << std::endl;
  std::cout << "gY  = " << (Y,P*Y) + (Q,Y) + R << std::endl;
}
*/

