// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <cstdlib>
#include "vbspline.h"
#include <bvalues.h>
#include <integr.h>
#include <geotyp.h>

#include "constrai.h"
#include "potentia.h"

void TestPointToPointAttractor(VariationalBSpline& curve, const PointToPointAttractor& attr)
{
  double t = attr.t;
  double weight = attr.weight;
  double3 point = attr.point;

  SparseQuadraticFunction* f = curve.GetPointToPointAttractor(attr);

  double3 F = curve.GetPoint(t);
  DoubleVector X = curve.GetControlVector();

  std::cout << "Testing PointToPointAttractor!" << std::endl;
  std::cout << "d(F(t),line)^2 = " << ((F-point),(F-point)) << std::endl;
  std::cout << "f(X) = " << (*f)(X) << std::endl;
}

void TestPointToLineAttractor(VariationalBSpline& curve, const PointToLineAttractor& attr)
{
  double t = attr.t;
  double weight = attr.weight;
  Line3D line = attr.line;

  SparseQuadraticFunction* f = curve.GetPointToLineAttractor(attr);

  double3 F = curve.GetPoint(t);
  DoubleVector X = curve.GetControlVector();

  double tmin;
  double dist = line.Distance(F,tmin);

  std::cout << "Testing PointToLineAttractor!" << std::endl;
  std::cout << "d(F(t),line)^2 = " << dist*dist << std::endl;
  std::cout << "f(X)             = " << (*f)(X) << std::endl;
}

void TestPointToPlaneAttractor(VariationalBSpline& curve, const PointToPlaneAttractor& attr)
{
  double t = attr.t;
  double weight = attr.weight;
  Plane3D plane = attr.plane;

  SparseQuadraticFunction* f = curve.GetPointToPlaneAttractor(attr);

  double3 F = curve.GetPoint(t);
  DoubleVector X = curve.GetControlVector();

  double dist = plane.Distance(F);

  std::cout << "Testing Planeattractor!" << std::endl;
  std::cout << "d(F(t),plane)^2 = " << dist*dist << std::endl;
  std::cout << "f(X) = " << (*f)(X) << std::endl;
}

void TestStretchEnergy(VariationalBSpline& curve, const StretchEnergy& stretch)
{
  double weight = stretch.weight;

  List<SparseQuadraticFunction> f1 = curve.GetSimpleStretchEnergy(stretch);
  List<SparseQuadraticFunction> f2 = curve.GetNumSimpleStretchEnergy(stretch);

  DoubleVector X = curve.GetControlVector();

  std::cout << "Testing Stretchenergy!" << std::endl;
  for (int i=0; i<f1.Count; i++) {
    std::cout << "f1(X) = " << f1[i](X) << std::endl;
    std::cout << "f2(X) = " << f2[i](X) << std::endl;
  }
}

void TestBendEnergy(VariationalBSpline& curve, const BendEnergy& bend)
{
  double weight = bend.weight;

  List<SparseQuadraticFunction> f1 = curve.GetSimpleBendEnergy(bend);
  List<SparseQuadraticFunction> f2 = curve.GetNumSimpleBendEnergy(bend);

  DoubleVector X = curve.GetControlVector();

  std::cout << "Testing Bendenergy!" << std::endl;
  for (int i=0; i<f1.Count; i++) {
    std::cout << "f1(X) = " << f1[i](X) << std::endl;
    std::cout << "f2(X) = " << f2[i](X) << std::endl;
  }
}

int main(int argc, char *argv[])
{
  // define a curve
  VariationalBSpline curve("curve.dat");

  // define a point to point attractor
  double t = 0.3;
  Parameter param(t);
  double3 point(1,2,3);
  double weight = 1;
  PointToPointAttractor p2p(param, weight, point);
  TestPointToPointAttractor(curve, p2p);

  // define a point to line attractor
  double3 r(1,1,1);
  double3 s(1,0,4);
  Line3D line(r,s);
  PointToLineAttractor p2l(param, weight, line);
  TestPointToLineAttractor(curve, p2l);

  // define a point to plane attractor
  double3 normal(1,2,3);
  double d = -2;
  Plane3D plane(normal,d);
  PointToPlaneAttractor p2v(param, weight, plane);
  TestPointToPlaneAttractor(curve, p2v);

  // define a stretch energy
  double v = 0.2;
  double w = 0.9;
  Interval interval(v,w);
  StretchEnergy stretch(interval, weight);
  TestStretchEnergy(curve, stretch);

  // define a bend energy
  BendEnergy bend(interval, weight);
  TestBendEnergy(curve, bend);

  return 0;
}

