// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 07-09-1996
// -> NUBSplineDerivs wegwerken

#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>

#include "bspline.h"

#include <bvalues.h>
#include <newton.h>
#include <splinval.h>
#include <poly.h>
#include <integr.h>
#include <numtyp.h>
#include <numalg.h>

//-----------------------------------------------------------------------//
//                          BSpline                                      //
//-----------------------------------------------------------------------//

//--- BSpline ------------------------------------------------------//

BSpline::BSpline() : V(10, 10)
{
  for (int i = 0; i < NUMCTLPTS; i++) knots[i] = i;
}

//--- BSpline ------------------------------------------------------//

BSpline::BSpline(int n) : V(n, 10)
{
  for (int i = 0; i < NUMCTLPTS; i++) knots[i] = i;
}

//--- BSpline ------------------------------------------------------//

BSpline::BSpline(std::string FileName) : V(10, 10)
{
  for (int i = 0; i < NUMCTLPTS; i++) knots[i] = i;
  std::ifstream from(FileName.c_str());

  if (!from)
  {
    std::cerr << "BSpline::BSpline(): could not open file " << FileName << std::endl;
  }
  else
    from >> (*this);
}

//--- Index --------------------------------------------------------//

int BSpline::Index(double u) const
{
  int result;

  if (u < MinPar()) result = 2;
  else if (u >= MaxPar()) result = NumKnots() - 4;
  else result = int(floor(u));

  return result;
}

//--- GetControlVector -----------------------------------------//

DoubleVector BSpline::GetControlVector() const
{
  DoubleVector x(3 * NumControlPoints());

  for (int i = 0; i < NumControlPoints(); i++)
  {
    const double3& Point = ControlPoint(i);
    int row = 3 * i;
    x[row] = Point.x();
    x[row + 1] = Point.y();
    x[row + 2] = Point.z();
  }

  return x;
}

void BSpline::SetLine(int NumPoints, double Length)
{
  Flush();
  double du = Length / (NumPoints - 3);

  for (int i = 0; i < NumPoints; i++)
  {
    double u = (i - 1) * du;
    InsertControlPoint(new double3(u, 0, 0));
  }
}

void BSpline::SetCircle(int NumPoints, double R, int Angle)
{
  Flush();
  double du = (M_PI * Angle / 180) / (NumPoints - 3);

  for (int i = 0; i < NumPoints; i++)
  {
    double u = (i - 1) * du;
    InsertControlPoint(new double3(R * cos(u), R * sin(u), 0));
  }
}

//--- SetControlVector -----------------------------------------//

void BSpline::SetControlVector(DoubleVector& x)
{
  for (int i = 0; i < NumControlPoints(); i++)
  {
    int row = 3 * i;
    ControlPoint(i) = double3(x[row], x[row + 1], x[row + 2]);
  }
}

//--- GetPoint ----------------------------------------------//

double3 BSpline::GetPoint(double u) const
{
  const int degree = 3;
  int index = Index(u);

  double* coeffx = new double[degree + 1];
  double* coeffy = new double[degree + 1];
  double* coeffz = new double[degree + 1];

  for (int i = 0; i <= degree; i++)
  {
    coeffx[i] = V[index - degree + 1 + i].x();
    coeffy[i] = V[index - degree + 1 + i].y();
    coeffz[i] = V[index - degree + 1 + i].z();
  }

  double x = DeBoor(degree, coeffx, knots + index - degree + 1, u, index);
  double y = DeBoor(degree, coeffy, knots + index - degree + 1, u, index);
  double z = DeBoor(degree, coeffz, knots + index - degree + 1, u, index);

  delete coeffx;
  delete coeffy;
  delete coeffz;

  return double3(x, y, z);
}

//--- GetDeriv ----------------------------------------------//

double3 BSpline::GetDeriv(double u, int d) const
{
  double dvals[4];
  double3 Deriv(0, 0, 0);
  int indx = Index(u);
  int k = Segment(u);

  NUBSplineDerivs(u, 4, d, knots, indx, dvals);
  for (int i = 0; i < 4; i++)
    Deriv += dvals[3 - i] * ControlPoint(k + i);

  return Deriv;
}

double BSpline::Curvature(double u) const
{
  CurvSample sample = GetSample(u);
  return sample.Curvature();
}

double3 BSpline::LaplaceBeltrami(double u, const BSpline& curve)
{
  double3 x = GetPoint(u);
  double3 xt = GetDeriv(u, 1);
  double3 xtt = GetDeriv(u, 2);
  double3 c = curve.GetPoint(u);
  double3 ct = curve.GetDeriv(u, 1);
  double3 ctt = curve.GetDeriv(u, 2);

  double cc = (ct, ct);
  return xtt / cc - xt * ((ct, ctt) / (cc * cc));
}

CurvSample BSpline::GetSample(double u) const
{
  double3 x = GetPoint(u);
  double3 xt = GetDeriv(u, 1);
  double3 xtt = GetDeriv(u, 2);

  return CurvSample(x, xt, xtt);
}

double BSpline::ComputeBendEnergy(int degree) const
{
  double energy = 0;

  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0, 1), degree);
  for (int i = 0; i < NumSegments(); i++)
  {
    double SegmentBeginPar = 2 + i;
    for (int j = 0; j < intpoints.Count; j++)
    {
      double tp = SegmentBeginPar + intpoints[j].Point();
      double wp = intpoints[j].Weight();
      CurvSample sample = GetSample(tp);
      double kappa = sample.Curvature();
      double length = sample.Ft.Length();
      double fp = kappa * kappa * length;
      energy += wp * fp;
    }
  }

  return energy;
}

double BSpline::ComputeSimpleBendEnergy(int degree) const
{
  double energy = 0;

  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0, 1), degree);
  for (int i = 0; i < NumSegments(); i++)
  {
    double SegmentBeginPar = 2 + i;
    for (int j = 0; j < intpoints.Count; j++)
    {
      CurvSample sample = GetSample(SegmentBeginPar + intpoints[j].Point());
      energy += intpoints[j].Weight() * sample.Ftt.Length2();
    }
  }

  return energy;
}

double BSpline::ComputeStretchEnergy(int degree) const
{
  double energy = 0;

  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0, 1), degree);
  for (int i = 0; i < NumSegments(); i++)
  {
    double SegmentBeginPar = 2 + i;
    for (int j = 0; j < intpoints.Count; j++)
    {
      CurvSample sample = GetSample(SegmentBeginPar + intpoints[j].Point());
      energy += intpoints[j].Weight() * sample.Ft.Length();
    }
  }

  return energy;
}

double BSpline::ComputeSimpleStretchEnergy(int degree) const
{
  double energy = 0;

  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0, 1), degree);
  for (int i = 0; i < NumSegments(); i++)
  {
    double SegmentBeginPar = 2 + i;
    for (int j = 0; j < intpoints.Count; j++)
    {
      CurvSample sample = GetSample(SegmentBeginPar + intpoints[j].Point());
      energy += intpoints[j].Weight() * sample.Ft.Length2();
    }
  }

  return energy;
}

//--- BBox ---------------------------------------------------------//

BoundingBox3D BSpline::BBox(int i, int n) const
{
  BoundingBox3D b;
  for (int k = i; k < i + n; k++) b.Insert(V[k]);
  return b;
}

//--- XPol ---------------------------------------------------------//

Polynomial BSpline::XPol(int index) const
{
  int i;

  double* a = new double[4];
  for (i = 0; i < 4; i++) a[i] = 0;
  Polynomial result(a, 3);

  Polynomial* bpols = BSplinePolynomials(3, knots + index, index);
  for (i = 0; i <= 3; i++)
  {
    bpols[i] *= V[index + i].x();
    result += bpols[i];
  }
  delete[] bpols;

  return result;
}

//--- YPol -----------------------------------------------------//

Polynomial BSpline::YPol(int index) const
{
  int i;

  double* a = new double[4];
  for (i = 0; i < 4; i++) a[i] = 0;
  Polynomial result(a, 3);

  Polynomial* bpols = BSplinePolynomials(3, knots + index, index);
  for (i = 0; i <= 3; i++)
  {
    bpols[i] *= V[index + i].y();
    result += bpols[i];
  }
  delete[] bpols;

  return result;
}

//--- ZPol -----------------------------------------------------//

Polynomial BSpline::ZPol(int index) const
{
  int i;

  double* a = new double[4];
  for (i = 0; i < 4; i++) a[i] = 0;
  Polynomial result(a, 3);

  Polynomial* bpols = BSplinePolynomials(3, knots + index, index);
  for (i = 0; i <= 3; i++)
  {
    bpols[i] *= V[index + i].z();
    result += bpols[i];
  }
  delete[] bpols;

  return result;
}

double BSpline::FindParameter(const Line3D& line)
{
  double dmin = 1e100; // current minimal distance
  double kmin = 0; // segment number of point with minimal distance
  double tmin = 0; // parameter of point with minimal distance (with
  // respect to the segment kmin!)

  int maxk = NumControlPoints() - 1; // number of segments on control polygon
  for (int k = 0; k < maxk; k++)
  {
    double t;
    double d = LineSegment3D(V[k], V[k + 1]).Distance(line, t);
    //std::cout << "k,d,t = " << k << " " << d << " " << t << std::endl;
    if (d < dmin)
    {
      dmin = d;
      kmin = k;
      tmin = t;
    }
    //std::cout << "kmin,dmin,tmin = " << kmin << " " << dmin << " " << tmin << std::endl;
  }

  return (kmin + tmin) / maxk;
}

//--- FindParameter --------------------------------------------//

int BSpline::FindParameter(double u0, int index, const double3& p,
    double& result) const
// pre: knots[i+2] <= u0 < knots[i+3]
{
  const double TOL = 1.e-4;
  const int MAXITER = 10;

  BoundingBox3D b = BBox(index, 4);
  b.Grow(20);

  if (!b.Inside(p))
  {
    result = u0;
    return 0;
  }

  Polynomial sx = XPol(index);
  Polynomial sy = YPol(index);

  sx[0] -= p.x();
  sy[0] -= p.y();

  Polynomial f = sx * sx.Derivative() + sy * sy.Derivative();

  // search for a zero of f in the interval [ knots[index+2], knots[index+3] ]
  if (FindZero(f, u0, knots[index + 2], knots[index + 3], TOL, MAXITER, result))
    return 0;
  else
    return 1;
}

//--- FindParameter --------------------------------------------//

int BSpline::FindParameter(const double3& p, double& result) const
{
  double distance = 1.e100;
  int found = 0;

  for (int i = 0; i < NumSegments(); i++)
  {
    double u;
    double u0 = (knots[i + 2] + knots[i + 3]) / 2;
    int b = FindParameter(u0, i, p, u);
    double3 q = GetPoint(u);
    double distpq = Dist(p, q);
    if (distpq < distance)
    {
      result = u;
      distance = distpq;
      found = b;
    }
  }

  if (!found)
  {
    std::cerr << "Geen minimum gevonden in BSpline::FindParameter()" << std::endl;
    if (distance < 20) std::cerr << "Afstand is toch klein genoeg!" << std::endl;
  }

  return (distance < 20);
}

//--- ControlPolygonLength -----------------------------------------//

double BSpline::ControlPolygonLength() const
{
  double length = 0;
  for (int i = 1; i < V.Count; i++)
  {
    length += Dist(V[i - 1], V[i]);
  }

  return length;
}

//--- CalcSegments -------------------------------------------------//

void BSpline::CalcSegments(double v, double w, BSplineSegmentList& sl) const
// N.B. In this function it is assumed that the knot vector is 0, 1, 2, ...!!!
{
  int firstSegment, lastSegment;

  sl.DeleteAll();

  if (v < MinPar())
  {
    firstSegment = 0;
    v = MinPar();
  }
  else
    firstSegment = Segment(v);

  if (v > MaxPar())
  {
    lastSegment = NumSegments() - 1;
    w = MaxPar();
  }
  else
    lastSegment = Segment(w);

  if (firstSegment == lastSegment)
  {
    sl.Insert(new BSplineSegment(v, w, firstSegment));
  }
  else
  {
    sl.Insert(new BSplineSegment(v, Knot(firstSegment + 3), firstSegment));
    for (int k = firstSegment + 1; k < lastSegment; k++)
      sl.Insert(new BSplineSegment(Knot(k + 2), Knot(k + 3), k));
    sl.Insert(new BSplineSegment(Knot(lastSegment + 2), w, lastSegment));
  }
}

//--- Show ---------------------------------------------------------//

void BSpline::ShowControlPoints() const
{
  std::cout << "  control points:" << std::endl;
  for (int i = 0; i < V.Count; i++) std::cout << "    " << V[i] << std::endl;
}

//--- Show ---------------------------------------------------------//

void BSpline::Show() const
{
  int i;

  std::cout << "BSpline::Show()" << std::endl;
  std::cout << "  knots: ";
  for (i = 0; i < NumKnots(); i++) std::cout << knots[i] << " ";
  std::cout << std::endl;

  ShowControlPoints();
}

std::istream & operator>>(std::istream& from, BSpline& curve)
{
  char ch;
  char buf[100];
  double3 point;

  curve.Flush(); // empty the list of control points

  std::string line;
  while (std::getline(from, line))
  {
    if (line.size() == 0)
    {
      continue;
    }
    switch (line[0])
    {
      case 'c': // control point
        line.erase(line.begin());
        std::stringstream ss(line);
        ss >> point;
        std::cout << "read point " << point << std::endl;
        curve.InsertControlPoint(new double3(point));
        break;
    }
  }

  //  while (from >> ch) {
  //    switch (ch) {
  //std::cout << "read char " << ch << std::endl;
  //      case 'c': // control point
  //        from >> point;
  //std::cout << "read point " << point << std::endl;
  //        curve.InsertControlPoint(new double3(point));
  //        from.get(buf, 100, '\n');  // read until end of line
  //        break;
  //      default :
  //        from.get(buf, 100, '\n');  // read until end of line
  //        break;
  //    }

  return from;
}

std::ostream & operator<<(std::ostream &to, const BSpline& curve)
{
  int i;

  // save control points
  for (i = 0; i < curve.NumControlPoints(); i++)
    to << "c " << curve.ControlPoint(i) << std::endl;

  return to;
}

int BSpline::SaveFile(std::string FileName) const
{
  int i;

  std::ofstream to(FileName.c_str());

  if (!to)
  {
    std::cerr << "BSpline::SaveFile(): could not open file " << FileName << std::endl;
    return 0;
  }

  // save control points
  for (i = 0; i < NumControlPoints(); i++)
    to << "c " << ControlPoint(i) << std::endl;

  to.close();
  return 1;
}

/*
//--- LoadFile -----------------------------------------------------//
int BSpline::LoadFile(std::string FileName)
{
  char ch;
  char buf[100];
  double3 point;

  std::ifstream from(FileName);

  if (!from) {
    std::cerr << "BSpline::LoadFile(): could not open file " << FileName << std::endl;
    return 0;
  }

  Flush();

  while (from >> ch) {
    switch (ch) {
      case 'c': // control point
        from >> point;
        InsertControlPoint(new double3(point));
        from.get(buf, 100, '\n');  // read until end of line
        break;
      default :
        from.get(buf, 100, '\n');  // read until end of line
        break;
    }
  }

  from.close();
  return 1;
}

//--- SaveFile -----------------------------------------------------//
int BSpline::SaveFile(std::string FileName) const
{
  int i;

  std::ofstream to(FileName);

  if (!to) {
    std::cerr << "BSpline::SaveFile(): could not open file " << FileName << std::endl;
    return 0;
  }

  // save control points
  for (i=0; i<V.Count; i++)
    to << "c " << V[i] << std::endl;

  to.close();
  return 1;
}
 */

