// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-09-1996

#include <cstdio>      // exit()
#include <cstdlib>     // exit()
#include <fstream>
#include <iostream>
#include <bvalues.h>
#include <splinval.h>
#include <bipoly.h>

TBSplineEvaluator BSplineEvaluator;

// N[i,3] = i-th B-spline basisfunction N of degree 3 with
// knotvector u = [0 1 2 3 4 5]
//
//--------------------------------------------------//
//                                                  //
//  M[i]        =   N[i,3]                          //
//                                                  //
//--------------------------------------------------//
//                                                  //
//  Mx[i]       =   N'[i,3]                         //
//                                                  //
//--------------------------------------------------//
//                                                  //
//  Mxx[i]      =   N''[i,3]                        //
//                                                  //
//--------------------------------------------------//
//                                                  //
//  MM[i][j]    =   N[i,3] N[j,3]                   //
//                                                  //
//--------------------------------------------------//
//                                                  //
//  MxMx[i][j]  =   N'[i,3] N'[j,3]                 //
//                                                  //
//--------------------------------------------------//
//                                                  //
//              3                                   //
//              -                                   //
//  S[i][j] =  | N'[i,3](x) N'[j,3](x) dx           //
//            -                                     //
//            2                                     //
//                                                  //
//--------------------------------------------------//
//                                                  //
//              3                                   //
//              -                                   //
//  B[i][j] =  | N''[i,3](x) N''[j,3](x) dx         //
//            -                                     //
//            2                                     //
//                                                  //
//--------------------------------------------------//
//                                                  //
//            -                                     //
//  N[i] =   |   N[i,3](x) dx                       //
//          -                                       //
//                                                  //
//--------------------------------------------------//
//                                                  //
//            -                                     //
//  Nx[i] =  |   N'[i,3](x) dx                      //
//          -                                       //
//                                                  //
//--------------------------------------------------//
//                                                  //
//             -                                    //
//  xN[i] =   |   x N[i,3](x) dx                    //
//           -                                      //
//                                                  //
//--------------------------------------------------//
//                                                  //
//             -                                    //
//  xxN[i] =  |   x^2 N[i,3](x) dx                  //
//           -                                      //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                -                                 //
//  NN[i][j] =   |   N[i,3](x) N[j,3](x) dx         //
//              -                                   //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                -                                 //
//  xNN[i][j] =  |  x N[i,3](x) N[j,3](x) dx        //
//              -                                   //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                -                                 //
//  NNxx[i][j] = |   N[i,3](x) N''[j,3](x) dx       //
//              -                                   //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                  -                               //
//  NxNx[i][j] =   |   N'[i,3](x) N'[j,3](x) dx     //
//                -                                 //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                   -                              //
//  xNxNx[i][j] =   |  x N'[i,3](x) N'[j,3](x) dx   //
//                 -                                //
//--------------------------------------------------//
//                                                  //
//                    -                             //
//  NxxNxx[i][j] =   |   N''[i,3](x) N''[j,3](x) dx //
//                  -                               //
//                                                  //
//--------------------------------------------------//
//                                                  //
//                   -                              //
//  xNxxNxx[i][j] = |  x N''[i,3](x) N''[j,3](x) dx //
//                 -                                //
//--------------------------------------------------//

//-----------------------------------------------------------------------//
//                          MatIndex                                     //
//-----------------------------------------------------------------------//
inline static int MatIndex(int i, int j)
{
  return 4*i + j;
}

void TBSplineEvaluator::GetNValues(double t, double V[4])
{
  for (int i=0; i<4; i++)
    V[i] = M[i](t);
}

void TBSplineEvaluator::GetNxValues(double t, double V[4])
{
  for (int i=0; i<4; i++)
    V[i] = Mx[i](t);
}

void TBSplineEvaluator::GetNxxValues(double t, double V[4])
{
  for (int i=0; i<4; i++)
    V[i] = Mxx[i](t);
}

void TBSplineEvaluator::GetNNValues(double t, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = MM[i][j](t);
}

void TBSplineEvaluator::GetNxNxValues(double t, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = MxMx[i][j](t);
}

void TBSplineEvaluator::GetNxNxxValues(double t, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = Mx[i](t)*Mxx[j](t);
}

void TBSplineEvaluator::GetNxxNxxValues(double t, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = Mxx[i](t)*Mxx[j](t);
}

void TBSplineEvaluator::GetNNValues(double u, double v, double m[4][4])
// the matrix M is computed such that
//
//       M[i][j] = N[i,3](u) N[j,3](v)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = M[i](u)*M[j](v);
}

void TBSplineEvaluator::GetNNuValues(double u, double v, double m[4][4])
// the matrix M is computed such that
//
//       M[i][j] = N'[i,3](u) N[j,3](v)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = Mx[i](u)*M[j](v);
}

void TBSplineEvaluator::GetNNvValues(double u, double v, double m[4][4])
// the matrix M is computed such that
//
//       M[i][j] = N[i,3](u) N'[j,3](v)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = M[i](u)*Mx[j](v);
}

void TBSplineEvaluator::GetNNSamples(double u, double v, FuncSample2D m[4][4])
// returns samples of the functions
//
//      f(u,v) = N[i,3](u) N[j,3](v),   0 <= i,j <= 3
//
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      FuncSample2D& sample = m[i][j];
      sample.f   = M[i](u)   * M[j](v);
      sample.fu  = Mx[i](u)  * M[j](v);
      sample.fv  = M[i](u)   * Mx[j](v);
      sample.fuu = Mxx[i](u) * M[j](v);
      sample.fuv = Mx[i](u)  * Mx[j](v);
      sample.fvv = M[i](u)   * Mxx[j](v);
    }
}

void TBSplineEvaluator::GetSimpleStretchEnergyMatrix(DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             3  3
//             -  -
// a_ijkl  =  |  |  N_ij  N_kl   +  N_ij   N_kl   du dv
//           -  -       u     u         v      v
//           2  2
{
  A.SetSize(16,16);

  double nn[4][4];
  double nxnx[4][4];
  GetNNIntegrals(2, 3, nn);
  GetNxNxIntegrals(2, 3, nxnx);
      
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);
          A(row,col) = nxnx[i][k]*nn[j][l] + nn[i][k]*nxnx[j][l];
        }
  A.CopyUpperToLower();
}

void TBSplineEvaluator::GetSimpleBendEnergyMatrix(DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             3  3
//             -  -
// a_ijkl  =  |  |  N_ij  N_kl   +  2 N_ij   N_kl   +  N_ij   N_kl   du dv
//           -  -       uu    uu          uv     uv        vv     vv
//           2  2
//
{
  A.SetSize(16,16);

  double nn[4][4];
  double nxnx[4][4];
  double nxxnxx[4][4];
  GetNNIntegrals(2, 3, nn);
  GetNxNxIntegrals(2, 3, nxnx);
  GetNxxNxxIntegrals(2, 3, nxxnxx);
      
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);
          A(row,col) = nxxnxx[i][k]*nn[j][l] + 2*nxnx[i][k]*nxnx[j][l] + nn[i][k]*nxxnxx[j][l];
        }
  A.CopyUpperToLower();
}

void TBSplineEvaluator::GetNNNNValues(double u, double v, DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//           
// a_ijkl  =  N (u) N (v) N (u) N (v)         0 <= i,j,k,l <= 3
//             i     j     k     l
//
// where 2 <= u,v <= 3
{
  A.SetSize(16,16);

  double nn[4][4];
  GetNNValues(u, v, nn);
  
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);
          A(row,col) = nn[i][j]*nn[k][l];
        }
  A.CopyUpperToLower();
}

void TBSplineEvaluator::GetNuNuNvNvValues(double u, double v, DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//
// a_ijkl  =  N'(u) N (v) N'(u) N (v) + N (u) N'(v) N (u) N'(v)       0 <= i,j,k,l <= 3
//             i     j     k     l       i     j     k     l   
//
// where 2 <= u,v <= 3
{
  A.SetSize(16,16);

  double nnu[4][4];
  double nnv[4][4];
  GetNNuValues(u, v, nnu);
  GetNNvValues(u, v, nnv);
  
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);
          A(row,col) = nnu[i][j]*nnu[k][l] + nnv[i][j]*nnv[k][l];
        }
  A.CopyUpperToLower();
}
    
void TBSplineEvaluator::GetNNValues(double u, double v, DoubleVector& b)
// The vector B is computed with coefficients B(row) = b_ij such that
//           
// b_ij  =  N (u) N (v)         0 <= i,j <= 3
//           i     j   
//
// where 2 <= u,v <= 3
{
  b.SetSize(16);

  double nn[4][4];
  GetNNValues(u, v, nn);
  
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++) {
      int row = MatIndex(i,j);
      b(row) = nn[i][j];
    }
}

void TBSplineEvaluator::GetNNIntegrals(const VerticalTrapezoid2D& gamma, DoubleVector& b)
// The vector B is computed with coefficients B(row) = b_ij such that
//
//           -
// b_ij  =  |   N (u) N (v) du dv
//         -     i     j
//        gamma
{
  b.SetSize(16);

  // this probably isn't the most efficient way of computing these integrals...
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++) {
      int row = MatIndex(i,j);
      BiPolynomial f(M[i], M[j]);
      b(row) = f.Integrate(gamma);
    }
}

void TBSplineEvaluator::GetNNNNIntegrals(const VerticalTrapezoid2D& gamma, DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             -
// a_ijkl  =  |   N (u) N (v) N (u) N (v) du dv
//           -     i     j     k     l
//          gamma
{
  A.SetSize(16,16);

  BiPolynomial nn[4][4];

  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      nn[i][j] = BiPolynomial(M[i],M[j]);

  // this certainly isn't the most efficient way of computing these integrals...
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);         
          A(row,col) = (nn[i][j]*nn[k][l]).Integrate(gamma);
        }
  A.CopyUpperToLower();
}
    
void TBSplineEvaluator::GetNuNuNvNvIntegrals(const VerticalTrapezoid2D& gamma, DoubleMatrix& A)
// The matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             -
// a_ijkl  =  |   N'(u) N (v) N'(u) N (v) + N (u) N'(v) N (u) N'(v) du dv
//           -     i     j     k     l       i     j     k     l    
//          gamma
{
  A.SetSize(16,16);

  BiPolynomial nnu[4][4];
  BiPolynomial nnv[4][4];

  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++) {
      nnu[i][j] = BiPolynomial(Mx[i],M[j]);
      nnv[i][j] = BiPolynomial(M[i],Mx[j]);
    }

  // this certainly isn't the most efficient way of computing these integrals...
  for (int i=0; i<=3; i++)
    for (int j=0; j<=3; j++)
      for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
        for (int l=0; l<=3; l++) {
          int row = MatIndex(i,j);
          int col = MatIndex(k,l);         
          A(row,col) = (nnu[i][j]*nnu[k][l] + nnv[i][j]*nnv[k][l]).Integrate(gamma);
        }
  A.CopyUpperToLower();
}
    
void TBSplineEvaluator::GetNNIntegrals(double v, double w, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = NN[i][j](w) - NN[i][j](v);
}

void TBSplineEvaluator::GetNNxxIntegrals(double v, double w, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = NNxx[i][j](w) - NNxx[i][j](v);
}

void TBSplineEvaluator::GetNxNxIntegrals(double v, double w, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = NxNx[i][j](w) - NxNx[i][j](v);
}

void TBSplineEvaluator::GetNxxNxxIntegrals(double v, double w, double m[4][4])
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      m[i][j] = NxxNxx[i][j](w) - NxxNxx[i][j](v);
}

void TBSplineEvaluator::GetNIntegrals(double v, double w, double V[4])
{
  for (int i=0; i<4; i++)
    V[i] = N[i](w) - N[i](v);
}

void TBSplineEvaluator::GetNxIntegrals(double v, double w, double V[4])
{
  for (int i=0; i<4; i++)
    V[i] = Nx[i](w) - Nx[i](v);
}

TBSplineEvaluator::TBSplineEvaluator()
{
  ReadMatrices();
}

void TBSplineEvaluator::ReadMatrices()
{
  std::cout << "Reading B-spline files..." << std::flush;
  ReadPolynomialVector("M.mtx", M);
  ReadPolynomialVector("Mx.mtx", Mx);
  ReadPolynomialVector("Mxx.mtx", Mxx);
  ReadPolynomialMatrix("MM.mtx", MM);
  ReadPolynomialMatrix("MxMx.mtx", MxMx);
  ReadDoubleMatrix("S.mtx", S);
  ReadDoubleMatrix("B.mtx", B);
  ReadPolynomialVector("N.mtx", N);
  ReadPolynomialVector("Nx.mtx", Nx);
  ReadPolynomialVector("xN.mtx", xN);
  ReadPolynomialVector("xxN.mtx", xxN);
  ReadPolynomialMatrix("NN.mtx", NN);
  ReadPolynomialMatrix("xNN.mtx", xNN);
  ReadPolynomialMatrix("NNxx.mtx", NNxx);
  ReadPolynomialMatrix("NxNx.mtx", NxNx);
  ReadPolynomialMatrix("xNxNx.mtx", xNxNx);
  ReadPolynomialMatrix("NxxNxx.mtx", NxxNxx);
  ReadPolynomialMatrix("xNxxNxx.mtx", xNxxNxx);
  std::cout << " done!" << std::endl;
}

void TBSplineEvaluator::ReadDoubleMatrix(const char *FileName, double m[4][4])
{
  std::ifstream from(FileName);
  if (!from) {
    std::cerr << "could not open file " << FileName << std::endl << "Program halted" << std::endl;
    std::exit(1);
  }

  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      from >> m[i][j];
  from.close();
}

void TBSplineEvaluator::ReadPolynomialVector(const char *FileName, Polynomial V[4])
{
  std::ifstream from(FileName);
  if (!from) {
    std::cerr << "could not open file " << FileName << std::endl << "Program halted" << std::endl;
    std::exit(1);
  }

  for (int i=0; i<4; i++)
    from >> V[i];
  from.close();
}

void TBSplineEvaluator::ReadPolynomialMatrix(const char *FileName, Polynomial m[4][4])
{
  std::ifstream from(FileName);
  if (!from) {
    std::cerr << "could not open file " << FileName << std::endl << "Program halted" << std::endl;
    std::exit(1);
  }

  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      from >> m[i][j];
  from.close();
}

void TBSplineEvaluator::ShowDoubleMatrix(double m[4][4], const char* s)
{
  std::cout << s << std::endl;
  for (int i=0; i<4; i++) {
    for (int j=0; j<4; j++) std::cout << m[i][j] << " ";
    std::cout << std::endl;
  }
  std::cout << std::endl;
}

void TBSplineEvaluator::ShowMatrices()
{
  ShowDoubleMatrix(S, "matrix S:");
  ShowDoubleMatrix(B, "matrix B:");
  for (int i=0; i<4; i++) {
    std::cout << M[i] << std::endl;
    std::cout << Mx[i] << std::endl;
  }
}

