// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include "distfunc.h"

//-----------------------------------------------------------------------//
//                          PointDistFunc
//-----------------------------------------------------------------------//

PointDistFunc::PointDistFunc(const BSpline& curve, const double3& point): x(curve), P(point)
{
}

PointDistFunc::PointDistFunc(const PointDistFunc& f): x(f.x), P(f.P)
{
}

PointDistFunc::~PointDistFunc()
{
}

double PointDistFunc::operator()(double t) const
{
  double3 XminP = x(t) - P;
  return (XminP,XminP);
}
 
double PointDistFunc::Deriv(double t) const
{
  double3 X = x(t);
  double3 Xt = x.GetDeriv(t,1);
  return 2*(X-P, Xt);
}
 
double PointDistFunc::DDeriv(double t) const
{
  double3 X   = x(t);
  double3 Xt  = x.GetDeriv(t,1);
  double3 Xtt = x.GetDeriv(t,2);
  return 2*( (Xtt,X-P) + (Xt,Xt) );
}

int PointDistFunc::MinimizeNewton(double& t, double eps, int maxiter, double a, double b)
{
  return Function1D::MinimizeNewton(t, eps, maxiter, a, b);
}

std::ostream& operator<<(std::ostream& to, const PointDistFunc&) 
{
  return to << "this is a PointDistFunc" << std::endl;
}

std::istream& operator>>(std::istream& from, PointDistFunc&) 
{
  return from;
}

//-----------------------------------------------------------------------//
//                          LineDistFunc
//-----------------------------------------------------------------------//

LineDistFunc::LineDistFunc(const BSpline& curve, const Line3D& line): x(curve)
{
  line.GetDistanceMap(A,b);
}

LineDistFunc::LineDistFunc(const LineDistFunc& f): x(f.x), A(f.A), b(f.b)
{
}

LineDistFunc::~LineDistFunc()
{
}

double LineDistFunc::operator()(double t) const
{
  double3 d = (A*x(t) + b);
  return (d,d);
}
 
double LineDistFunc::Deriv(double t) const
{
  double3  d = (A*x(t) + b);
  double3 dd = A*x.GetDeriv(t,1);
  return 2*(d,dd);
}
 
double LineDistFunc::DDeriv(double t) const
{
  double3   d = (A*x(t) + b);
  double3  dd = A*x.GetDeriv(t,1);
  double3 ddd = A*x.GetDeriv(t,2);
  return 2*((d,ddd) + (dd,dd));
}

int LineDistFunc::MinimizeNewton(double& t, double eps, int maxiter, double a0, double b0)
{
  return Function1D::MinimizeNewton(t, eps, maxiter, a0, b0);
}

std::ostream& operator<<(std::ostream& to, const LineDistFunc&) 
{
  return to << "this is a LineDistFunc" << std::endl;
}

std::istream& operator>>(std::istream& from, LineDistFunc&) 
{
  return from;
}
