// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 29-3-95

#include <iostream>
#include <math.h>
#include "tequation.h"

//-----------------------------------------------------------------------//
//                          Equation                                     //
//-----------------------------------------------------------------------//
void Equation::Add(Equation& Equation, double t)
{
  for (int i=0; i<n; i++)
    if (!Equation[i].IsZero())
      a[i] += t*Equation[i].Value();
  b += t*Equation.Const();
}

void Equation::operator+= (Equation& Equation)
{
  for (int i=0; i<n; i++)
    a[i] += Equation[i].Value();
  b += Equation.Const();
}

void Equation::operator*= (double t)
{
  for (int i=0; i<n; i++)
    a[i] *= t;
  b *= t;
}

std::ostream& operator<<(std::ostream& s, Equation& Equation)
{
  for (int i=0; i<Equation.GetSize(); i++) {
    s << Equation[i] << " ";
  }
  std::cout << "  " << Equation.Const();
  return s;
}

//-----------------------------------------------------------------------//
//                          EquationSystem                               //
//-----------------------------------------------------------------------//
void EquationSystem::GetVector(DoubleVector& b)
{
  b.SetSize(3*NumRows());
  for (int i=0; i<NumRows(); i++) {
    b[3*i  ] = rows[i].Const().x();
    b[3*i+1] = rows[i].Const().y();
    b[3*i+2] = rows[i].Const().z();
  }
}

void EquationSystem::GetRightMatrix(DoubleMatrix& A)
{
  int numcols = NumCols() - NumRows();
  A.SetSize(NumRows(), numcols);
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<numcols; j++)
    A[i][j] = rows[i][j+NumRows()].Value();
}

void EquationSystem::GetMatrix(DoubleMatrix& A)
{
  A.SetSize(NumRows(), NumCols());
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      A[i][j] = rows[i][j].Value();
}

void EquationSystem::SwapColumns(int i, int j)
{
  if (i != j) {
    for (int k=0; k<NumRows(); k++)
      rows[k].Swap(i,j);
    sigma.Swap(i,j);
  }
}

void EquationSystem::GetPivotElement(int& row, int& col)
{
  TDouble max; // is initialized with 0

  for (int i=0; i<NumRows(); i++) {
    if (rows[i].IsDone())
      continue;
    for (int j=0; j<rows[i].GetSize(); j++) {
      TDouble& aij = rows[i][j];
      if (aij.IsAbsBigger(max)) {
        max = aij;
        row = i;
        col = j;
      }
    }
  }
}

void EquationSystem::ClearColumn(int row, int col)
{
  Equation& therow = rows[row];
  therow.SetDone();
  therow *= (1/therow[col].Value());

  for (int i=0; i<NumRows(); i++) {
    if (i == row || rows[i][col].IsZero())
      continue;
    rows[i].Add(therow, -rows[i][col].Value());
    rows[i][col].MakeZero();
  }
}

void EquationSystem::Clear()
{
  int i;

  if (NumRows() < 1) {
    std::cerr << "cannot clear an empty system!" << std::endl;
    return;
  }

  for (i=0; i<NumRows(); i++)
    rows[i].SetDone(0);
  sigma.SetSize(NumCols());
  sigma.MakeSequence();

  int row = -1;
  int col = -1;
  for (i=0; i<NumRows(); i++) {
    GetPivotElement(row, col);
    if (row == -1)
      std::cerr << "singular matrix in EquationSystem::Clear()!!!!!!" << std::endl;
    if (fabs(rows[row][col].Value()) < 1e-10)
      std::cerr << "nearly singular matrix in EquationSystem::Clear()!!!!" << std::endl;
    ClearColumn(row, col);
    SwapColumns(row, col);
  }
}

void EquationSystem::GetSimpleForm(DoubleMatrix& C, DoubleVector& D)
{
  int i,j;

  int m = NumRows();
  int n = NumCols();

  // first we transform the system into  [I(m) F] sigma(X) = G
  Clear();

  DoubleMatrix F;
  DoubleVector G;
  GetRightMatrix(F);
  GetVector(G);

  // compute the matrix T = [  -F    ]
  //                        [ I(n-m) ]
  DoubleMatrix T;

  T.SetSize(n,n-m);
  T.MakeZero();

  for (i=0; i<m; i++)
    for (j=0; j<n-m; j++)
      T[i][j] = -F[i][j];

  for (i=0; i<n-m; i++)
    T[m+i][i] = 1;

  // create a vector H = [ G ]
  //                     [ 0 ]
  DoubleVector H(3*n);
  H.MakeZero();
  for (i=0; i<m; i++) {
    H[3*i]   = G[3*i];
    H[3*i+1] = G[3*i+1];
    H[3*i+2] = G[3*i+2];
  }

  // compute the answer from T and H
  C.SetSize(n,n-m);
  for (i=0; i<n; i++)
    for (j=0; j<n-m; j++)
      C[sigma[i]][j] = T[i][j];

  D.SetSize(3*n);
  for (i=0; i<n; i++) {
    D[3*sigma[i]]   = H[3*i];
    D[3*sigma[i]+1] = H[3*i+1];
    D[3*sigma[i]+2] = H[3*i+2];
  }
}

std::ostream& operator<<(std::ostream& s, EquationSystem& System)
{
  s << std::endl;
  for (int i=0; i<System.NumRows(); i++) {
    if (System[i].IsDone())
      s << "d ";
    else
      s << "u ";
    s << System[i] << std::endl;
  }
  s << "sigma = " << System.sigma << std::endl;
  return s;
}

