// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 07-09-1996

#include <fstream>
#include <iostream>
#include <cmath>
#include <iomanip>

#include <bvalues.h>
#include <dvector.h>
#include <dmatrix.h>
#include <integr.h>

#include "vbspline.h"

//-----------------------------------------------------------------------//
//                          VariationalBSpline                           //
//-----------------------------------------------------------------------//

QuadraticFunction* VariationalBSpline::GetSimpleStretchEnergy() const
// computes a quadratic function f such that
//
//           -
//  f(P) =  |  <Ft|Ft> dt
//         -
{
  BandMatrix* A = new BandMatrix(3*NumControlPoints(), 3*NumControlPoints(), 9, 9);

  double M[4][4];
  BSplineEvaluator.GetNxNxIntegrals(2,3,M);

  for (int k=0; k<NumSegments(); k++) {
    for (int i=0; i<4; i++)
      for (int j=0; j<4; j++) {
        int row = 3*(i+k);
        int col = 3*(j+k);
        (*A)(row,col)     += M[i][j];
        (*A)(row+1,col+1) += M[i][j];
        (*A)(row+2,col+2) += M[i][j];
      }  
  }
  
  return new QuadraticFunction(A, 0, 0);
}

QuadraticFunctionList VariationalBSpline::GetDataStretchEnergy(const BSpline& refcurve, int degree) const
// computes a list of quadratic functions f such that
//
//           - 
//  f(P) =  |  xxx
//         -   
//
// using numerical integration (valid degrees are 3,5,7 and 9)
{
  BandMatrix* A = new BandMatrix(3*NumControlPoints(), 3*NumControlPoints(), 9, 9);

  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0,1), degree);

  for (int k=0; k<NumSegments(); k++) {
    double SegmentBeginPar = k+2;
    for (int p=0; p<intpoints.Count; p++) {
      double tp = SegmentBeginPar + intpoints[p].Point();
      double wp = intpoints[p].Weight();                    // (tp,wp) is integration point
      double ctlen = refcurve.GetTangent(tp).Length();

      double M[4][4];
      BSplineEvaluator.GetNxNxValues(tp-k, M);
      double factor = wp/ctlen;
      for (int ii=0; ii<4; ii++)
        for (int jj=0; jj<4; jj++)
          M[ii][jj] *= factor;

/////////////////////////////////////////////////////////////////////////////////
// std::cout << "TESTLEN" << std::endl;
// {
//   double sum=0;
//   for (int i=0; i<4; i++)
//     for (int j=0; j<4; j++)
//       sum += (M[i][j]/factor)*(ControlPoint(i+k),ControlPoint(j+k));
// 
//   double len1 = sum;
//   double len2 = GetTangent(tp).Length2();
//   std::cout << "len1 = " << len1 << std::endl;
//   std::cout << "len2 = " << len2 << std::endl;
//   std::cout << "ctlen = " << ctlen << std::endl;
// }
/////////////////////////////////////////////////////////////////////////////////

      for (int i=0; i<4; i++)
        for (int j=0; j<4; j++) {
          int row = 3*(i+k);
          int col = 3*(j+k);
          (*A)(row,col)     += M[i][j];
          (*A)(row+1,col+1) += M[i][j];
          (*A)(row+2,col+2) += M[i][j];
        }  
    }
  }

  return new QuadraticFunction(A, 0, 0);
}

QuadraticFunction* VariationalBSpline::GetSimpleBendEnergy() const
// computes a quadratic function f such that
//
//           -
//  f(P) =  |  <Ftt|Ftt> dt
//         -
{
  BandMatrix* A = new BandMatrix(3*NumControlPoints(), 3*NumControlPoints(), 9, 9);

  double M[4][4];
  BSplineEvaluator.GetNxNxIntegrals(2,3,M);

  for (int k=0; k<NumSegments(); k++) {
    for (int i=0; i<4; i++)
      for (int j=0; j<4; j++) {
        int row = 3*(i+k);
        int col = 3*(j+k);
        (*A)(row,col)     += M[i][j];
        (*A)(row+1,col+1) += M[i][j];
        (*A)(row+2,col+2) += M[i][j];
      }  
  }
  
  return new QuadraticFunction(A, 0, 0);
}

QuadraticFunctionList VariationalBSpline::GetDataBendEnergy(const BSpline& refcurve, int degree) const
// computes a list of quadratic functions f such that
//
//           - 
//  f(P) =  |  xxx
//         -   
//
// using numerical integration (valid degrees are 3,5,7 and 9)
{
  BandMatrix* A = new BandMatrix(3*NumControlPoints(), 3*NumControlPoints(), 9, 9);
  
  IntPoint1DList intpoints = IntPointEvaluator.GetIntervalIntPoints(double2(0,1), degree);
 
  for (int k=0; k<NumSegments(); k++) {
    double SegmentBeginPar = k+2;
    for (int p=0; p<intpoints.Count; p++) {
      double tp = SegmentBeginPar + intpoints[p].Point();
      double wp = intpoints[p].Weight();                    // (tp,wp) is integration point
      double3 ct  = refcurve.GetDeriv(tp,1);
      double3 ctt = refcurve.GetDeriv(tp,2);

      double ctct   = (ct,ct);
      double ctctt  = (ct,ctt);
      //double cttctt = (ctt,ctt);
      double ctlen   = ct.Length();

      double factor = wp*ctlen;

      double Nx[4];
      double Nxx[4];
      double HessN[4];
      BSplineEvaluator.GetNxValues(tp-k, Nx);
      BSplineEvaluator.GetNxxValues(tp-k, Nxx);
      for (int ii=0; ii<4; ii++)
        HessN[ii] = (Nxx[ii] - Nx[ii]*ctctt/ctct)/ctct;

/* 
      double fac1 = ctctt*ctctt/(ctct*ctct*ctct*ctct);
      double fac2 = -2*ctctt/(ctct*ctct*ctct);
      double fac3 = 1/(ctct*ctct);

      double M1[4][4];
      double M2[4][4];
      double M3[4][4];
      double M[4][4];
      BSplineEvaluator.GetNxNxValues(tp-k, M1);
      BSplineEvaluator.GetNxNxxValues(tp-k, M2);
      BSplineEvaluator.GetNxxNxxValues(tp-k, M3);
      for (int ii=0; ii<4; ii++)
        for (int jj=0; jj<4; jj++)
          M[ii][jj] = factor*(fac1*M1[ii][jj] + fac2*M2[ii][jj] + fac3*M3[ii][jj]);

/////////////////////////////////////////////////////////////////////////////////
// std::cout << "TESTCUR" << std::endl;
// {
//   double sum1=0;
//   double sum2=0;
//   double sum3=0;
//   for (int i=0; i<4; i++)
//     for (int j=0; j<4; j++) {
//       sum1 += (M1[i][j])*(ControlPoint(i+k),ControlPoint(j+k));
//       sum2 += (M2[i][j])*(ControlPoint(i+k),ControlPoint(j+k));
//       sum3 += (M3[i][j])*(ControlPoint(i+k),ControlPoint(j+k));
//     }
// 
//   double3 xt = GetDeriv(tp,1); 
//   double3 xtt = GetDeriv(tp,2); 
//   std::cout << "sum1 = " << sum1 << " " << (xt,xt) << std::endl;
//   std::cout << "sum2 = " << sum2 << " " << (xt,xtt) << std::endl;
//   std::cout << "sum3 = " << sum3 << " " << (xtt,xtt) << std::endl;
// 
//   double f1 = ctctt*ctctt/(ctct*ctct*ctct*ctct);
//   double f2 = -2*ctctt/(ctct*ctct*ctct);
//   double f3 = 1/(ctct*ctct);
// 
//   double kap = f1*sum1 + f2*sum2 + f3*sum3;
//   double kap2 = Curvature(tp); kap2 = kap2*kap2;
// 
//   std::cout << "curvature: " << kap << " " << kap2 << std::endl;
// }
/////////////////////////////////////////////////////////////////////////////////

{
  std::cout << "M123-test" << std::endl;
  double mm = 0;
  double m1 = 0;
  double m2 = 0;
  double m3 = 0;
  for (int yy=0; yy<4; yy++) {
    for (int zz=0; zz<4; zz++) {
      mm += fabs(M[yy][zz] - M[zz][yy]);
      m1 += fabs(M1[yy][zz] - M1[zz][yy]);
      m2 += fabs(M2[yy][zz] - M2[zz][yy]);
      m3 += fabs(M3[yy][zz] - M3[zz][yy]);
    }
  }
  std::cout << m1 << " " << m2 << " " << m3 << " " << mm << std::endl;
}
*/

      for (int i=0; i<4; i++)
        for (int j=0; j<4; j++) {
          int row = 3*(i+k);
          int col = 3*(j+k);
          double val = factor*HessN[i]*HessN[j];
          (*A)(row,col)     += val;
          (*A)(row+1,col+1) += val;
          (*A)(row+2,col+2) += val;
        }
    }
  }
  
  return new QuadraticFunction(A, 0, 0);
}

LinearEquation* VariationalBSpline::GetPointToPointConstraint(double t, const double3& point) const
// returns the equation that represents the constraint
//
//                F(t) = point
{
  int NC = NumControlPoints(); 
  int k = Segment(t);
  double N[4];
  BSplineEvaluator.GetNValues(t-k, N);

  IntVector index(4);
  for (int j=0; j<4; j++)
    index(j) = j+k;

  DoubleVector a(4);
  for (int i=0; i<4; i++)
    a(i) = N[i];
 
  return new PointConstraintEquation(NC, index, a, point);
}

LinearEquation* VariationalBSpline::GetPointToLineConstraint(double t, const Line3D& line) const
// returns the equation that represents the constraint
//
//                F(t) on line
{
  int NC = NumControlPoints(); 
  int k = Segment(t);
  double N[4];
  BSplineEvaluator.GetNValues(t-k, N);

  IntVector index(4);
  for (int j=0; j<4; j++)
    index(j) = j+k;

  DoubleVector a(4);
  for (int i=0; i<4; i++)
    a(i) = N[i];
 
  return new LineConstraintEquation(NC, index, a, line);
}

LinearEquation* VariationalBSpline::GetPointToPlaneConstraint(double t, const Plane3D& plane) const
// returns the equation that represents the constraint
//
//                F(t) in plane
{
  int NC = NumControlPoints(); 
  int k = Segment(t);
  double N[4];
  BSplineEvaluator.GetNValues(t-k, N);

  IntVector index(4);
  for (int j=0; j<4; j++)
    index(j) = j+k;

  DoubleVector a(4);
  for (int i=0; i<4; i++)
    a(i) = N[i];
 
  return new PlaneConstraintEquation(NC, index, a, plane);
}

LinearEquation* VariationalBSpline::GetTangentToVectorConstraint(double t, const double3& tangent) const
// returns the equation that represents the constraint
//
//                x'(t) parallel to tangent
{
  int NC = NumControlPoints(); 
  int k = Segment(t);
  double Nx[4];
  BSplineEvaluator.GetNxValues(t-k, Nx);

  IntVector index(4);
  for (int j=0; j<4; j++)
    index(j) = j+k;

  DoubleVector a(4);
  for (int i=0; i<4; i++)
    a(i) = Nx[i];
 
  return new TangentConstraintEquation(NC, index, a, tangent);
}

