// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <fstream>
#include <cmath>
#include <double3x3.h>
#include <pause.h>

#include "nieuw.h"
#include "vrml.h"

void TestVRML()
{
  std::cout << "Testing VRML" << std::endl;

  std::ofstream out("testvrml.wrl");

  double3 red(1,0,0);
  double3 green(0,1,0);
  double3 blue(0,0,1);
  double3 yellow(1,1,0);

  double3 x0(1,2,3);
  double3 x1(2,2,8);
  double3 point1(10, 0, 0);
  double3 point2(20, 0, 0);
  double radius = 1;

  VRMLWriteFileHeader(out);

  VRMLWriteBeginSeparator(out);
  VRMLWriteColor(out, red);
  VRMLWriteSphere(out, point1, radius, 0);
  VRMLWriteEndSeparator(out);

  VRMLWriteBeginSeparator(out);
  VRMLWriteColor(out, yellow);
  VRMLWriteSphere(out, point2, radius, 0);
  VRMLWriteEndSeparator(out);

  VRMLWriteCylinder(out, x0, x1, radius/2);

  out.close();

  std::cout << "output written in file 'testvrml.wrl'..." << std::endl;
}

void TestPolyline()
{
  std::cout << "Testing Polyline" << std::endl;

  double3 red(1,0,0);
  double3 green(0,1,0);
  double3 yellow(1,1,0);
  double radius = 0.03;

  Polyline3D poly1, poly2;
  for (double t=0; t<10; t += 0.2) {
    poly1.Insert(double3(0.1*t*cos(t), 0.1*t*sin(t), t));
    poly2.Insert(double3(t, cos(t), 0));
  }

  std::ofstream out("polyline.wrl");
  VRMLWriteFileHeader(out);
  VRMLWritePolylineAsCylinders(out, poly1, radius, green);
  VRMLWritePolylineAsCylinders(out, poly2, radius, yellow);

//  VRMLWriteColor(out, red);
//  for (int i=0; i<poly1.Count; i++) {
//    VRMLWriteSphere(out, poly1[i], radius);
//    VRMLWriteSphere(out, poly2[i], radius);
//  }

  out.close();

  std::cout << "output written in file 'polyline.wrl'..." << std::endl;
}

void TestCrossProduct()
{
  std::cout << "Testing CrossProduct" << std::endl;

  double3 a(1,2,3);
  double3 b(1,7,0);

  double3x3 A = GetCrossProductMatrix(a);
  std::cout << "A = " << A << std::endl;

  double3 c1 = (a^b);
  double3 c2 = A*b;
  std::cout << "c1 = " << c1 << std::endl;
  std::cout << "c2 = " << c2 << std::endl;

  std::cout << std::endl;
  Pause();
}

void TestAxialRotation()
{
  std::cout << "Testing AxialRotation" << std::endl;

  double3 n(1,2,3);
  double phi = 1.2;
  double3 x(1,7,0);

  double3x3 A = GetAxialRotationMatrix(n,phi);
  double3x3 B = GetAxialRotationMatrix(n,-phi);
  double3x3 At = A.Transpose();
  std::cout << "A = " << A << std::endl;

  std::cout << "Check if it is a rotation..." << std::endl;
  std::cout << "determinant = " << A.Det() << std::endl;
  std::cout << "A At = " << A*At << std::endl;
  std::cout << "At A = " << At*A << std::endl;
  std::cout << "At - B = " << At - B << std::endl;

  std::cout << "Check if it is the right rotation..." << std::endl;
  std::cout << "n = " << n << std::endl;
  std::cout << "An = " << A*n << std::endl;

  Pause();
}

void TestVectorRotation()
{
  std::cout << "Testing VectorRotation" << std::endl;

  double3 a(1,2,3);
  double3 b(1,7,0);
  double3 x(2,-2,2);

  double3x3 A = GetVectorRotationMatrix(a,b);
  double3x3 B = GetVectorRotationMatrix(b,a);
  double3x3 At = A.Transpose();
  std::cout << "A = " << A << std::endl;

  std::cout << "Check if it is a rotation..." << std::endl;
  std::cout << "determinant = " << A.Det() << std::endl;
  std::cout << "A At = " << A*At << std::endl;
  std::cout << "At A = " << At*A << std::endl;
  std::cout << "At - B = " << At - B << std::endl;

  std::cout << "Check if it is the right rotation..." << std::endl;
  std::cout << "a = " << a << std::endl;
  std::cout << "b = " << b << std::endl;
  std::cout << "A a.Unit() - b.Unit() = " << A*(a.Unit()) - b.Unit() << std::endl;

  Pause();
}

int main()
{
  TestVRML();
  TestPolyline();
  TestCrossProduct();
  TestAxialRotation();
  TestVectorRotation();

  return 0;
}
