// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 

#include <cmath>
#include <iostream>
#include <mystuff.h>
#include <dvector.h>
#include <dmatrix.h>

//-----------------------------------------------------------------------//
//                          DoubleMatrix                                 //
//-----------------------------------------------------------------------//

DoubleVector DoubleMatrix::Row(int row) const
{
  int n = NumCols();
  DoubleVector result(n);

  for (int i=0; i<n; i++)
    result(i) = elem(row,i);

  return result;
}

DoubleVector DoubleMatrix::Col(int col) const
{
  int m = NumRows();
  DoubleVector result(m);

  for (int i=0; i<m; i++)
    result(i) = elem(i,col);

  return result;
}

void DoubleMatrix::AddDouble3x3(int row, int col, const double3x3& B)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      elem(row+i, col+j) += B[i][j];
}

DoubleVector DoubleMatrix::EigenValues() const
{
  int m = NumRows();
  DoubleVector result(m);

  symmeig(mat(), MNULL, result.vec());

  return result;
}

DoubleMatrix DoubleMatrix::GetSubMatrix(int row1, int row2, int col1, int col2) const
{
  int m = row2 - row1 + 1;
  int n = col2 - col1 + 1;

  DoubleMatrix sub(m,n);
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      sub(i,j) = elem(row1+1+i, col1+1+j);

  return sub;
}

DoubleMatrix DoubleMatrix::EigenVectors() const
{
 int m= NumRows();
 int n= NumCols();
 DoubleMatrix result_re(m,n);
 DoubleMatrix result_im(m,n);

 DoubleMatrix T(*this),Q(m,m);

 schur(T.mat(),Q.mat());

 schur_vecs(T.mat(),Q.mat(),result_re.mat(),result_im.mat());

 return result_re;
}

DoubleMatrix DoubleMatrix::EigenVectors(DoubleMatrix &result_im) const
{
 int m= NumRows();
 int n= NumCols();
 DoubleMatrix result_re(m,n);

 DoubleMatrix T(*this),Q(m,m);

 schur(T.mat(),Q.mat());

 schur_vecs(T.mat(),Q.mat(),result_re.mat(),result_im.mat());

 return result_re;
}

void DoubleMatrix::GetColumnPermutation(DoubleMatrix& B, const IntVector& sigma) const
{
  B.SetSize(NumRows(), NumCols());
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      B(i,j) = elem(i,sigma[j]);
}

void DoubleMatrix::GetInverseColumnPermutation(DoubleMatrix& B, const IntVector& sigma) const
{
  B.SetSize(NumRows(), NumCols());
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      B(i,sigma[j]) = elem(i,j);
}

void DoubleMatrix::GetRowPermutation(DoubleMatrix& B, const IntVector& sigma) const
{
  B.SetSize(NumRows(), NumCols());
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      B(i,j) = elem(sigma[i],j);
}

void DoubleMatrix::GetInverseRowPermutation(DoubleMatrix& B, const IntVector& sigma) const
{
  B.SetSize(NumRows(), NumCols());
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      B(sigma[i],j) = elem(i,j);
}

void DoubleMatrix::RowHouse(const DoubleVector& v)
// Golub p. 197
{
  double beta = -2*(v,v);
  DoubleVector w = beta*(v*(*this));
  (*this) += DyadeMatrix(v,w);
}

void DoubleMatrix::ColHouse(const DoubleVector& v)
// Golub p. 197
{
  double beta = -2*(v,v);
  DoubleVector w = beta*((*this)*v);
  (*this) += DyadeMatrix(w,v);
}

void DoubleMatrix::RowRot(int col1, int col2, int startrow, double c, double s)
// see Golub p. 203
// the submatrix B = A(col1,col2, startrow:n) from A is overwritten
// with M^T B where M = M(c,s)
//
// Given A in R[m,n] the following algorithm overwrites A with
//   G(col1, col2, theta(c,s))^T * A
{
  double tau1, tau2;

  for (int j=startrow; j<=NumCols(); j++) {
    tau1 = Elem(col1,j);
    tau2 = Elem(col2,j);
    Elem(col1,j) = c*tau1 - s*tau2;
    Elem(col2,j) = s*tau1 + c*tau2;
  }
}

void DoubleMatrix::ColRot(int row1, int row2, int startcol, double c, double s)
// see Golub p. 203
// Given A in R[m,n] the following algorithm overwrites the submatrix
// B = A(startcol:m, row1,row2) from A with B M
// where M = M(c,s)
{
  double tau1, tau2;
 
  for (int i=startcol; i<=NumRows(); i++) {
    tau1 = Elem(i,row1);
    tau2 = Elem(i,row2);
    Elem(i,row1) = c*tau1 - s*tau2;
    Elem(i,row2) = s*tau1 + c*tau2;
  }
}

void DoubleMatrix::RowFastRot(int col, int row, double alpha, double beta, int type)
// derived from Golub p. 208
// Given A in R[m,n] the following algorithm overwrites the submatrix
// B = A(col:col+1, row:n) from A with M^T B
// where M = M(alpha, beta, type)
{
  int j;
  double tau1, tau2;
 
  if (type == 1) {
    for (j=row; j<=NumCols(); j++) {
      tau1 = Elem(col,j);
      tau2 = Elem(col+1,j);
      Elem(col,j) = beta*tau1 + tau2;
      Elem(col+1,j) = tau1 + alpha*tau2;
    }
  }
  else {
    for (j=row; j<=NumCols(); j++) {
      tau1 = Elem(col,j);
      tau2 = Elem(col+1,j);
      Elem(col,j) = tau1 + beta*tau2;
      Elem(col+1,j) = alpha*tau1 + tau2;
    }
  }
}

DoubleMatrix DoubleMatrix::Transpose() const
{
  int m = NumRows();
  int n = NumCols();

  DoubleMatrix B(n,m);
  for (int i=0; i<n; i++)
    for (int j=0; j<m; j++)
      B(i,j) = elem(j,i);

  return B;
}

double DoubleMatrix::FrobeniusNorm() const
{
  double norm = 0;
  int m = NumRows();
  int n = NumCols();

  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      norm += elem(i,j)*elem(i,j);

  return sqrt(norm);
}

void DoubleMatrix::Fill(double fill)
{
  int m = NumRows();
  int n = NumCols();

  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      elem(i,j) = fill;
}

void DoubleMatrix::FillDiag(double fill)
{
  int n = (NumRows()<NumCols())?NumRows():NumCols();
  for (int i=0; i<n; i++)
    elem(i,i) = fill;
}

void DoubleMatrix::FillRandom(double min, double max)
{
  for (int i=0; i<NumRows(); i++)
    for (int j=0; j<NumCols(); j++)
      elem(i,j) = Random(min,max);
}

void DoubleMatrix::MakeSymmetric()
{
  int m = NumRows();
  if (NumCols() == m) {
    for (int i=0; i<m; i++)
      for (int j=i+1; j<m; j++) {
        double t = (elem(i,j) + elem(j,i))/2;
        elem(i,j) = elem(j,i) = t;
      }
  }
  else {
    std::cerr << "dimension error in DoubleMatrix::MakeSymmetric()" << std::endl;
  }
}

void DoubleMatrix::CopyUpperToLower()
{
  int m = NumRows();
  if (NumCols() == m) {
    for (int i=1; i<m; i++)
      for (int j=0; j<i; j++)
        elem(i,j) = elem(j,i);
  }
  else {
    std::cerr << "dimension error in DoubleMatrix::CopyUpperToLower()" << std::endl;
  }
}

void DoubleMatrix::SwapCols(int col1, int col2)
{
  if (col1 != col2) {
    for (int i=0; i<NumRows(); i++) {
      double temp  = elem(i,col1);
      elem(i,col1) = elem(i,col2);
      elem(i,col2) = temp;
    }
  }
}

void DoubleMatrix::SwapRows(int row1, int row2)
{
  if (row1 != row2) {
     for (int i=0; i<NumCols(); i++) {
       double temp  = elem(row1,i);
       elem(row1,i) = elem(row2,i);
       elem(row2,i) = temp;
     }
  }
}

void DoubleMatrix::SetColumn(int col, const DoubleVector& a)
// overwrites the i-th column with the vector a
{
  for (int i=0; i<NumRows(); i++) elem(i,col) = a(i);
}

void DoubleMatrix::SetRow(int row, const DoubleVector& a)
// overwrites the i-th row with the vector a
{
  for (int i=0; i<NumCols(); i++) elem(row,i) = a(i);
}

void DoubleMatrix::MultColumn(double t, int col)
// multiplies the i-th column of the matrix by t
{
  for (int i=0; i<NumRows(); i++) elem(i,col) *= t;
}

void DoubleMatrix::MultRow(double t, int row)
// multiplies the Row(i) of the matrix by t
{
  for (int i=0; i<NumCols(); i++) elem(row,i) *= t;
}

void DoubleMatrix::MultAddColumn(double t, int col1, int col2)
// adds t*Col(col1) to Col(col2)
{
  for (int i=0; i<NumRows(); i++) elem(i,col2) += (t*elem(i,col1));
}

void DoubleMatrix::MultAddRow(double t, int row1, int row2)
// adds t*Row(row1) to Row(row2)
{
  for (int i=0; i<NumCols(); i++) elem(row2,i) += (t*elem(row1,i));
}

DoubleMatrix operator-(const DoubleMatrix& A)
{
  int m = A.NumRows();
  int n = A.NumCols();
  DoubleMatrix B(m,n);
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      B(i,j) = -A(i,j);
  return B;
}

DoubleMatrix operator+(const DoubleMatrix& A, const DoubleMatrix& B)
{
  int m = A.NumRows();
  int n = A.NumCols();

  DoubleMatrix C(m,n);
  m_add(A.mat(), B.mat(), C.mat());

  return C;
}

DoubleMatrix operator-(const DoubleMatrix& A, const DoubleMatrix& B)
{
  int m = A.NumRows();
  int n = A.NumCols();

  DoubleMatrix C(m,n);
  m_sub(A.mat(), B.mat(), C.mat());

  return C;
}

DoubleMatrix operator*(double t, const DoubleMatrix& A)
{
  int m = A.NumRows();
  int n = A.NumCols();

  DoubleMatrix B(m,n);
  sm_mlt(t, A.mat(), B.mat());

  return B;
}

DoubleMatrix operator*(const DoubleMatrix& A, double t)
{
  return t*A;
}

DoubleMatrix operator/(const DoubleMatrix& A, double t)
{
  int m = A.NumRows();
  int n = A.NumCols();

  DoubleMatrix B(m,n);
  sm_mlt(1/t, A.mat(), B.mat());

  return B;
}

DoubleMatrix operator*(const DoubleMatrix& A, const DoubleMatrix& B)
{
  int m = A.NumRows();
  int n = B.NumCols();

  DoubleMatrix C(m,n);
  m_mlt(A.mat(), B.mat(), C.mat());

  return C;
}

DoubleVector operator*(const DoubleMatrix& A, const DoubleVector& X)
{
  int n = A.NumCols();

  DoubleVector Y(n);
  mv_mlt(A.mat(), X.vec(), Y.vec());

  return Y;
}

DoubleVector DoubleMatrix::LeftMultiplyXYZ(const DoubleVector& X) const
{
  if (3*NumRows() != X.NumRows())
         std::cerr << "dimension error in DoubleMatrix::LeftMultiplyXYZ()" << std::endl;

  int m = NumRows();
  int n = NumCols();

  DoubleVector Y(3*n);
  Y.MakeZero();

  for (int j=0; j<n; j++)
    for (int i=0; i<m; i++) {
      int xrow = 3*i;
      int yrow = 3*j;
      Y(yrow)   += X(xrow)  *elem(i,j);
      Y(yrow+1) += X(xrow+1)*elem(i,j);
      Y(yrow+2) += X(xrow+2)*elem(i,j);
    }

  return Y;
}

DoubleVector DoubleMatrix::RightMultiplyXYZ(const DoubleVector& X) const
{
  if (3*NumCols() != X.NumRows())
         std::cerr << "dimension error in DoubleMatrix::RightMultiplyXYZ()" << std::endl;

  int m = NumRows();
  int n = NumCols();

  DoubleVector Y(3*m);
  Y.MakeZero();

  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++) {
      int xrow = 3*j;
      int yrow = 3*i;
      Y(yrow)   += X(xrow)  *elem(i,j);
      Y(yrow+1) += X(xrow+1)*elem(i,j);
      Y(yrow+2) += X(xrow+2)*elem(i,j);
    }

  return Y;
}

DoubleVector operator*(const DoubleVector& X, const DoubleMatrix& A)
{
  int n = A.NumCols();

  DoubleVector Y(n);
  vm_mlt(A.mat(), X.vec(), Y.vec());

  return Y;
}

std::ostream& operator<<(std::ostream& s, const DoubleMatrix& A)
{
  s << std::endl;
  s << A.NumRows() << " " << A.NumCols() << std::endl;
  for (int i=0; i<A.NumRows(); i++) {
    int j;
    for (j=0; j<A.NumCols()-1; j++)
      s << A(i,j) << " ";
    s << A(i,j) << std::endl;
  }
  return s;
}

std::istream& operator>>(std::istream& s, DoubleMatrix& A)
{
  int m,n;

  s >> m;
  s >> n;

  A.SetSize(m,n);
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      s >> A(i,j);
  return s;
}

DoubleMatrix DiagonalMatrix(const DoubleVector& x)
{
  int m = x.NumRows();
  DoubleMatrix A(m,m);
  A.MakeZero();
  for (int i=0; i<m; i++)
    A(i,i) = x(i);

  return A;
}

DoubleMatrix DyadeMatrix(const DoubleVector& x, const DoubleVector& y)
{
  int m = x.NumRows();
  int n = y.NumRows();
  DoubleMatrix A(m,n);
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      A(i,j) = x(i)*y(j);

  return A;
}

void DoubleMatrix::DeleteColumns(const IntVector& cols)
{
// a very naive approach is followed: copy the remaining columns to
// a new MAT structure and then replace the current matrix with this
// one

  int m = NumRows();
  int n = NumCols() - cols.Size();
  MAT* Anew = m_get(m, n);
  
  // first create a vector containing the indices of the remaining columns
  IntVector newcols(n);
  {
    int col=0;
    int newcol=0;
    for (int k=0; k<cols.Size(); k++) {
      for (; col<cols[k]; col++)
        newcols[newcol++] = col;
      col++;
    }
    while (newcol<n)
      newcols[newcol++] = col++;
  }

  // copy the remaining columns to Anew
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      Anew->me[i][newcols[j]] = elem(i,j);
      
  // replace A with Anew
  M_FREE(A);
  A = Anew;
}

void DoubleMatrix::DeleteRows(const IntVector& rows)
{
// a very naive approach is followed: copy the remaining rows to
// a new MAT structure and then replace the current matrix with this
// one

  int m = NumRows() - rows.Size();
  int n = NumCols();
  MAT* Anew = m_get(m, n);
  
  // first create a vector containing the indices of the remaining rows
  IntVector newrows(n);
  {
    int row=0;
    int newrow=0;
    for (int k=0; k<rows.Size(); k++) {
      for (; row<rows[k]; row++)
        newrows[newrow++] = row;
      row++;
    }
    while (newrow<m)
      newrows[newrow++] = row++;
  }

  // copy the remaining rows to Anew
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      Anew->me[newrows[i]][j] = elem(i,j);
      
  // replace A with Anew
  M_FREE(A);
  A = Anew;
}

DoubleVector DoubleMatrix::SolveBKP(const DoubleVector& b)
{
  DoubleBKPFact BKP(*this);
  return BKP.solve(b);
}

DoubleVector DoubleMatrix::SolveCH(const DoubleVector& b)
{
  DoubleCHFact CH(*this);
  return CH.solve(b);
}

DoubleVector DoubleMatrix::SolveLDL(const DoubleVector& b)
{
  DoubleLDLFact LDL(*this);
  return LDL.solve(b);
}

DoubleVector DoubleMatrix::SolveLU(const DoubleVector& b)
{
  DoubleLUFact LU(*this);
  return LU.solve(b);
}

DoubleVector DoubleMatrix::SolveQR(const DoubleVector& b)
{
  DoubleQRFact QR(*this);
  return QR.solve(b);
}

//-----------------------------------------------------------------------//
//                          DoubleBKPFact                                //
//-----------------------------------------------------------------------//

DoubleBKPFact::DoubleBKPFact(const DoubleMatrix& A)
{
  BKP = m_copy(A.mat(), MNULL);
  pivot = px_get(A.NumRows());
  blocks = px_get(A.NumRows()); // ???
  BKPfactor(BKP, pivot, blocks);
}

DoubleBKPFact::~DoubleBKPFact()
{
  M_FREE(BKP);
  PX_FREE(pivot);
  PX_FREE(blocks);
}

DoubleVector DoubleBKPFact::solve(const DoubleVector& b)
{
  DoubleVector x(BKP->n);
  VEC* y = BKPsolve(BKP, pivot, blocks, b.vec(), VNULL);
  return DoubleVector(y);
}

//-----------------------------------------------------------------------//
//                          DoubleCHFact                                 //
//-----------------------------------------------------------------------//

DoubleCHFact::DoubleCHFact(const DoubleMatrix& A)
{
  CH = m_copy(A.mat(), MNULL);
  CHfactor(CH);
}

DoubleCHFact::~DoubleCHFact()
{
  M_FREE(CH);
}

DoubleVector DoubleCHFact::solve(const DoubleVector& b)
{
  DoubleVector x(CH->n);
  VEC* y = CHsolve(CH, b.vec(), VNULL);
  return DoubleVector(y);
}

//-----------------------------------------------------------------------//
//                          DoubleLDLFact                                //
//-----------------------------------------------------------------------//

DoubleLDLFact::DoubleLDLFact(const DoubleMatrix& A)
{
  LDL = m_copy(A.mat(), MNULL);
  LDLfactor(LDL);
}

DoubleLDLFact::~DoubleLDLFact()
{
  M_FREE(LDL);
}

DoubleVector DoubleLDLFact::solve(const DoubleVector& b)
{
  DoubleVector x(LDL->n);
  VEC* y = LDLsolve(LDL, b.vec(), VNULL);
  return DoubleVector(y);
}

//-----------------------------------------------------------------------//
//                          DoubleLUFact                                 //
//-----------------------------------------------------------------------//

DoubleLUFact::DoubleLUFact(const DoubleMatrix& A)
{
  LU = m_copy(A.mat(), MNULL);
  pivot = px_get(A.NumRows()); // ???
  LUfactor(LU, pivot);
}

DoubleLUFact::~DoubleLUFact()
{
  M_FREE(LU);
  PX_FREE(pivot);
}

DoubleVector DoubleLUFact::solve(const DoubleVector& b)
{
  DoubleVector x(LU->n);
  VEC* y = LUsolve(LU, pivot, b.vec(), VNULL);
  return DoubleVector(y);
}

//-----------------------------------------------------------------------//
//                          DoubleQRFact                                 //
//-----------------------------------------------------------------------//

DoubleQRFact::DoubleQRFact(const DoubleMatrix& A)
{
  QR = m_copy(A.mat(), MNULL);
  diag = v_get(A.NumRows()); // ???
  QRfactor(QR, diag);
}

DoubleQRFact::~DoubleQRFact()
{
  M_FREE(QR);
  V_FREE(diag);
}

DoubleVector DoubleQRFact::solve(const DoubleVector& b)
{
  DoubleVector x(QR->n);
  VEC* y = QRsolve(QR, diag, b.vec(), VNULL);
  return DoubleVector(y);
}

