// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#include <cmath>
#include <iostream>
#include <double2.h>
#include <double2x2.h>

//-----------------------------------------------------------------------//
//                          double2x2                                    //
//-----------------------------------------------------------------------//

double2x2::double2x2(const double2& row0, const double2& row1)
{
  for (int i=0; i<2; i++) {
    A[0][i] = row0[i];
    A[1][i] = row1[i];
  }
}               
    
double double2x2::Det() const
{
  return A[0][0]*A[1][1] - A[0][1]*A[1][0];
}

double double2x2::Trace() const
{
  return A[0][0] + A[1][1];
}

int double2x2::GetEigenvalues(double& l1, double& l2)
// The eigenvalues are the zeroes of the characteristic polynomial of the matrix.
// This polynomial p is defined as
//
//         p(t) = det(t*I - A),
//
// where I is the 2x2 unit matrix.
//
// Suppose that p(t) = t^2 + b*t + c.
{
  double b = - Trace();
  double c = Det();

  double D = b*b - 4*c;

  if (D < 0)
    return 0;

  double sqrtD = sqrt(D);

  l1 = (-b - sqrtD)/2;
  l2 = (-b + sqrtD)/2;

  return 1;
}

double2x2 double2x2::Inverse() const
{
  return double2x2(A[1][1], -A[0][1], -A[1][0], A[0][0])/Det();
}

double2x2 double2x2::Transpose() const
{
  return double2x2(Col(0), Col(1));
}

double double2x2::FrobeniusNorm() const
{
  double norm=0;

  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      norm += A[i][j]*A[i][j];
  
  return sqrt(norm);
}

void double2x2::Fill(double fill)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] = fill;
}

void double2x2::MakeSymmetric()
{
  for (int i=0; i<2; i++)
    for (int j=i+1; j<2; j++) {
      double t = (A[i][j] + A[j][i])/2;
	  A[i][j] = A[j][i] = t;
    }
}

void double2x2::MultCol(int col, double t)
{
  for (int i=0; i<2; i++) A[i][col] *= t;
}

void double2x2::MultRow(int row, double t)
{
  for (int i=0; i<2; i++) A[row][i] *= t;
}

void double2x2::operator=(const double2x2& B)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] = B.A[i][j];
}

void double2x2::operator+=(const double2x2& B)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] += B.A[i][j];
}

void double2x2::operator-=(const double2x2& B)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] -= B.A[i][j];
}

void double2x2::operator*=(double t)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] *= t;
}

void double2x2::operator/=(double t)
{
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      A[i][j] /= t;
}

double2x2 operator-(const double2x2& A)
{
  double2x2 B;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      B[i][j] = A[i][j];
  return B;
}

double2x2 operator+(const double2x2& A, const double2x2& B)
{
  double2x2 C;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      C[i][j] = A[i][j] + B[i][j];
  return C;
}

double2x2 operator-(const double2x2& A, const double2x2& B)
{
  double2x2 C;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      C[i][j] = A[i][j] - B[i][j];
  return C;
}

double2x2 operator*(double t, const double2x2& A)
{
  double2x2 C;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      C[i][j] = t*A[i][j];
  return C;
}

double2x2 operator/(const double2x2& A, double t)
{
  double2x2 C;
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      C[i][j] = A[i][j]/t;
  return C;
}

double2x2 operator*(const double2x2& A, const double2x2& B)
{
  double2x2 C;
  C.MakeZero();
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      for (int k=0; k<2; k++)
        C[i][j] += A[i][k]*B[k][j];
  return C;
}

std::ostream& operator<<(std::ostream& s, const double2x2& A)
{
  s << std::endl;
  for (int i=0; i<2; i++) {
    int j;
    for (j=0; j<1; j++)
      s << A[i][j] << ", ";
    s << A[i][j] << std::endl;
  }
  return s;
}

std::istream& operator>>(std::istream& s, double2x2& A)
{
  for (int i=0; i<2; i++) 
    for (int j=0; j<2; j++)
      s >> A[i][j];
  return s;
}

double2x2 DiagonalMatrix(const double2& x)
{
  double2x2 A;
  A.MakeZero();
  for (int i=0; i<2; i++)
    A[i][i] = x[i];

  return A;
}

double2x2 DyadeMatrix(const double2& x, const double2& y)
{
  double2x2 A;
  for (int i=0; i<2; i++) 
    for (int j=0; j<2; j++)
      A[i][j] = x[i]*y[j];

  return A;
}

