// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-18-1996

#include <cmath>
#include <iostream>
#include <mystuff.h>
#include <fstream>
#include <dvector.h>

//-----------------------------------------------------------------------//
//                          DoubleVector                                 //
//-----------------------------------------------------------------------//

DoubleVector::DoubleVector(const DoubleVector& X, const DoubleVector& Y)
// constructor (creates the concatenation vector X Y)
{
  int m = X.NumRows();
  int n = Y.NumRows();
  x = v_get(m+n);
 
  for (int i=0; i<m; i++)
    elem(i) = X(i);
    
  for (int i=0; i<n; i++)
    elem(i+m) = Y(i);
}
 
DoubleVector::DoubleVector(const DoubleVector& X, const DoubleVector& Y, const DoubleVector& Z)
{
  int n = X.NumRows();
  x = v_get(3*n);
 
  for (int i=0; i<n; i++) {
    int row = 3*i;
    elem(row) = X(i);
    elem(row+1) = Y(i);
    elem(row+2) = Z(i);
  }
}

double DoubleVector::Length2() const
{
  double length = 0;

  for (int i=0; i<size(); i++)
    length += elem(i)*elem(i);

  return length;
}

double DoubleVector::Min() const
// returns the minimum value of the components
{
  if (size() <= 0) {
    std::cerr << "empty vector in DoubleVector::Min()" << std::endl;
    return 0;
  }

  double min = elem(0);
  for (int i=1; i<size(); i++) {
    if (elem(i) < min)
      min = elem(i);
  }

  return min;
}

double DoubleVector::Max() const
// returns the maximum value of the components
{
  if (size() <= 0) {
    std::cerr << "empty vector in DoubleVector::Max()" << std::endl;
    return 0;
  }

  double max = elem(0);
  for (int i=1; i<size(); i++) {
    if (elem(i) > max)
      max = elem(i);
  }

  return max;
}

double DoubleVector::Sum() const
{
  double Result = 0;
  
  for (int i=0; i<size(); i++)
    Result += elem(i);
    
  return Result;
}

double DoubleVector::Median() const
{
  DoubleVector temp = (*this);
  temp.Sort();
  return temp(temp.Size()/2);
}

void DoubleVector::Scale(double min, double max)
// scales the components to the interval [min,max] using an affine transformation
{
  double a = Min();
  double b = Max();
  
  if (b-a < 1e-50) {
    std::cerr << "cannot scale in DoubleVector::Scale";
    return;
  }
  
  for (int i=0; i<size(); i++)
    elem(i) = min + (elem(i)-a)*(max-min)/(b-a);
}

void DoubleVector::Fill(double fill)
{
  for (int i=0; i<size(); i++)
    elem(i) = fill;
}

void DoubleVector::FillRandom(double min, double max)
{
  for (int i=0; i<size(); i++)
    elem(i) = Random(min,max);
}

DoubleVector DoubleVector::House()
// Golub p. 196
{
  float mu = Length();
  DoubleVector v(*this);
     
  if (mu != 0) {
    double beta;
    if (Elem(1) > 0)
      beta = Elem(1) + mu;
    else
      beta = Elem(1) - mu;
    for (int i=2; i<=size(); i++)
      v.Elem(i) /= beta;
   }

  v.Elem(1) = 1;
      
  return v;
}

DoubleVector DoubleVector::Permute(const IntVector& Sigma) const
{
  DoubleVector result(size());

  for (int i=0; i<size(); i++)
    result[i] = elem(Sigma[i]);

  return result;
}

DoubleVector DoubleVector::PermuteInverse(const IntVector& Sigma) const
{
  DoubleVector result(size());
   
  for (int i=0; i<size(); i++)
    result[Sigma[i]] = elem(i);
  
  return result;
}

void DoubleVector::Split(int k, DoubleVector& y, DoubleVector& z) const
{
  int i;

  y.SetSize(k);
  z.SetSize(size()-k);

  for (i=0; i<k; i++)
    y(i) = elem(i);

  for (i=0; i<size()-k; i++)
    z(i) = elem(k+i);
}

void DoubleVector::SplitXYZ(DoubleVector& X, DoubleVector& Y, DoubleVector& Z) const
{
  int n = NumRows()/3;
  X.SetSize(n);
  Y.SetSize(n);
  Z.SetSize(n);

  for (int i=0; i<n; i++) {
    int row = 3*i;
    X(i) = elem(row);
    Y(i) = elem(row+1);
    Z(i) = elem(row+2);
  }
}

DoubleVector DoubleVector::SubVector(int i0, int i1) const
{
  int size = (i1-i0+1);
  DoubleVector result(size);
  for (int i=0; i<size; i++)
    result[i] = elem(i0+i);

  return result;
}

DoubleVector DoubleVector::Concat(const DoubleVector& y) const
{
  int i;
  int m = size();
  int n = y.NumRows();

  DoubleVector result(m+n);
  for (i=0; i<m; i++)
    result[i] = elem(i);
  for (i=0; i<n; i++)
    result[m+i] = y[i];

  return result;
}

DoubleVector operator-(const DoubleVector& x)
{
  DoubleVector result(x.size());

  v_sub(result.vec(), x.vec(), result.vec());
  return result;
}

DoubleVector operator+(const DoubleVector& x, const DoubleVector& y)
{
  DoubleVector result(x.size());

  v_add(x.vec(), y.vec(), result.vec());
  return result;
}

DoubleVector operator-(const DoubleVector& x, const DoubleVector& y)
{
  DoubleVector result(x.size());

  v_sub(x.vec(), y.vec(), result.vec());
  return result;
}

DoubleVector operator*(double t, const DoubleVector& x)
{
  DoubleVector result(x.size());

  sv_mlt(t, x.vec(), result.vec());
  return result;
}

DoubleVector operator*(const DoubleVector& x, double t)
{
  return t*x;
}
 
DoubleVector operator/(const DoubleVector& x, double t)
{
  DoubleVector result(x.size());

  sv_mlt(1/t, x.vec(), result.vec());
  return result;
}

double operator,(const DoubleVector&x, const DoubleVector& y)
{
  return in_prod(x.vec(), y.vec());
}

double operator*(const DoubleVector&x, const DoubleVector& y)
{
  return in_prod(x.vec(), y.vec());
}

void DoubleVector::WriteToFile(std::string FileName) const
{
  std::ofstream to(FileName.c_str());

  bool file_exists = false;
  {
    if (std::ifstream(FileName.c_str()))
      file_exists = true;
  }

  if (file_exists) {
    std::cerr << "DoubleVector::WriteToFile: file " << FileName 
         << " already exists!" << std::endl;
    return;
  }

  to << size() << std::endl;
  for (int i=0; i<size(); i++)
    to << elem(i) << std::endl;
}

std::ostream& operator<<(std::ostream& s, const DoubleVector& x)
{
//  s << '(';
  s << x.size() << " ";
  if (x.NumRows() > 0) {
    int i;
    for (i=0; i<x.NumRows()-1; i++)
      s << x[i] << " ";
    s << x[i];
  }
//  s << ')';
  return s;
}

std::istream& operator>>(std::istream& s, DoubleVector& x)
{
  int m;

  s >> m;

  x.SetSize(m);
  for (int i=0; i<m; i++)
    s >> x[i];
  return s;
}

void DoubleVector::AddDouble3(int row, const double3& X)
{
  elem(row)   += X[0];
  elem(row+1) += X[1];
  elem(row+2) += X[2];
}

