// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sat 06-15-1996

#include <function.h>
#include <double3.h>
#include <cmath>

//-----------------------------------------------------------------------//
//                          Function                                     //
//-----------------------------------------------------------------------//

double Function::val(const DoubleVector& x) const
{
  std::cerr << "method Function::val() undefined!" << std::endl; 
  return 0;
}

DoubleVector Function::grad(const DoubleVector& x) const
{
  std::cerr << "method Function::grad() undefined!" << std::endl;
  int n = dimension();
  return DoubleVector(n);
}

DoubleMatrix Function::hess(const DoubleVector& x) const
{
  std::cerr << "method Function::hess() undefined!" << std::endl;
  int n = dimension();
  return DoubleMatrix(n,n);
}

void Function::operator*=(double t)
{
  std::cerr << "method Function::operator*=() undefined!" << std::endl;
}

int Function::MinimizeConjugateGradient(DoubleVector& x, int imax,
         double epsCG, int jmax, double epsNewton)
{
  int n = dimension();
  int i=0;
  int k=0;
  DoubleVector r = -grad(x);
  DoubleVector d = r;
  double delta_new = (r,r);
  double delta_0 = delta_new;

  std::cout << "-------------------------------------------------------" << std::endl;
  std::cout << "-        Function::MinimizeConjugateGradient()        -" << std::endl;
  std::cout << "-------------------------------------------------------" << std::endl;
  std::cout << "f(x) = " << val(x) << std::endl;

  while ( (i<imax) && (delta_new > epsCG*epsCG*delta_0) ) {
    std::cout << "CG iteration " << i << "; " << std::flush;
    int j=0;
    double delta_d = (d,d);
    double alpha;
    do {
      alpha = - (grad(x),d) / hessleftright(x,d,d);
      x += alpha*d;
      j++;
    } while ( (j<jmax) && (alpha*alpha*delta_d > epsNewton*epsNewton) );

    std::cout << "linesearch: " << j << " iterations" << std::endl;
    
    r = -grad(x);
    double delta_old = delta_new;
    delta_new = (r,r);
    double beta = delta_new/delta_old;
    d = r + beta*d;
    k++;
    if ( (k==n) || (r,d) <= 0 ) {
      d = r;
      k = 0;
    }
    i++;
    std::cout << "f(x) = " << val(x) << std::endl;
  }
  
  return (i>=imax);
}        

int Function::TestConjugateGradient(DoubleVector& x, int imax,
         double epsCG, int jmax, double epsNewton)
{
  int n = dimension();
  int i=0;
  int k=0;
  DoubleVector r = -grad(x);
  DoubleVector d = r;
  double delta_new = (r,r);
  double delta_0 = delta_new;

  std::cout << "-------------------------------------------------------" << std::endl;
  std::cout << "-   Testing Function::MinimizeConjugateGradient()     -" << std::endl;
  std::cout << "-------------------------------------------------------" << std::endl;

//std::cout << "x0     = " << x << std::endl;
  std::cout << "f(x0)  = " << val(x) << std::endl;
  std::cout << "delta0 = " << delta_0 << std::endl;
  
  while ( (i<imax) && (delta_new > epsCG*epsCG*delta_0) ) {
    std::cout << "i = " << std::endl;
//  std::cout << "direction d = " << d << std::endl;
    std::cout << "performing line search..." << std::endl;

    int j=0;
    double delta_d = (d,d);
    double alpha;
    do {
      alpha = - (grad(x),d) / (d,hess(x)*d);
      x += alpha*d;
      std::cout << "   j              = " << j << std::endl;
      std::cout << "   alpha          = " << alpha << std::endl;
//    std::cout << "   x + alpha*d    = " << x << std::endl;
      std::cout << "   f(x + alpha*d) = " << val(x) << std::endl;
      j++;
    } while ( (j<jmax) && (alpha*alpha*delta_d > epsNewton*epsNewton) );

    std::cout << "finished line search" << std::endl;
    if (j >= jmax)
      std::cout << "error: (j>jmax)!" << std::endl;
    if (alpha*alpha*delta_d > epsNewton*epsNewton)
      std::cout << "aad, ee = " << alpha*alpha*delta_d << epsNewton*epsNewton << std::endl;
    
    r = -grad(x);
    double delta_old = delta_new;
    delta_new = (r,r);
    std::cout << "delta = " << delta_new << std::endl;
    double beta = delta_new/delta_old;
    d = r + beta*d;
//  std::cout << "direction d = " << d << std::endl;
    k++;
    if ( (k==n) || (r,d) <= 0 ) {
      std::cout << "error: (k==n) || (r,d) <= 0 " << std::endl;
      std::cout << "  k,n, (r,d) = " << k << " " << n << " " << (r,d) << std::endl;     
      d = r;
      k = 0;
    }
    i++;
  }
  
  return (i>=imax);
}        

//-----------------------------------------------------------------------//
//                          DivFunction                                  //
//-----------------------------------------------------------------------//

DoubleVector DivFunction::grad(const DoubleVector& x) const
{
  double fx = (*f).val(x);
  return (*f).grad(x)/(-fx*fx);
}

DoubleMatrix DivFunction::hess(const DoubleVector& x) const
{
  double fx = (*f).val(x);
  DoubleVector gradx = (*f).grad(x);
  DoubleMatrix F = (*f).hess(x);
  return ((-fx)*F + DyadeMatrix(2*gradx, gradx)) / (fx*fx*fx);
}

//-----------------------------------------------------------------------//
//                          SumFunction                                  //
//-----------------------------------------------------------------------//

DoubleVector SumFunction::grad(const DoubleVector& x) const
{
  return (*f).grad(x) + (*g).grad(x);
}

DoubleMatrix SumFunction::hess(const DoubleVector& x) const
{
 return (*f).hess(x) + (*g).hess(x);
}

//-----------------------------------------------------------------------//
//                          FunctionList
//-----------------------------------------------------------------------//

FunctionList::FunctionList(const FunctionList& g)
{
}

double FunctionList::val(const DoubleVector& x) const
{
  double sum=0;
  for (int i=0; i<f.Count; i++)
    sum += f[i].val(x);
    
  return sum;
}

DoubleVector FunctionList::grad(const DoubleVector& x) const
{
  DoubleVector Result(dimension());
  Result.MakeZero();
  
  for (int i=0; i<f.Count; i++)
    Result += f[i].grad(x);
    
  return Result;
}

DoubleMatrix FunctionList::hess(const DoubleVector& x) const
{
  int n=dimension();
  DoubleMatrix Result(n,n);
  Result.MakeZero();
  
  for (int i=0; i<f.Count; i++)
    Result += f[i].hess(x);
    
  return Result;
}

double FunctionList::hessleftright(const DoubleVector& x, const DoubleVector& d, const DoubleVector& e) const
{
  double Result = 0;
  for (int i=0; i<f.Count; i++)
    Result += f[i].hessleftright(x,d,e);
  return Result;    
}

int FunctionList::dimension() const
{
  int n=0;
  if (f.Count > 0)
    n = f[0].dimension();
    
  return n;
}

void FunctionList::operator+= (FunctionList& g)
{
  for (int i=0; i<g.Size(); i++)
    f.Insert(&g[i]);
}

void FunctionList::operator*= (double t)
{
  for (int i=0; i<f.Count; i++)
    f[i] *= t;
}

std::ostream& operator<<(std::ostream& to, const FunctionList&) 
{
  return to << "this is a FunctionList" << std::endl;
}

std::istream& operator>>(std::istream& from, FunctionList&) 
{
  return from;
}

//-----------------------------------------------------------------------//
//                          Function1D
//-----------------------------------------------------------------------//

int Function1D::MinimizeNewton(double& t, double eps, int maxiter, double a, double b) const
{
  double ft  = Deriv(t);
  double ftt = DDeriv(t); 

  int iter = 0;  

std::cout << "inside MinimizeNewton()" << std::endl;
std::cout << "a,b,eps,maxiter = " << a << " " << b << " " << eps << " " << maxiter << std::endl;

std::cout << "t      = " << t << std::endl;
std::cout << "f(t)   = " << (*this)(t) << std::endl;
std::cout << "f'(t)  = " << ft << std::endl;
std::cout << "f''(t) = " << ftt << std::endl;
  
  do {
std::cout << "iter = " << iter << std::endl;  

    if (fabs(ftt) < 1e-20) {
      std::cerr << "|f''(t)| < 1e-20 in Function1D::MinimizeNewton!!" << std::endl;
      ftt = 1e-20;
    }
    t -= (ft/ftt);
    ft  = Deriv(t);
    ftt = DDeriv(t); 

std::cout << "t      = " << t << std::endl;
std::cout << "f(t)   = " << (*this)(t) << std::endl;
std::cout << "f'(t)  = " << ft << std::endl;
std::cout << "f''(t) = " << ftt << std::endl;
    
    if (fabs(ft) < eps) break; // convergence test
    iter++;
  } while ((a <= t) && (t <= b) && (iter < maxiter));

  return ((a <= t) && (t <= b) && (iter < maxiter));
}

void Function1D::TestTaylor(double t, double h, int n) const
{
  double f   = (*this)(t);
  double ft  = Deriv(t);
  double ftt = DDeriv(t);

  std::cout << "       h              O(h)           O(h^2)          O(h^3)" << std::endl;

  for (int i=0; i<n; i++) {
    double g = (*this)(t+h);
    double val1 = g - f;
    double val2 = g - f - h*ft;
    double val3 = g - f - h*ft - 0.5*h*h*ftt;
    std::cout << std::setw(15) << h << std::setw(15) << val1  << std::setw(15) << val2 << std::setw(15) << val3 << std::endl;
    h *= 0.5;
  }
}

//-----------------------------------------------------------------------//
//                          Function3D
//-----------------------------------------------------------------------//

double3 Function3D::Deriv(double t) const
// returns f'(t)
{
  std::cerr << "error: Function3D::Deriv is not implemented!";
  return double3(0,0,0);
}

double3 Function3D::DDeriv(double t) const
// returns f''(t)
{
  std::cerr << "error: Function3D::DDeriv is not implemented!";
  return double3(0,0,0);
}

void Function3D::TestTaylor(double t, double h, int n) const
{
  double3 F   = (*this)(t);
  double3 Ft  = Deriv(t);
  double3 Ftt = DDeriv(t);

  std::cout << "       h              O(h)           O(h^2)          O(h^3)" << std::endl;

  for (int i=0; i<n; i++) {
    double3 G = (*this)(t+h);
    double val1 = (G - F).Length();
    double val2 = (G - F - h*Ft).Length();
    double val3 = (G - F - h*Ft - 0.5*h*h*Ftt).Length();
    std::cout << std::setw(15) << h << std::setw(15) << val1  << std::setw(15) << val2 << std::setw(15) << val3 << std::endl;
    h *= 0.5;
  }
}

//-----------------------------------------------------------------------//
//                          PiecewiseFunction3D                          //
//-----------------------------------------------------------------------//
 
PiecewiseFunction3D::PiecewiseFunction3D(double tleft, const double3& left, double tmiddle,
                                    const double3& middle, double tright, const double3& right)
{
  t0 = tmiddle;
  a1 = (middle - left) / (tmiddle - tleft);
  a0 = left - tleft*a1;
  b1 = (right - middle) / (tright - tmiddle);
  b0 = right - tright*b1;
}
 
double3 PiecewiseFunction3D::operator()(double t) const
// returns f(t)
{
  if (t<=t0)
    return a0+a1*t;
  else
    return b0+b1*t;
}

