// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-12-1996

#include <geotyp.h>
#include <pause.h>

//-----------------------------------------------------------------------//
//                          Min                                          //
//-----------------------------------------------------------------------//
static inline double Min(double a, double b)
{
  return (a<b) ? a : b;
}

//-----------------------------------------------------------------------//
//                          Min                                          //
//-----------------------------------------------------------------------//
static inline double Min(double a, double b, double c)
{
  return (a<b) ? Min(a,c) : Min(b,c);
}

//-----------------------------------------------------------------------//
//                          Max                                          //
//-----------------------------------------------------------------------//
static inline double Max(double a, double b)
{
  return (a>b) ? a : b;
}

//-----------------------------------------------------------------------//
//                          Max                                          //
//-----------------------------------------------------------------------//
static inline double Max(double a, double b, double c)
{
  return (a>b) ? Max(a,c) : Max(b,c);
}

//-----------------------------------------------------------------------//
//                          EqualSigns                                   //
//-----------------------------------------------------------------------//
static inline double EqualSigns(double a, double b, double c)
{
  return ( ((a >= 0) && (b >= 0) && (c >= 0)) ||
           ((a <= 0) && (b <= 0) && (c <= 0))    
  );
}

//-----------------------------------------------------------------------//
//                          BaryCentricDeterminant                       //
//-----------------------------------------------------------------------//
// This function computes the determinant of the following 3x3-matrix:
//
//                      [ 1  1  1 ]
//                      [ A  B  C ]
//
// See also [Fong, p. 4]
static inline double BarycentricDeterminant(const double2& A,
                                            const double2& B,
                                            const double2& C)
{
  return (A.x()*B.y() + B.x()*C.y() + C.x()*A.y() - A.y()*B.x() - B.y()*C.x() - C.y()*A.x());
}

//-----------------------------------------------------------------------//
//                          Intersect2DLines                             //
//-----------------------------------------------------------------------//
// this function computes the intersection parameters of the two lines
// l = { x in R^2 | x = a + s b } and m = { x in R^2 | x =  c + t d }
// if the lines are (nearly) parallel, false is returned

static int Intersect2DLines(const double2& a, const double2& b, double& s,
                            const double2& c, const double2& d, double& t)
//
// s b - t d = c - a
//
// |b.x -d.x| (s)    (c.x - a.x)
// |b.y -d.y| (t)  = (c.y - a.y)
//
// (s)                         |-d.y  d.x| (c.x - a.x)
// (t) = 1/(b.y*d.x - b.x*d.y) |-b.y  b.x| (c.y - a.y)
{
  double det = b.y()*d.x() - b.x()*d.y();

  if (fabs(det) < 1e-20) {
    return 0;
  }

  double cax = c.x() - a.x();
  double cay = c.y() - a.y();
  s = (-d.y()*cax + d.x()*cay)/det;
  t = (-b.y()*cax + b.x()*cay)/det;

  return 1;
}

//-----------------------------------------------------------------------//
//                          Line2D                                       //
//-----------------------------------------------------------------------//

double Line2D::VerticalDistance(const double2& point) const
{
  double2 left = r;
  double2 right = r+s;
  double alpha = (point.x() - left.x()) / (right.x() - left.x());
  
  return point.y() - ((1-alpha) * left.y() + alpha * right.y());
}

int Line2D::IntersectLineSegment(const double2& A, const double2& B, double& t) const
{
  double alpha;

  if (Intersect2DLines(A, (B-A), alpha, r, s, t)) {
    return (0 <= alpha) && (alpha <= 1);
  }
  else
    return 0;
}

int Line2D::IntersectLine(const double2& A, const double2& B, double& t) const
// if n is perpendicular to AB then we have to find the solution of
// the equation (n, r + t s - A) = 0
{
  double2 n(A.y() - B.y(), B.x() - A.x());
  double ns = (n,s);

  if (fabs(ns) < 1e-20)
    return 0;
  else {
    t = (n, A-r)/ns;
    return 1;
  }
}

//-----------------------------------------------------------------------//
//                          LineSegment2D                                //
//-----------------------------------------------------------------------//

double LineSegment2D::Distance(const double2& x) const
{
  double s,t;

  double2 n(A.y() - B.y(), B.x() - A.x());

  Intersect2DLines(A, B-A, s, x, n, t);

  double ds = Max(-s, 0, s-1);

  return sqrt( (t*t+ds*ds)*(B-A,B-A) );
}

//-----------------------------------------------------------------------//
//                          Line3D                                       //
//-----------------------------------------------------------------------//

void Line3D::GetPlanes(Plane3D& V0, Plane3D& V1) const
// returns two planes V0 and V1, such that this line is the intersection
// of V0 and V1
{
  s.PerpendicularVectors(V0.n, V1.n);
  V0.d = (V0.n,r);
  V1.d = (V1.n,r);
}  

double Line3D::Distance(const double3& p, double& tmin) const
// We have to minimize the squared distance function
//
//     f(t) = || r + s t - p ||^2
//
//          = (s,s) t^2 + 2(s,r-p) t + (r-p,r-p)
{
  double srp = (s,r-p);
  double ss  = (s,s);

  tmin = - srp/ss;                         // minimum of f(t)
  double dist = (r-p,r-p) - (srp*srp)/ss;  // f(tmin)

  return sqrt(fabs(dist));                 // fabs is necessary!!!
}

double Line3D::Distance(const Line3D& l, double& tmin) const
// Suppose the distance from a point x to l is given by ||Ax + b||
// Then we have to minimize the function
//
//          f(t) = ||A(r + s t) + b||^2
//
//               = (v,v) t^2 + 2(v,w) t + (w,w)
//
// with v = As and w = Ar + b.
{
  double3x3 A;
  double3 b;

  l.GetDistanceMap(A,b);

  double3 v = A*s;
  double3 w = A*r + b;
  double vv = (v,v);
  double vw = (v,w);
  double ww = (w,w);

  if (v.NearZero())
    tmin = 0;
  else
    tmin = - vw/vv;

  return ww + tmin*(2*vw + tmin*vv);
}

void Line3D::GetDistanceMap(double3x3& A, double3& b) const
{
  double ss = (s,s);
  A.MakeUnit();
  A -= DyadeMatrix(s,s)/ss;

  b = ((r,s)/ss)*s - r;
}

double3 Line3D::Rotate(const double3& p, double angle)
// the rotation p' of the point p along the line 
// { x | x = t n, t in R, ||n|| = 1 }, over the angle a
// is given by the formula
//
//  p'  =  cos a p + (1 - cos a)(p,n)n + sin a (n^p)
{
  double cosa = cos(angle);
  double sina = sin(angle);
  double3 n = s.Unit();   // the direction vector must have unit length

  double3 xtrans = p - r; // translate the point p first
  double3 image = cosa*xtrans + (1-cosa)*(xtrans,n)*n + sina * (n^xtrans);
  return image + r;       // translate back
}

//-----------------------------------------------------------------------//
//                          LineSegment3D                                //
//-----------------------------------------------------------------------//
// Suppose the distance from a point x to l is given by ||A0x + b||
// and suppose r = A and s = B-A.
// Then we have to minimize the function f: [0,1] -> R
//
//          f(t) = ||A0(r + s t) + b||^2
//
//               = (v,v) t^2 + 2(v,w) t + (w,w)
//
// with v = A0s and w = A0r + b.

double LineSegment3D::Distance(const Line3D& l, double& tmin) const
{
  double3 r = A;
  double3 s = B-A;

  double3x3 A0;
  double3 b;

  l.GetDistanceMap(A0,b);

  double3 v = A0*s;
  double3 w = A0*r + b;
  double vv = (v,v);
  double vw = (v,w);
  double ww = (w,w);

  if (v.NearZero())
    tmin = 0.5;
  else {
    tmin = - vw/vv;
    if (tmin < 0) tmin = 0;
    if (tmin > 1) tmin = 1;
  }

  return sqrt(ww + tmin*(2*vw + tmin*vv));
}

//-----------------------------------------------------------------------//
//                          Plane3D                                      //
//-----------------------------------------------------------------------//

int Plane3D::IntersectLine(const Line3D& line, double& t) const
{
  double ns = (n, line.s);
  if (fabs(ns) < 1e-20) {
    return 0;
  }

  t = (d - (n,line.r))/ns;
  return 1;
}

void Plane3D::GetParallelProjectionMap(const double3& r, double3x3& A, double3& b)
{
  double3x3 I3;
  I3.MakeUnit();
  double nr = (n,r);
  A = (I3 - DyadeMatrix(r,n)) / nr;
  b = (d/nr)*r;
}

//-----------------------------------------------------------------------//
//                          Rectangle2D                                  //
//-----------------------------------------------------------------------//

void Rectangle2D::MakeCube()
{
  leftdown = double2(-1,-1);
  width = 2;
  height = 2;
}

void Rectangle2D::GetRectangleToCubeMap(double2x2& A, double2& b) const
{
  A = double2x2(2/width, 0, 0, 2/height);
  b = - (A*Center());
}

void Rectangle2D::GetCubeToRectangleMap(double2x2& A, double2& b) const
{
  A = double2x2(width/2, 0, 0, height/2);
  b = Center();
}

void Rectangle2D::GetSquareToRectangleMap(double2x2& A, double2& b) const
// computes a linear mapping x -> A x + b such that
//
//              A (0,0) + b = r0
//              A (1,0) + b = r1
//              A (1,1) + b = r2
//              A (0,1) + b = r3
//
// where r0 ... r3 are the corners of this rectangle
{
  A = double2x2(width, 0, 0, height);
  b = leftdown;
}

//-----------------------------------------------------------------------//
//                          Triangle2D                                   //
//-----------------------------------------------------------------------//

double2 Triangle2D::operator[](int i) const
{
  switch (i) {
    case 0: return A;
    case 1: return B;
    case 2: return C;
    default:
      std::cerr << "Bad index in Triangle2D::operator[]" << std::endl;
  }

  return A;
}

double2& Triangle2D::operator[](int i)
{
  switch (i) {
    case 0: return A;
    case 1: return B;
    case 2: return C;
    default:
      std::cerr << "Bad index in Triangle2D::operator[]" << std::endl;
  }

  return A;
}

double Triangle2D::Area() const
{
  double2 AminB = A-B;
  double2 AminC = A-C;
  double X = AminB^AminC;
  return sqrt(X*X)/2;
}

void Triangle2D::GetBarycentricCoords(const double2& x, double& u, double& v) const
{
  double det = BarycentricDeterminant(A, B, C);
  u = BarycentricDeterminant(x, B, C) / det;
  v = BarycentricDeterminant(A, x, C) / det;
}

int Triangle2D::ContainsPoint(const double2& x, double EPS) const
{
  double u, v;

  GetBarycentricCoords(x, u, v);
  double w = 1-u-v;

  return ( (u >= 0) && (u <= 1) &&
           (v >= 0) && (v <= 1) &&
           (w >= 0) && (w <= 1)    ) || (BoundaryDistance(x) < EPS);
}

void Triangle2D::MakeEquiLateral()
{
  A = double2(0,0);
  B = double2(1,0);
  C = double2(0.5, 0.866025403784439);
}

void Triangle2D::MakeSimplex()
{
  A = double2(0,0);
  B = double2(1,0);
  C = double2(0,1);
}

int Triangle2D::IntersectEdge(int edge, const Line2D& line, double& t) const
{
  switch (edge) {
    case 0: return line.IntersectLineSegment(A,B,t);
    case 1: return line.IntersectLineSegment(B,C,t);
    case 2: return line.IntersectLineSegment(C,A,t);
  }

  return 0;
}

double Triangle2D::BoundaryDistance(const double2& x) const
{
  double distAB = LineSegment2D(A,B).Distance(x);
  double distBC = LineSegment2D(B,C).Distance(x);
  double distCA = LineSegment2D(C,A).Distance(x);

  return Min(distAB, distBC, distCA);
}

void Triangle2D::GetBarycentricMap(double2x2& AA, double2& b) const
{
  AA = double2x2(B.y()-C.y(), C.x()-B.x(), C.y()-A.y(), A.x()-C.x());
  b  = double2(B.x()*C.y() - B.y()*C.x(), A.y()*C.x() - A.x()*C.y());
}

void Triangle2D::GetTriangleToSimplexMap(double2x2& AA, double2& b) const
//                 
//             AA (B-A  C-A) = (1 0)       etc.
//                           = (0 1)
{
  double2x2 M(B.x()-A.x(), C.x()-A.x(), B.y()-A.y(), C.y()-A.y());
  AA = M.Inverse();
  b = - (AA*A);
}

void Triangle2D::GetSimplexToTriangleMap(double2x2& AA, double2& b) const
{
  AA = double2x2(B.x()-A.x(), C.x()-A.x(), B.y()-A.y(), C.y()-A.y());
  b  = A;
}

List<double2> Triangle2D::GetGridPoints(int n)
//
//                                     ...
//        
//        
//                              ...             ...
//        
//        
//                      x               x               ...
//                  (n-1,0,1)        (n-2,1,1)   
//        
//             x                 x                 x             ...
//          (n,0,0)          (n-1,1,0)         (n-2,2,0)
//
//--------------------------------------------------------------------------
//
//                      x
//               (n-i-1,i-j,j+1)
//
//             x               x
//        (n-i,i-j,j)     (n-i-1,i-j+1,j)
//
//                      x
//               (n-i,i-j+1,j-1)
//
//--------------------------------------------------------------------------
//
// numbering:
//
//
//         .
//       5   .
//     2   4   .
//   0   1   3   6     etc.
//
//   i = int(ceil(n + 0.5 - 0.5*sqrt(1 + 8*index)));
//   j = (n-i)*(n-i+3)/2 - index;
//   k = n-i-j;
// 
//   index(i,j,k) = (j+k)*(j+k+1)/2 + k
// 
{
  int NumPoints = (n+1)*(n+2)/2;
  List<double2> Result(NumPoints,1);

  for (int i=n; i>=0; i--)
    for (int j=n-i; j>=0; j--)
      Result.Insert(new double2((i*A + j*B + (n-i-j)*C)/n));

  return Result;
}

static inline int Index(int i, int j, int k) { return (j+k)*(j+k+1)/2 + k; }

List<int3> Triangle2D::GetGridTriangles(int n)
{
  int NumTriangles = n*n;                     // 1 + 3 + 5 + ...
  List<int3> Result(NumTriangles,1);

  for (int i=0; i<n; i++) {
    int j;

    // two loops for the triangles
    // (n-i,i-j,j) (n-i-1,i-j+1,j) (n-i-1,i-j,j+1) and
    // (n-i,i-j,j) (n-i,i-j+1,j-1) (n-i-1,i-j+1,j)  

    for (j=0; j<=i; j++) {
      int v0 = Index(n-i,i-j,j);
      int v1 = Index(n-i-1,i-j+1,j);
      int v2 = Index(n-i-1,i-j,j+1);
      Result.Insert(new int3(v0,v1,v2));
    }
    for (j=1; j<=i; j++) {
      int v0 = Index(n-i,i-j,j);
      int v1 = Index(n-i,i-j+1,j-1);
      int v2 = Index(n-i-1,i-j+1,j);
      Result.Insert(new int3(v0,v1,v2));
    }
  }

  return Result;
}

//-----------------------------------------------------------------------//
//                          Triangle3D                                   //
//-----------------------------------------------------------------------//

double Triangle3D::Area() const
{
  double3 AminB = A-B;
  double3 AminC = A-C;
  double3 X = AminB^AminC;
  return sqrt((X,X))/2;
}

double Triangle3D::BoundaryDistance(const Line3D& l) const
{
  double dummy;

  double distAB = LineSegment3D(A,B).Distance(l, dummy);
  double distBC = LineSegment3D(B,C).Distance(l, dummy);
  double distCA = LineSegment3D(C,A).Distance(l, dummy);

  return Min(distAB, distBC, distCA);
}

int Triangle3D::IntersectLine(const Line3D& l) const
// Suppose the line has the following representation:
//
//     l = { x in R^3 | x = r + t s, t in R }
//
// We define three planes V0, V1 and V2 as follows:
//
//   V0 is the plane through the points A, B and parallel to s
//   V1 is the plane through the points B, C and parallel to s
//   V2 is the plane through the points C, A and parallel to s
//
// Then there will be an intersection point if the point r is on the 'right' side of all
// three planes.
{
  double3 BminA = B-A;
  double3 AminC = A-C;
  double3 CminB = C-B;

  double3 nAB = BminA^l.s;
  double3 nBC = CminB^l.s;
  double3 nCA = AminC^l.s;

  double t0 = (nAB, (l.r-A));
  double t1 = (nBC, (l.r-B));
  double t2 = (nCA, (l.r-C));

  return EqualSigns(t0, t1, t2);
}

