// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 12-13-1995

#include <cmath>
#include <iostream>
#include <mystuff.h>
#include <ivector.h>

//-----------------------------------------------------------------------//
//                          IntVector                                    //
//-----------------------------------------------------------------------//

IntVector::IntVector(const IntVector& X, const IntVector& Y, const IntVector& Z)
{
  int n = X.NumRows();
  x = iv_get(3*n);
 
  for (int i=0; i<n; i++) {
    int row = 3*i;
    elem(row) = X(i);
    elem(row+1) = Y(i);
    elem(row+2) = Z(i);
  }
}

int IntVector::Length2() const
{
  int length = 0;

  for (int i=0; i<size(); i++)
    length += elem(i)*elem(i);

  return length;
}

void IntVector::Fill(int fill)
{
  for (int i=0; i<size(); i++)
    elem(i) = fill;
}

void IntVector::FillRandom(int min, int max)
{
  for (int i=0; i<size(); i++)
    elem(i) = Random(min,max);
}

void IntVector::MakeSequence()
{
  for (int i=0; i<size(); i++)
    elem(i) = i;
}

void IntVector::Split(int k, IntVector& y, IntVector& z) const
{
  int i;

  y.SetSize(k);
  z.SetSize(size()-k);

  for (i=0; i<k; i++)
    y[i] = elem(i);

  for (i=0; i<size()-k; i++)
    z[i] = elem(k+i);
}

IntVector IntVector::SubVector(int i0, int i1) const
{
  int size = (i1-i0+1);
  IntVector result(size);
  for (int i=0; i<size; i++)
    result[i] = elem(i0+i);

  return result;
}

IntVector IntVector::Concat(const IntVector& y) const
{
  int i;
  int m = size();
  int n = y.NumRows();

  IntVector result(m+n);
  for (i=0; i<m; i++)
    result[i] = elem(i);
  for (i=0; i<n; i++)
    result[m+i] = y[i];

  return result;
}

void IntVector::operator *= (int t)
{
  for (int i=0; i<size(); i++)
    elem(i) *= t;
}

IntVector operator-(const IntVector& x)
{
  IntVector result(x.size());

  iv_sub(result.vec(), x.vec(), result.vec());
  return result;
}

IntVector operator+(const IntVector& x, const IntVector& y)
{
  IntVector result(x.size());

  iv_add(x.vec(), y.vec(), result.vec());
  return result;
}

IntVector operator-(const IntVector& x, const IntVector& y)
{
  IntVector result(x.size());

  iv_sub(x.vec(), y.vec(), result.vec());
  return result;
}

IntVector operator*(int t, const IntVector& x)
{
  IntVector result(x.Size());

  for (int i=0; i<x.Size(); i++)
    result(i) = t*x(i);

  return result;
}

IntVector operator*(const IntVector& x, int t)
{
  return t*x;
}
 
int operator,(const IntVector& x, const IntVector& y)
{
  int result = 0;

  for (int i=0; i<x.Size(); i++)
    result += x(i)*y(i);
    
  return result;
}

int operator*(const IntVector&x, const IntVector& y)
{
  return (x,y);
}

std::ostream& operator<<(std::ostream& s, const IntVector& x)
{
//  s << '(';
  s << x.size() << " ";
  if (x.NumRows() > 0) {
    int i;
    for (i=0; i<x.NumRows()-1; i++)
      s << x[i] << " ";
    s << x[i];
  }
//  s << ')';
  return s;
}

std::istream& operator>>(std::istream& s, IntVector& x)
{
  int m;

  s >> m;

  x.SetSize(m);
  for (int i=0; i<m; i++)
    s >> x[i];
  return s;
}

int IntVector::Contains(int k) const
{
  for (int i=0; i<size(); i++)
    if (elem(i)==k) return 1;

  return 0;
}
