// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 15-5-95
#define DEBUG

#include <iostream>
#include <fstream>
#include <numtyp.h>
#include <numalg.h>

//-----------------------------------------------------------------------//
//                          ConjGradSolve                                //
//-----------------------------------------------------------------------//

int ConjGradSolve(Function& f, DoubleVector& x0, DoubleVector& x, int MaxIter, double Eps)
{
  DoubleVector x_new, d_new, g_new;

  int Converged = 0;
  x = x0;
  DoubleVector g = f.grad(x0);
  DoubleVector d = -g;

std::cout << "f(x) = " << f.val(x) << std::endl;
  for (int i=0; (i<MaxIter) && (!Converged); i++) {
    DoubleMatrix Fx = f.hess(x);
    DoubleVector Fxd = Fx*d;
    double dFxd = (d,Fxd);
    double alpha = -(g,d)/dFxd;
    x_new = x + alpha*d;
    g_new = f.grad(x_new);
    double beta = (g_new, Fxd)/dFxd;
    d_new = - g_new + beta*d;
    x = x_new;
    g = g_new;
    d = d_new;

std::cout << "i = " << i << " f(x) = " << f.val(x) << std::endl;
    Converged = (g.Length() < Eps);
  }

  if (!Converged) {
    char ch;
    std::cout << "ConjugateGradientMethod::Minimize() did not find a minimum." << std::endl;
    std::cout << "Use the current solution? y/n " << std::endl;
    std::cin >> ch;
    Converged = (ch == 'y');
  }

  return Converged;
}

//-----------------------------------------------------------------------//
//                          StedescSolve                                 //
//-----------------------------------------------------------------------//
// this function solves the linear system Ax = b
//
// in fact, it minimizes f(x) = 1/2 x^T A x - b^T x using steepest descent
//
// [see Shewchuk, app. B1]

int StedescSolve(SparseMatrix& A, DoubleVector& b, DoubleVector& x, int imax, double eps)
{
  int m = b.NumRows();

  if ( (x.NumRows() != m) || (A.NumRows()  != m) ||
       (A.NumCols()  != m) || (b.NumRows()  != m)    ) {
    std::cerr << "dimension error in StedescSolve()!" << std::endl;
    return 0;
  }

  int i=0;

  DoubleVector r = b - A*x;
  double delta = (r,r);
  double delta_0 = delta;
  
  while ( (i<imax) && (delta > (eps*eps)*delta_0) ) {

    double fx = 0.5*(x,A*x) - (b,x);
    std::cout << "|Ax - b| = " << std::setw(25) << std::setprecision(16) << (A*x - b).Length()
	 << "    f(x) = " << fx << std::setprecision(6) << std::endl;

    DoubleVector q = A*r;

    double alpha = delta/(r,q);

#ifdef DEBUG
{
  DoubleVector y = x + alpha*r;
  DoubleVector z = x + (1e-5)*r;
  double fy = 0.5*(y,A*y) - (b,y);
  double fz = 0.5*(z,A*z) - (b,z);
  if (fy > fx) {
    std::cout << "error: fy > fx !!     alpha = " << alpha << std::endl;
  }
  if (fz > fx) {
    std::cout << "error: direction ascending!!" << std::endl;
  }
}
#endif

    x += alpha*r;

    if (i%50 == 0)
      r = b - A*x;
    else
      r -= alpha*q;
    delta = (r,r);
    i++;
  }

  double fx = 0.5*(x,A*x) - (b,x);
  std::cout << "|Ax - b| = " << std::setw(25) << (A*x - b).Length()
       << "    f(x) = " << std::setprecision(16) << fx << std::setprecision(6) << std::endl;
  std::cout << "StedescSolve(): " << i << " iterations" << std::endl;
  std::cout << std::endl;

  return (i<imax);
}

//-----------------------------------------------------------------------//
//                          ConjGradSolve                                //
//-----------------------------------------------------------------------//
// this function solves the linear system Ax = b
//
// in fact, it minimizes f(x) = 1/2 x^T A x - b^T x using conjugate gradients
//
// [see Shewchuk, app. B2]

int ConjGradSolve(SparseMatrix& A, DoubleVector& b, DoubleVector& x, int imax, double eps)
{
  int m = b.NumRows();

  if ( (x.NumRows() != m) || (A.NumRows()  != m) ||
       (A.NumCols()  != m) || (b.NumRows()  != m)    ) {
    std::cerr << "dimension error in StedescSolve()!" << std::endl;
    return 0;
  }

#ifdef DEBUG
{
  std::ofstream conjgrad("conjgrad.mtx");
  conjgrad << b << std::endl;
  DoubleMatrix AA;
  A.GetMatrix(AA);
  conjgrad << AA << std::endl;
  conjgrad.close();
}
#endif

#ifdef DEBUG
  std::ofstream conjgradinfo("conjgrad.info");
  conjgradinfo << std::setprecision(16) << (A*x - b).Length() << std::endl;
#endif

  int i=0;

  DoubleVector r = b - A*x;
  DoubleVector d = r;
  double delta_new = (r,r);
  double delta_0 = delta_new;
  
  while ( (i<imax) && (delta_new > (eps*eps)*delta_0) ) {

    double fx = 0.5*(x,A*x) - (b,x);
    std::cout << "|Ax - b| = " << std::setw(25) << std::setprecision(16) << (A*x - b).Length()
	 << "    f(x) = " << fx << std::setprecision(6) << std::endl;

    DoubleVector q = A*d;

    double alpha = delta_new/(d,q);

#ifdef DEBUG
{
  DoubleVector y = x + alpha*d;
  DoubleVector z = x + (1e-5)*d;
  double fy = 0.5*(y,A*y) - (b,y);
  double fz = 0.5*(z,A*z) - (b,z);
  if (fy > fx) {
    std::cout << "error: fy > fx !     alpha = " << alpha << std::endl;
  }
  if (fz > fx) {
    std::cout << "error: direction ascending!!" << std::endl;
  }
}
#endif

    x += alpha*d;

    if (i%50 == 0)
      r = b - A*x;
    else
      r -= alpha*q;

    double delta_old = delta_new;
    delta_new = (r,r);
    double beta = delta_new/delta_old;
    d = r + beta*d;
    i++;
  }

  double fx = 0.5*(x,A*x) - (b,x);
  std::cout << "|Ax - b| = " << std::setw(25) << (A*x - b).Length()
       << "    f(x) = " << std::setprecision(16) << fx << std::setprecision(6) << std::endl;
  std::cout << "ConjGradSolve(): " << i << " iterations" << std::endl;
  std::cout << std::endl;

#ifdef DEBUG
  conjgradinfo << std::setprecision(16) << (A*x - b).Length() << std::endl;
  conjgradinfo << i << std::endl;
#endif

  return (i<imax);
}

//-----------------------------------------------------------------------//
//                          LUSolve                                      //
//-----------------------------------------------------------------------//

int LUSolve(const DoubleMatrix& A, const DoubleVector& b, DoubleVector& x)
{
  DoubleLUFact LU(A);
  x = LU.solve(b);

  std::cout << "LUSolve: |Ax - b| = " << (A*x - b).Length() << std::endl;

  return 1;
}

//-----------------------------------------------------------------------//
//                          QuadMinLUSolve                               //
//-----------------------------------------------------------------------//
// First the constraint equation is rewritten as
//
//                 Cx = d => x = Fy + g
//
// Substitution of this equation in the original problem yields an
// unconstrained minimization problem in y:
//
// minimize y^T F^T A F y + (2g^T A + b^T)F y + const
//
// grad(x^T A x + b^T x) = 2 A x + b = 0 etc.

int QuadMinLUDecomp(DoubleMatrix& A, DoubleVector& b, DoubleMatrix& C,
                    DoubleVector& d, DoubleVector& x)
{
  DoubleMatrix F;
  DoubleVector g;

  RewriteEquations(C, d, F, g);

  DoubleMatrix FtAF = F.Transpose()*A*F;
  DoubleVector u = (2*g*A + b)*F;

  DoubleVector y;

  // los op: FtAF y = -u/2
  if (LUSolve(FtAF, -0.5*u, y)) {
std::cout << "QuadMinLUSolve(): y = " << y << std::endl;
    x = F*y + g;
    return 1;
  }
  else
    return 0;
}

//-----------------------------------------------------------------------//
//                          SwapInts                                     //
//-----------------------------------------------------------------------//
inline void SwapInts(int& i, int& j)
{
  int tmp = i;
  i = j;
  j = tmp;
}

//-----------------------------------------------------------------------//
//                          SwapDoubles                                   //
//-----------------------------------------------------------------------//
inline void SwapDoubles(double& i, double& j)
{
  double tmp = i;
  i = j;
  j = tmp;
}
/*
//-----------------------------------------------------------------------//
//                          RewriteEquations                             //
//-----------------------------------------------------------------------//
void RewriteEquations(DoubleMatrix& A, DoubleVector& b, RewriteMatrix& C, DoubleVector& d)
{
  int i,icol,irow,j,k,l,ll,*ivector(int, int);
  double big,dum,pivinv;

  int m = A.NumRows();
  int n = A.NumCols();

  IntVector indxc(m);
  IntVector indxr(m);
  IntVector ipiv(n);
  ipiv.MakeZero();

  IntVector sigma(n);
  for (i=1; i<=n; i++)
    sigma.Elem(i) = i;

  for (i=1;i<=m;i++) {
    big=0.0;
    for (j=1;j<=m;j++)
      if (ipiv.Elem(j) != 1)
      for (k=1;k<=n;k++) {
        if (ipiv.Elem(k) == 0) {
          if (fabs(A.Elem(j,k)) >= big) {
            big=fabs(A.Elem(j,k));
            irow=j;
            icol=k;
          }
        }
        else
        if (ipiv.Elem(k) > 1)
          nrerror("GAUSSJ2: Singular Matrix-1");
      }
    ++(ipiv.Elem(irow));
    if (irow != icol) {
      for (l=1;l<=m;l++) SwapDoubles(A.Elem(l,irow), A.Elem(l,icol));
    }
    indxr.Elem(i)=irow;
    indxc.Elem(i)=icol;
    SwapInts(sigma.Elem(irow), sigma.Elem(icol));
    if (A.Elem(irow,irow) == 0.0) nrerror("GAUSSJ2: Singular Matrix-2");
    pivinv=1.0/A.Elem(irow,irow);
// A.Elem(irow,irow)=1.0; Dit moet of na de volgende regel of helemaal weg!!!!!!!!!
    for (l=1;l<=n;l++) A.Elem(irow,l) *= pivinv;
    b.Elem(irow) *= pivinv;
    for (ll=1;ll<=m;ll++)
    if (ll != irow) {
      dum=A.Elem(ll,irow);
// A.Elem(ll,irow)=0.0; Dit moet of na de volgende regel of helemaal weg!!!!!!!!!
      for (l=1;l<=n;l++) A.Elem(ll,l) -= A.Elem(irow,l)*dum;
      b.Elem(ll) -= b.Elem(irow)*dum;
    }
  }

  DoubleMatrix* F = new DoubleMatrix(m, n-m);
  for (i=1; i<=m; i++)
    for (j=1; j<=m; j++)
      F->Elem(i,j) = A.Elem(i,j+n-m);

  C = RewriteMatrix(F, sigma);

  d.SetSize(n);
  for (i=1; i<=m; i++)
    d.Elem(i) = b.Elem(sigma.Elem(i));
}
*/
//-----------------------------------------------------------------------//
//                          RewriteEquations                             //
//-----------------------------------------------------------------------//
void RewriteEquations(DoubleMatrix& A, DoubleVector& b, DoubleMatrix& C, DoubleVector& d)
{
/*
  int i,icol,irow,j,k,l,ll,*ivector(int, int);
  double big,dum,pivinv;

  int m = A.NumRows();
  int n = A.NumCols();

  C.SetSize(n, n-m);
  d.SetSize(n);

  IntVector indxc(m);
  IntVector indxr(m);
  IntVector ipiv(n);
  ipiv.MakeZero();

  for (i=1;i<=m;i++) {
    big=0.0;
    for (j=1;j<=m;j++)
      if (ipiv.Elem(j) != 1)
      for (k=1;k<=n;k++) {
        if (ipiv.Elem(k) == 0) {
          if (fabs(A.Elem(j,k)) >= big) {
            big=fabs(A.Elem(j,k));
            irow=j;
            icol=k;
          }
        }
        else
        if (ipiv.Elem(k) > 1)
	  std::cerr << "GAUSSJ2: Singular Matrix-1" << std::endl;
      }
    ++(ipiv.Elem(irow));
    if (irow != icol) {
      for (l=1;l<=m;l++) SwapDoubles(A.Elem(l,irow), A.Elem(l,icol));
    }
    indxr.Elem(i)=irow;
    indxc.Elem(i)=icol;
    if (A.Elem(irow,irow) == 0.0)
      std::cerr<< "GAUSSJ2: Singular Matrix-2" << std::endl;
    pivinv=1.0/A.Elem(irow,irow);
// A.Elem(irow,irow)=1.0; Dit moet of na de volgende regel of helemaal weg!!!!!!!!!
	 for (l=1;l<=n;l++) A.Elem(irow,l) *= pivinv;
	 b.Elem(irow) *= pivinv;
	 for (ll=1;ll<=m;ll++)
	 if (ll != irow) {
		dum=A.Elem(ll,irow);
// A.Elem(ll,irow)=0.0; Dit moet of na de volgende regel of helemaal weg!!!!!!!!!
		for (l=1;l<=n;l++) A.Elem(ll,l) -= A.Elem(irow,l)*dum;
		b.Elem(ll) -= b.Elem(irow)*dum;
	 }
  }

  for (i=1;i<=m;i++)
	 for (j=1;j<=n-m;j++) C.Elem(i,j) = -A.Elem(i,m+j);
  for (i=m+1;i<=n;i++)
	 for (j=1;j<=n-m;j++) C.Elem(i,j) = 0.0;
  for (i=1;i<=n-m;i++) C.Elem(m+i,i) = 1.0;
  for (i=1;i<=m;i++) d.Elem(i) = b.Elem(i);
  for (i=m+1;i<=n;i++) d.Elem(i) = 0.0;

  for (l=m;l>=1;l--) {
	 if (indxr.Elem(l) != indxc.Elem(l)) {
		for (k=1;k<=n-m;k++)
		  SwapDoubles(C.Elem(indxr.Elem(l), k), C.Elem(indxc.Elem(l), k));
		SwapDoubles(d.Elem(indxr.Elem(l)), d.Elem(indxc.Elem(l)));
	 }
  }
*/
}

const double SMALLFLOAT = 1.0E-20;

//-----------------------------------------------------------------------//
//                          Givens                                       //
//-----------------------------------------------------------------------//
void Givens(double a, double b, double& c, double& s)
{
  double tau;

  if (fabs(b) < SMALLFLOAT) {
    c=1; s=0;
  }
  else {
    if (fabs(b) > fabs(a)) {
      tau = -a/b; s = 1/sqrt(1+tau*tau); c = s*tau;
    }
    else {
      tau = -b/a; s = 1/sqrt(1+tau*tau); s = c*tau;
    }
  }
}

//-----------------------------------------------------------------------//
//                          FastGivens                                   //
//-----------------------------------------------------------------------//
void FastGivens(const double2& x, const double2& d0, double& alpha, double& beta, int& type)
{
  double gamma, tau;
  double2 d = d0;

  if (x.Elem(2) != 0) {
    alpha = -x.Elem(1)/x.Elem(2);
    beta  = - alpha*d.Elem(2)/d.Elem(1);
    gamma = - alpha*beta;
    if (gamma <= 1) {
      type = 1;
      tau = d.Elem(1);
      d.Elem(1) = (1+gamma)*d.Elem(2);
      d.Elem(2) = (1+gamma)*tau;
    }
    else {
      type = 2;
      alpha = 1/alpha; beta  = 1/beta; gamma = 1/gamma;
      d.Elem(1) = (1+gamma)*d.Elem(1);
      d.Elem(2) = (1+gamma)*d.Elem(2);
    }
  }
  else {
    type = 2;
    alpha = 0; beta = 0;
  }
}

//-----------------------------------------------------------------------//
//                          GivensQR                                     //
//-----------------------------------------------------------------------//
void GivensQR(DoubleMatrix& B, DoubleMatrix& A)
{
  int i,j;
  double c,s;

  int m = A.NumRows();
  int n = A.NumCols();

  if (m<n) {
    std::cerr << "dimension error in GivensQR()" << std::endl;
    return;
  }

  if ( (B.NumRows() != m) || (B.NumCols() != m) ) {
    if ((B.NumRows() != 0) || (B.NumCols() != 0))
      std::cerr << "dimension error in GivensQR()" << std::endl;
    B.SetSize(m,m);
  }

  B.MakeUnit();

  for (j=1; j<=n; j++) {
    for (i=m; i>=j+1; i--) {
      Givens(A.Elem(i-1,j), A.Elem(i,j), c, s);
      A.RowRot(i-1, i, j, c, s);
      B.RowRot(i-1, i, 1, c, s);

// N.B. the effect of the last two statements is equal to
//              A = GivensMatrix(m, i-1, i, c, s).Transpose()*A;
//              B = GivensMatrix(m, i-1, i, c, s).Transpose()*Q;
    }
  }
}

//-----------------------------------------------------------------------//
//                          FastGivensQR                                 //
//-----------------------------------------------------------------------//
void FastGivensQR(DoubleMatrix& A, DoubleVector& d)
{
  int i,j;

  int m = A.NumRows();
  int n = A.NumCols();

  if (m<n || d.NumRows() != m) {
    std::cerr << "dimension error in FastGivensQR()" << std::endl;
    return;
  }

  double alpha, beta;
  int type;

  for (i=1; i<=m; i++) d.Elem(i) = 1;
  for (j=1; j<=n; j++) {
    for (i=m; i>=j+1; i--) {
      FastGivens(double2(A.Elem(i-1,j), A.Elem(i,j)),
                 double2(d.Elem(i-1), d.Elem(i)),
                 alpha, beta, type );
      A.RowFastRot(i-1, j, alpha, beta, type);
    }
  }
}

//-----------------------------------------------------------------------//
//                          GivensMatrix                                 //
//-----------------------------------------------------------------------//
DoubleMatrix GivensMatrix(int n, int i, int k, double c, double s)
{
  DoubleMatrix G(n,n);
  G.MakeUnit();
  G.Elem(i,i) = c;
  G.Elem(k,k) = c;
  G.Elem(i,k) = s;
  G.Elem(k,i) = -s;

  return G;
}

/*--- old stuff -------------------------------------------------------
// nog:
// -> testen van de vorm 'double == 0' eruit halen

void ModifiedGramSchmidt(DoubleMatrix& A, DoubleMatrix& R)
// See [Golub p. 219]
// Given A in R[m,n] with rank(A)= n, the following algorithm computes
// the factorization A = QR where Q in R[m,n] has orthonormal columns
// and R in R[n,n] is upper triangular
{
}

// De routine QuadMin lost het volgende kwadratische programmeringsprobleem op:
//
// minimaliseer          1/2 x^T Q x + c^T x
//
// onder voorwaarde dat  A x = b
//
// met Q een symmetrische positief (semi)definiete nxn-matrix
// en  A een mxn-matrix met rang m (m <= n)
//
// Zie ook Luenberger p. 423 e.v.
//
// De Lagrange necessary conditions voor dit probleem zijn:
//
//           Q x + A^T lambda + c = 0
//           A x - b              = 0
// oftewel
//
//  | Q   A^T | |  x   |   |-c|
//  |         | |      | = |  |
//  | A    0  | |lambda|   | b|
//
// Dit lineaire stelsel wordt m.b.v. de geconjugeerde gradientmethode
// opgelost.

int QuadMin(DoubleMatrix& Q, DoubleVector& c, DoubleMatrix& A, DoubleVector& b,
            int MaxIter, double Tol, DoubleVector& x, DoubleVector& lambda)
// x contains startvalue!!
{
//std::cout << "Q = " << Q << std::endl;
//std::cout << "c = " << c << std::endl;
//std::cout << "A = " << A << std::endl;
//std::cout << "b = " << b << std::endl;
//std::cout << "x = " << x << std::endl;

  LagrangeMatrix L(Q,A);

  int m = A.NumRows();

  // write the problem in Lagrange form L x = d

//  DoubleVector lambda(m); // vector of Lagrange multipliers
//  lambda.MakeZero();
  DoubleVector x0 = x.Concat(lambda);
  DoubleVector d  = (-c).Concat(b);

//std::cout << "lambda = " << lambda << std::endl;
//std::cout << "x0 = " << x0 << std::endl;
//std::cout << "d = " << d << std::endl;

  int ReturnValue = ConjugateGradientSolve(L, d, x0, Tol, MaxIter);
  if (ReturnValue) {
    x      = x0.SubVector(0,x.NumRows()-1);
    lambda = x0.SubVector(x.NumRows(), x.NumRows()+m-1);
  }

  return ReturnValue;
}

//-----------------------------------------------------------------------//
//                          ConjGradSolve                                //
//-----------------------------------------------------------------------//
// Internally the problem A(x+x0) = b is solved with x=0 as startvalue.
// The real solution is then x+x0.

int ConjGradSolve(SparseMatrix& A, DoubleVector& b, DoubleVector& x0, DoubleVector& x, int MaxIter, double Eps)
{
  int m = b.NumRows();

  if ( (x0.NumRows() != m) || (A.NumRows()  != m) ||
       (A.NumCols()  != m) || (b.NumRows()  != m)    ) {
    std::cerr << "dimension error in ConjGradSolve()!" << std::endl;
    return 0;
  }

  double rho_prev;
  DoubleVector p(m);
  DoubleVector w(m);

  x.SetSize(m);
  x.MakeZero();

  DoubleVector r = (b-A*x0);
  double rho = (r,r);

  int k=0;
  double MaxRho = Eps*(b,b);

  while (rho > MaxRho && k < MaxIter) {
std::cout << "x = " << x << std::endl;
std::cout << "|Ax - b| = " << (A*x - b).Length() << std::endl;
std::cout << "f(x) = " << std::setprecision(16) << 0.5*(x,A*x) - (b,x) << std::setprecision(6) << std::endl;

    k++;
    if (k==1)
      p=r;
    else {
      double beta = rho/rho_prev;
      p = r + beta*p;
    }
    w = A*p;
    double alpha = rho/(p,w);
    x += alpha*p;
    r -= alpha*w;
    rho_prev = rho;
    rho = (r,r);
//std::cout << "ConjGradSolve(): |Ax-b| = " << (A*(x+x0) - b).Length() << std::endl;
  }

std::cout << "ConjGradSolve(): |Ax-b| = " << (A*(x+x0) - b).Length() << std::endl;
std::cout << "ConjGradSolve(): " << k << " iterations" << std::endl;
std::cout << std::endl;

  x += x0;

  return (k<MaxIter);
}

*/
