// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Fri 06-21-1996

#include <cmath>
#include <iostream>
#include <numtyp.h>

//-----------------------------------------------------------------------//
//                          SparseVector                                 //
//-----------------------------------------------------------------------//

void SparseVector::SelfTest() const
// this method tests if the virtual methods are implemented correctly
{
  DoubleVector b;
  GetVector(b);

  if (NumRows() != b.NumRows())
    std::cerr << "Dimension error in SparseVector implementation!" << std::endl;

  DoubleVector x(NumRows());
  x.FillRandom(-100,100);

  std::cout << "SparseVector::SelfTest: " << (b,x) - (*this,x) << std::endl;
}

//-----------------------------------------------------------------------//
//                          IndexedVector                                //
//-----------------------------------------------------------------------//
// The class IndexedVector is a SparseVector x for which the nonzero elements
// are explicitly stored.
//
//         x(index(i)) = b(i)    0 <= i < b.Size() 
//                        0      otherwise

void IndexedVector::AddToVector(DoubleVector& Y) const
{
  for (int i=0; i<index.Size(); i++)
    Y(index(i)) += b(i);
}

void IndexedVector::MultAddToVector(double t, DoubleVector& Y) const
{
  for (int i=0; i<index.Size(); i++)
    Y(index(i)) += t*b(i);
}

double IndexedVector::GetInnerProduct(const DoubleVector& Y) const
{
  double Result = 0;

  for (int i=0; i<index.Size(); i++)
    Result += b(i)*Y(index(i));

  return Result;
}

void IndexedVector::Show() const
{
  std::cout << "numrows, b = " << numrows << " " << b << std::endl;
  std::cout << "index   = " << index << std::endl;
}

void IndexedVector::Read(std::ifstream& from)
{
  from >> numrows >> index >> b;
}

void IndexedVector::Write(std::ofstream& to) const
{
  to << numrows << " " << index << std::endl << b << std::endl;
}

//-----------------------------------------------------------------------//
//                          QuadVector                                   //
//-----------------------------------------------------------------------//

void QuadVector::AddToVector(DoubleVector& Y) const
{
  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    double3 Q = b[i]*P;
    int row = 3*index[i];
    Y.AddDouble3(row, Q);
  }
}

void QuadVector::MultAddToVector(double t, DoubleVector& Y) const
{
  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    double3 Q = t*b[i]*P;
    int row = 3*index[i];
    Y.AddDouble3(row, Q);
  }
}
   
double QuadVector::GetInnerProduct(const DoubleVector& Y) const
{
  double Result = 0;

  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    double3 Q = b[i]*P;
    int row = 3*index[i];
    Result += (Q[0]*Y(row) + Q[1]*Y(row+1) + Q[2]*Y(row+2));
  }

  return Result;
}

void QuadVector::Show() const
{
  std::cout << "numrows, b, P = " << numrows << " " << b << " " << P << std::endl;
  std::cout << "index   = " << index << std::endl;
}

void QuadVector::Read(std::ifstream& from)
{
  from >> numrows >> index >> b >> P;
}

void QuadVector::Write(std::ofstream& to) const
{
  to << numrows << " " << index << std::endl << b << " " << P << std::endl;
}

//-----------------------------------------------------------------------//
//                          SparseMatrix                                 //
//-----------------------------------------------------------------------//

void SparseMatrix::SelfTest() const
// this method tests if the virtual methods are implemented correctly
{
  DoubleMatrix A;
  GetMatrix(A);

  if ((NumRows() != A.NumRows()) || (NumCols() != A.NumCols()) )
    std::cerr << "Dimension error in SparseMatrix implementation!" << std::endl;

  DoubleVector left(NumRows());
  left.FillRandom(-100,100);
  DoubleVector right(NumCols());
  right.FillRandom(-100,100);

  std::cout << "SparseMatrix::SelfTest left:  " << (left*(*this) - left*A).Length() << std::endl;
  std::cout << "SparseMatrix::SelfTest right: " << ((*this)*right - A*right).Length() << std::endl;
}

void SparseMatrix::SolveBKP(const DoubleVector& b, DoubleVector& x) const
{
  DoubleMatrix BKP;
  GetMatrix(BKP);

  PERM* pivot = px_get(NumRows());
  PERM* blocks = px_get(NumRows()); // ???
  BKPfactor(BKP.mat(), pivot, blocks);

  VEC* y = BKPsolve(BKP.mat(), pivot, blocks, b.vec(), VNULL);
  x = DoubleVector(y);

  PX_FREE(pivot);
  PX_FREE(blocks); 
}

void SparseMatrix::SolveCH(const DoubleVector& b, DoubleVector& x) const
{
  DoubleMatrix CH;
  GetMatrix(CH);

  CHfactor(CH.mat());
  VEC* y = CHsolve(CH.mat(), b.vec(), VNULL);
  x = DoubleVector(y);
}

void SparseMatrix::SolveLDL(const DoubleVector& b, DoubleVector& x) const
{
  DoubleMatrix LDL;
  GetMatrix(LDL);

  LDLfactor(LDL.mat());
  VEC* y = LDLsolve(LDL.mat(), b.vec(), VNULL);
  x = DoubleVector(y);
}

void SparseMatrix::SolveLU(const DoubleVector& b, DoubleVector& x) const
{
  DoubleMatrix LU;
  GetMatrix(LU);
  PERM* pivot = px_get(NumRows()); // ???
  LUfactor(LU.mat(), pivot);

  VEC* y = LUsolve(LU.mat(), pivot, b.vec(), VNULL);
  x = DoubleVector(y);
  
  PX_FREE(pivot);
}

void SparseMatrix::SolveQR(const DoubleVector& b, DoubleVector& x) const
{
  DoubleMatrix QR;
  GetMatrix(QR);
  
  VEC* diag = v_get(NumRows()); // ???
  
  QRfactor(QR.mat(), diag);
  VEC* y = QRsolve(QR.mat(), diag, b.vec(), VNULL);
  x = DoubleVector(y);
  
  V_FREE(diag);
}

int SparseMatrix::SolveSteepestDescent(const DoubleVector& b, DoubleVector& x, int imax, double eps) const
// Solves the system Ax=b using the steepest descent method
{
  int i=0;
  DoubleVector r = b - (*this)*x;
  double delta = (r,r);
  //double delta0 = delta;
  
  while ( (i<imax) && (delta > eps*eps) ) {
    std::cout << "i = " << i << " " << delta << std::endl;
    DoubleVector q = (*this)*r;
    double alpha = delta / (r,q);
    x += (alpha*r);
    if (i%50 == 0)
      r = b - (*this)*x;
    else
      r -= (alpha*q);
    delta = (r,r);
    i++;
  }

  std::cout << "i = " << i << " " << delta << std::endl;
  return (delta <= eps*eps);
}

int SparseMatrix::SolveConjugateGradient(const DoubleVector& b, DoubleVector& x, int imax, double eps) const
// Solves the system Ax=b using the conjugate gradient method
{
  int i=0;
  DoubleVector r = b - (*this)*x;
  DoubleVector d = r;  
  double delta_new = (r,r);
  //double delta0 = delta_new;
  
  while ( (i<imax) && (delta_new > eps*eps) ) {
    std::cout << "i = " << i << " " << delta_new << std::endl;
    DoubleVector q = (*this)*d;
    double alpha = delta_new / (d,q);
    x += (alpha*d);
    if (i%50 == 0)
      r = b - (*this)*x;
    else
      r -= (alpha*q);
    double delta_old = delta_new;  
    delta_new = (r,r);
    double beta = delta_new / delta_old;
    d = r + beta*d;
    i++;
  }
  
  std::cout << "i = " << i << " " << delta_new << std::endl;
  return (delta_new <= eps*eps);
}

int SparseMatrix::SolveConjugateResidual(const DoubleVector& b, DoubleVector& x, int imax, double eps) const
// Solves the system Ax=b using the conjugate residual method
// N.B. Note that the stop criterion is very poor...
{
  std::cout << "SparseMatrix::SolveConjugateResidual()" << std::endl;
  DoubleVector r = b - (*this)*x;
  DoubleVector p = r;

  DoubleVector p_last(NumRows(), 0);
  DoubleVector Ap_last(NumRows(), 0);
  double pAp_last = 0;
  
  double error = r.Length();
  
  int i=0;
  
  while ( (error > eps*eps) && (i < imax) ) { 
    std::cout << "i = " << i << " " << error << std::endl;
    DoubleVector Ap  = (*this)*p;
    double pAp = (p,Ap);
    double ApAp = (Ap,Ap);
    
    double sigma = (p,r) / pAp;
    x += (sigma*p);
    r -= (sigma*Ap);
    
    double gamma = ApAp / pAp;
    double delta = (i==0) ? 0 : (Ap,Ap_last) / pAp_last;
    DoubleVector p_new = Ap - gamma*p - delta*p_last;
    
    p_last   = p;
    p        = p_new;
    Ap_last  = Ap;
    pAp_last = pAp;

    i++;
    error = ((*this)*x - b).Length2(); // this is expensive!
  }
  
  std::cout << "i = " << i << " " << error << std::endl;
  return (error < eps);
}

int SparseMatrix::SolveSTOD(const DoubleVector& b, DoubleVector& x, int imax, double eps) const
// Solves the system Ax=b using stabilized orthogonal directions
// N.B. Note that the stop criterion is very poor...
{
  std::cout << "SparseMatrix::SolveSTOD()" << std::endl;
  DoubleVector r = b - (*this)*x;
  DoubleVector p = r;

  DoubleVector p_last(NumRows(), 0);
  double ApAp_last = 0;

  double error = r.Length2();
  
  int i=0;
  
  while ( (error > eps*eps) && (i < imax) ) { 
    std::cout << "i = " << i << " " << error << std::endl;
    DoubleVector Ap  = (*this)*p;
    DoubleVector AAp = (*this)*Ap;
    double ApAp = (Ap,Ap);
    double ApAAp = (Ap,AAp);
    
    double sigma = (r,p) / ApAp;
    x += (sigma*Ap);
    r -= (sigma*AAp);
    
    double gamma = ApAAp / ApAp;
    double delta = (i==0) ? 0 : (ApAp / ApAp_last);
    DoubleVector p_new = Ap - gamma*p - delta*p_last;
    
    p_last    = p;
    p         = p_new;
    ApAp_last = ApAp;

    i++;
    error = ((*this)*x - b).Length2(); // this is expensive!
  }
  
  std::cout << "i = " << i << " " << error << std::endl;
  return (error < eps);
}

//-----------------------------------------------------------------------//
//                          SparseMatrixList                             //
//-----------------------------------------------------------------------//

void SparseMatrixList::AddToMatrix(DoubleMatrix& B) const
{
  for (int i=0; i<f.Count; i++)
    f[i].AddToMatrix(B);
}

DoubleVector SparseMatrixList::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector result(NumCols());
  result.MakeZero();
  
  for (int i=0; i<f.Count; i++)
    result += f[i].GetVectorMatrixProduct(X);
    
  return result;
}

DoubleVector SparseMatrixList::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector result(NumRows());
  result.MakeZero();
  
  for (int i=0; i<f.Count; i++)
    result += f[i].GetMatrixVectorProduct(X);
    
  return result;
}

void SparseMatrixList::operator*= (double t)
{
  for (int i=0; i<f.Count; i++)
    f[i] *= t;
}

int SparseMatrixList::NumRows() const
{
  if (f.Count <= 0)
    return 0;
  else
    return f[0].NumRows();
}

int SparseMatrixList::NumCols() const
{
  if (f.Count <= 0)
    return 0;
  else
    return f[0].NumCols();
}

//-----------------------------------------------------------------------//
//                          BlockMatrix                                  //
//-----------------------------------------------------------------------//

void BlockMatrix::Show() const
{
  std::cout << "dimension = " << 3*numrows << std::endl;
  std::cout << "index = " << index << std::endl;
  std::cout << "M = " << M << std::endl;
}

void BlockMatrix::AddToMatrix(DoubleMatrix& A) const
{
  int n = index.NumRows();
  for (int i=0; i<n; i++)
    for (int j=0; j<n; j++)	{
      int row = 3*index[i];
      int col = 3*index[j];
      A(row, col)   += M[i][j];
      A(row+1,col+1) += M[i][j];
      A(row+2,col+2) += M[i][j];
    }
}

DoubleVector BlockMatrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumRows());
  Result.MakeZero();

  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      Result(row)   += M[i][j] * X(col);
      Result(row+1) += M[i][j] * X(col+1);
      Result(row+2) += M[i][j] * X(col+2);
    }
  }
  
  return Result;
}

DoubleVector BlockMatrix::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      Result(col)   += M[i][j] * X(row);
      Result(col+1) += M[i][j] * X(row+1);
      Result(col+2) += M[i][j] * X(row+2);
    }
  }
  
  return Result;
}

//-----------------------------------------------------------------------//
//                          SharedBlockMatrix                            //
//-----------------------------------------------------------------------//

void SharedBlockMatrix::Show() const
{
  for (int i=0; i<indices.Count; i++)
    std::cout << i << " " << indices[i] << std::endl;
  std::cout << "M = " << M << std::endl;
}

void SharedBlockMatrix::AddToMatrix(DoubleMatrix& A) const
{
  for (int m=0; m<indices.Count; m++) {
    const IntVector& index = indices[m];
    int n = index.NumRows();
    for (int i=0; i<n; i++)
      for (int j=0; j<n; j++)	{
        for (int m=0; m<indices.Count; m++) {
          int row = 3*indices[m](i);
          int col = 3*indices[m](j);
          A(row, col)    += M(i,j);
          A(row+1,col+1) += M(i,j);
          A(row+2,col+2) += M(i,j);
        }  
      }
  }
}

DoubleVector SharedBlockMatrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumRows());
  Result.MakeZero();

  for (int m=0; m<indices.Count; m++) {
    const IntVector& index = indices[m];
    int n = index.NumRows();
    for (int i=0; i<n; i++) {
      int row = 3*index[i];
      for (int j=0; j<n; j++)	{
        int col = 3*index[j];
        Result(row)   += M(i,j) * X(col);
        Result(row+1) += M(i,j) * X(col+1);
        Result(row+2) += M(i,j) * X(col+2);
      }
    }
  }
  
  return Result;
}

DoubleVector SharedBlockMatrix::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  for (int m=0; m<indices.Count; m++) {
    const IntVector& index = indices[m];
    int n = index.NumRows();
    for (int i=0; i<n; i++) {
      int row = 3*index[i];
      for (int j=0; j<n; j++)	{
        int col = 3*index[j];
        Result(col)   += M[i][j] * X(row);
        Result(col+1) += M[i][j] * X(row+1);
        Result(col+2) += M[i][j] * X(row+2);
      }
    }
  }
  
  return Result;
}

//-----------------------------------------------------------------------//
//                          Quad2Matrix                                  //
//-----------------------------------------------------------------------//

void Quad2Matrix::Show() const
{
  std::cout << "index = " << index << std::endl;
  std::cout << "M = " << M << std::endl;
  std::cout << "A = " << A << std::endl;
}

void Quad2Matrix::AddToMatrix(DoubleMatrix& B) const
{
  int n = index.NumRows();
  for (int i=0; i<n; i++)
    for (int j=0; j<n; j++)	{
      int row = 3*index[i];
      int col = 3*index[j];
      B.AddDouble3x3(row, col, M[i][j]*A);
    }
}

DoubleVector Quad2Matrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumRows());
  Result.MakeZero();

  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      double3x3 B = M[i][j]*A;
      Result(row)   += ( B[0][0]*X(col) + B[0][1]*X(col+1) + B[0][2]*X(col+2) );
      Result(row+1) += ( B[1][0]*X(col) + B[1][1]*X(col+1) + B[1][2]*X(col+2) );
      Result(row+2) += ( B[2][0]*X(col) + B[2][1]*X(col+1) + B[2][2]*X(col+2) );
    }
  }
  
  return Result;
}

DoubleVector Quad2Matrix::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      double3x3 B = M[i][j]*A;
      Result(col)   += ( X(row)*B[0][0] + X(row+1)*B[1][0] + X(row+2)*B[2][0] );
      Result(col+1) += ( X(row)*B[0][1] + X(row+1)*B[1][1] + X(row+2)*B[2][1] );
      Result(col+2) += ( X(row)*B[0][2] + X(row+1)*B[1][2] + X(row+2)*B[2][2] );
    }
  }
  
  return Result;
}

//-----------------------------------------------------------------------//
//                          LagrangeMatrix                               //
//-----------------------------------------------------------------------//

DoubleVector 
LagrangeMatrix::GetMatrixVectorProduct(const DoubleVector& x) const
{
  DoubleVector x1, x2;
  x.Split((*A).NumRows(), x1, x2); 

  DoubleVector y1 = (*A)*x1 + x2*(*B);
  DoubleVector y2 = (*B)*x1;

  return DoubleVector(y1,y2);
}

DoubleVector LagrangeMatrix::GetVectorMatrixProduct(const DoubleVector& x) const
{
  DoubleVector x1, x2;
  x.Split((*A).NumRows(), x1, x2); 

  DoubleVector y1 = x1*(*A) + x2*(*B);
  DoubleVector y2 = (*B)*x1;

  return DoubleVector(y1,y2);
}

void LagrangeMatrix::AddToMatrix(DoubleMatrix& C) const
{
  if ( C.NumRows() != NumRows() || C.NumCols() != NumCols() ) {
  	std::cerr << "dimension error in LagrangeMatrix::AddToMatrix()" << std::endl;
  }
  std::cerr << "LagrangeMatrix::AddToMatrix not yet implemented!" << std::endl;
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  for (int i=0; i<A.NumRows(); i++) {
    for (int j=0; j<A.NumCols(); j++)
      C[i][j] += A[i][j];
  }

  for (i=0; i<B.NumRows(); i++) {
    for (int j=0; j<B.NumCols(); j++) {
      C[A.NumRows()+i][j] += B[i][j];
      C[j][A.NumCols()+i] += B[i][j];
    }  
  }
*/
}

//-----------------------------------------------------------------------//
//                          PermutationMatrix                            //
//-----------------------------------------------------------------------//

DoubleVector PermutationMatrix::GetMatrixVectorProduct(const DoubleVector& x) const
{
  return A*(x.Permute(sigma));
}

DoubleVector PermutationMatrix::GetVectorMatrixProduct(const DoubleVector& x) const
{
  return (x*A).Permute(sigma);
}

void PermutationMatrix::AddToMatrix(DoubleMatrix& B) const
{
  if ( B.NumRows() != NumRows() || B.NumCols() != NumCols() ) {
  	std::cerr << "dimension error in PermutationMatrix::AddToMatrix()" << std::endl;
  }

  DoubleMatrix matrixA;
  A.GetMatrix(matrixA);

  for (int i=0; i<A.NumRows(); i++)
    for (int j=0; j<A.NumCols(); j++)
      B[i][j] += matrixA[i][sigma[j]];
}

//-----------------------------------------------------------------------//
//                          BTABMatrix                                   //
//-----------------------------------------------------------------------//

DoubleVector BTABMatrix::GetMatrixVectorProduct(const DoubleVector& x) const
{
  return (A*(B*x)) * B;
}

DoubleVector BTABMatrix::GetVectorMatrixProduct(const DoubleVector& x) const
{
  return (B*x) * A * B;
}

void BTABMatrix::AddToMatrix(DoubleMatrix& C) const
{
  DoubleMatrix matrixA;
  DoubleMatrix matrixB;
  A.GetMatrix(matrixA);
  B.GetMatrix(matrixB);
  C += matrixB.Transpose()*matrixA*matrixB;
}

//-----------------------------------------------------------------------//
//                          BandMatrix                                   //
//-----------------------------------------------------------------------//

static int Min(int i, int j) { return (i<j) ? i : j; }
static int Max(int i, int j) { return (i<j) ? j : i; }

void BandMatrix::AddToMatrix(DoubleMatrix& A) const
{
  for (int i=0; i<numrows; i++) {
    int jmin = Max(0, i-lbw);
    int jmax = Min(numcols-1, i+ubw);

    for (int j=jmin; j<=jmax; j++)
      A(i,j) += elem(i,j);
  }    
}

DoubleVector BandMatrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  
  for (int i=0; i<numrows; i++) {
    int jmin = Max(0, i-lbw);
    int jmax = Min(numcols-1, i+ubw);
    
    for (int j=jmin; j<=jmax; j++)
      Result(i) += elem(i,j)*X(j);
  }    
  
  return Result;
}

DoubleVector BandMatrix::GetVectorMatrixProduct(const DoubleVector& X) const
{ 
  DoubleVector Result(NumRows());
  
  for (int j=0; j<numcols; j++) {
    int imin = Max(0, j-ubw);
    int imax = Min(numrows-1, j+lbw);
    
    for (int i=imin; i<=imax; i++)
      Result(j) += elem(i,j)*X(i);
  }    

  return Result;
}

void BandMatrix::Show() const
{
  DoubleMatrix A(NumRows(), NumCols());
  GetMatrix(A);
  std::cout << A << std::endl;
}

DoubleVector BandMatrix::SolveCholesky(const DoubleVector& b)
// see Golub p. 155
{
  int p = ubw;
  int n = NumRows();
  int lambda;

  for (int i=0; i<n; i++) {
    for (int j = Max(0,i-p); j<i; j++) {
      lambda = Min(j+p+1,n);
      for (int k=i; k<lambda; k++)
        elem(k,i) -= elem(i,j)*elem(k,j);
    }
    lambda = Min(i+p+1,n);
    double factor = sqrt(elem(i,i));

    for (int k=i; k<lambda; k++)
      elem(k,i) /= factor;
  }

  return DoubleVector();
}

DoubleVector BandMatrix::LowerTriangularSolveLeft(const DoubleVector& b)
{
  int n = NumRows();
  DoubleVector x(n);

  for (int j=n-1; j>=0; j--) {
    double d = b(j);
    int i;
    for (i=Min(n-1,j+lbw); i>j; i--)
      d -= elem(i,j)*x(i);
    x(i) = d/elem(i,i);
  }

  return x;
}

DoubleVector BandMatrix::LowerTriangularSolve(const DoubleVector& b)
{
  int n = NumRows();
  DoubleVector x(n);

  for (int i=0; i<n; i++) {
    double d = b(i);
    for (int j=Max(0,i-lbw); j<i; j++)
      d -= elem(i,j)*x(j);
    x(i) = d/elem(i,i);
  }

  return x;
}

//-----------------------------------------------------------------------//
//                          QuadMatrix                                   //
//-----------------------------------------------------------------------//

void QuadMatrix::Show() const
{
  std::cout << "index = " << index << std::endl;
  std::cout << "M = " << M << std::endl;
}

void QuadMatrix::AddToMatrix(DoubleMatrix& A) const
{
  int n = index.NumRows();
  for (int i=0; i<n; i++)
    for (int j=0; j<n; j++)	{
      int row = 3*index[i];
      int col = 3*index[j];
      A(row, col)   += M[i][j];
      A(row+1,col+1) += M[i][j];
      A(row+2,col+2) += M[i][j];
    }
}

DoubleVector QuadMatrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumRows());
  Result.MakeZero();

  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      Result(row)   += M[i][j] * X(col);
      Result(row+1) += M[i][j] * X(col+1);
      Result(row+2) += M[i][j] * X(col+2);
    }
  }
  
  return Result;
}

DoubleVector QuadMatrix::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  int n = index.NumRows();
  for (int i=0; i<n; i++) {
    int row = 3*index[i];
    for (int j=0; j<n; j++)	{
      int col = 3*index[j];
      Result(col)   += M[i][j] * X(row);
      Result(col+1) += M[i][j] * X(row+1);
      Result(col+2) += M[i][j] * X(row+2);
    }
  }
  
  return Result;
}

/*

int SparseMatrix::SolveSTOD(const DoubleVector& b, DoubleVector& x, int imax, double epsSTOD) const
// Solves the system Ax=b using stabilized orthogonal directions
{
  int hoog = NumRows();
  int breed = NumCols();

  if (hoog != breed) std::cerr << "hoog != breed";
  if (x.Size() != breed) std::cerr << "x.Size() != breed";

  DoubleVector p = b - (*this) * x;
  DoubleVector residu=p;
  DoubleVector lastp =p;
  lastp.Fill(0);
  DoubleVector newp = p;
 
  DoubleVector newAp, Ap, last_Ap;
  Ap = (*this)*p; last_Ap = (*this)*lastp;
  DoubleVector AAp = (*this)*Ap;
  int i = 0;
  DoubleVector sigmaMp;
  double Ap_len2, last_Ap_len2  = 0;
 
  while (i < imax) {
    // stoppen?
    std::cout << i << " " << ((*this)*x - b).Length2() << std::endl;
    if (residu.Length2() < epsSTOD)
      break;
    Ap_len2 = Ap.Length2();
    double sigma = residu * p / Ap_len2;
 
    AAp = (*this)*Ap;   // EXPENSIVE
    x += sigma * Ap;
    residu -= sigma * AAp;
 
    double delta = (i > 0) ? Ap_len2 / last_Ap_len2 : 0;
    double gamma = Ap* AAp/ (Ap_len2);
 
    newp = Ap - gamma * p - delta * lastp;
    lastp = p;
    p = newp;
 
    newAp= AAp - gamma*Ap-delta*last_Ap;
    last_Ap = Ap;
    Ap = newAp;
    last_Ap_len2 = Ap_len2;
    i++;
  }
 
std::cout << " |Ax - b| = " << ((*this)*x - b).Length() << std::endl;
std::cout << " # iteraties " << i << std::endl;

  return (i<imax);
}

*/ 
