// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-25-1996

#include "parsurf.h"

//-----------------------------------------------------------------------//
//                          ParametricSurface                            //
//-----------------------------------------------------------------------//

double3 ParametricSurface::GetDerivative(const double2& u, int k, int l) const
// returns the partial derivative
//
//              d^(k+l) F(u,v)
//              --------------
//                du^k dv^l
{
  std::cerr << "ParametricSurface::GetDerivative() not yet implemented!";
  return double3(0,0,0);
}

double3 ParametricSurface::GetDirectionalDerivative(const double2& u, const double2& d, int k) const
// returns the directional derivative
//
//               (k)
//              g   (t)             with g(t) = F(u+td)
{
  std::cerr << "ParametricSurface::GetDirectionalDerivative() not yet implemented!";
  return double3(0,0,0);
}

void ParametricSurface::GetFuncSamples(const double2& u, FuncSample2D& F1, FuncSample2D& F2, FuncSample2D& F3) const
// computes samples of the coordinate functions F1, F2 and F3 at the point with parameter u
{
  SurfSample sample = GetSurfSample(u);
  
  F1.f = sample.F.x();
  F2.f = sample.F.y();
  F3.f = sample.F.z();
  
  F1.fu = sample.Fu.x();
  F2.fu = sample.Fu.y();
  F3.fu = sample.Fu.z();
  
  F1.fv = sample.Fv.x();
  F2.fv = sample.Fv.y();
  F3.fv = sample.Fv.z();
  
  F1.fuu = sample.Fuu.x();
  F2.fuu = sample.Fuu.y();
  F3.fuu = sample.Fuu.z();
  
  F1.fuv = sample.Fuv.x();
  F2.fuv = sample.Fuv.y();
  F3.fuv = sample.Fuv.z();
  
  F1.fvv = sample.Fvv.x();
  F2.fvv = sample.Fvv.y();
  F3.fvv = sample.Fvv.z();
}         

double3 ParametricSurface::GetNormal(const double2& u) const
// returns the normal of the surface at the point with parameter u
{
  return GetSurfSample(u).Normal();
}

double3 ParametricSurface::GetOffsetPoint(const double2& u, double d) const
// returns F(u) + d*N(u), with N the unit normal
{
  SurfSample sample = GetSurfSample(u);
  return sample.F + d*sample.Normal();
}

double3 ParametricSurface::GetPoint(const double2& u) const
// returns the point of the surface with parameter u
{
  return GetSurfSample(u).F; 
}

SurfSample ParametricSurface::GetSurfSample(const double2& u) const
// computes a sample of F at u
{
  SurfSample Result;
  std::cerr << "ParametricSurface::GetDirectionalDerivative() not yet implemented!";
  return Result;
}

void ParametricSurface::GetSurfSampleExt(const double2& u, SurfSampleExt& sample) const
// computes a sample of F at u with some extra info
{
  SurfSample s = GetSurfSample(u);
  sample.SetValues(s.F, s.Fu, s.Fv, s.Fuu, s.Fuv, s.Fvv);
}

Plane3D ParametricSurface::GetTangentPlane(const double2& u) const
// returns the tangent plane at the point with parameter u
{
  SurfSample sample = GetSurfSample(u);
  double3 normal = sample.Normal();
  double d = (normal, sample.F);
  
  return Plane3D(normal, d);
}

double ParametricSurface::GetTraceExpr(const double2& u, const FuncSample2D& f) const
// returns the following expression:
//              2   2 
//     trace( (D  f)  ) * sqrt(g_F)
//              F
{
  SurfSampleExt sample;
  GetSurfSampleExt(u, sample);
  
  double2x2 D2f = sample.I_inv*( f.Hessian() - f.fu*sample.gamma1 - f.fv*sample.gamma2 );
  return (D2f*D2f).Trace()*sqrt(sample.g);
}

double ParametricSurface::SimpleStretchEnergy(const double2& u) const
{
   return GetSurfSample(u).SimpleArea();
}

double ParametricSurface::DataStretchEnergy(const ParametricSurface& G, const double2& u) const
{
  SurfSample sampleF = GetSurfSample(u);
  SurfSample sampleG = G.GetSurfSample(u);
  double2x2 I_F = sampleF.FirstFundamentalForm();
  double2x2 I_G = sampleG.FirstFundamentalForm();       
  return sqrt(I_G.Det()) * (I_G.Inverse()*I_F).Trace();
}  

double ParametricSurface::ExactStretchEnergy(const double2& u) const
{
  return sqrt(GetSurfSample(u).FirstFundamentalForm().Det());
}    

double ParametricSurface::SimpleBendEnergy(const double2& u) const
{
  return GetSurfSample(u).SimpleTotalCurvature();
}

double ParametricSurface::DataBendEnergy(const ParametricSurface& G, const double2& u) const
{
  FuncSample2D F1, F2, F3;
  GetFuncSamples(u, F1, F2, F3);
  return G.GetTraceExpr(u,F1) + G.GetTraceExpr(u,F2) + G.GetTraceExpr(u,F3);
}  

double ParametricSurface::ExactBendEnergy(const double2& u) const
{
  SurfSample sampleF = GetSurfSample(u);
  double2x2 I_F = sampleF.FirstFundamentalForm();
  return sampleF.TotalCurvature() * sqrt(I_F.Det());
}    

double ParametricSurface::PointDistance(const double2& u, const double3& point) const
// returns  d(F(u), point)^2
{
  return (GetPoint(u) - point).Length2();
}

double ParametricSurface::LineDistance(const double2& u, const Line3D& line) const
// returns  d(F(u), line)^2
{
  double dist = line.Distance(GetPoint(u));
  return dist*dist;
}

double ParametricSurface::PlaneDistance(const double2& u, const Plane3D& plane) const
// returns  d(F(u), plane)^2
{
  double dist = plane.Distance(GetPoint(u));
  return dist*dist;
}

double ParametricSurface::NormalDistance(const double2& u, const double3& normal) const
// returns  (Fu(u),normal)^2 + (Fv(u),normal)^2
{
  SurfSample sample = GetSurfSample(u);
  double d1 = (sample.Fu,normal);
  double d2 = (sample.Fv,normal);
  return d1*d1 + d2*d2;
}  

