// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <polygon.h>

//-----------------------------------------------------------------------//
//                          Min                                          //
//-----------------------------------------------------------------------//
static inline double Min(double a, double b)
{
  return (a<b) ? a : b;
}

//-----------------------------------------------------------------------//
//                          Polygon2D                                    //
//-----------------------------------------------------------------------//

Polygon2D::Polygon2D(const Polygon2D& polygon)
// copy constructor
{
  for (int i=0; i<polygon.NumPoints(); i++)
    Insert(polygon.Point(i));
}   

double2 Polygon2D::Center() const
// returns the center of the vertices of the polygon
{
  double2 Result(0,0);
  
  for (int i=0; i<NumPoints(); i++)
    Result += Point(i);
    
  return Result/NumPoints();
}

void Polygon2D::Scale(const double2& P, double t)
// scales the polygon with a factor t w.r.t. the point P
{
  for (int i=0; i<NumPoints(); i++)
    Point(i) = P + t*(Point(i)-P);
}

void Polygon2D::Transform(const double2x2& A, const double2& b)
// applies the mapping x -> Ax+b to all the vertices
{
  for (int i=0; i<points.Count; i++)
    points[i] = A*points[i] + b;
}

double Polygon2D::Area() const
// returns the area of the polygon
{
  double Result = 0;

  int left, right, top, bottom;
  GetLeftRightTopBottom(left, right, top, bottom);

  for (int i=left; i!=right; i = (i+1)%NumPoints()) {
    int next = (i+1)%NumPoints();
    double area = (x(next)-x(i))*(y(i)+y(next))/2;
    Result += area;
  }

  for (int i=left; i!=right; i = (i+NumPoints()-1)%NumPoints()) {
    int next = (i+NumPoints()-1)%NumPoints();
    double area = (x(next)-x(i))*(y(i)+y(next))/2;
    Result -= area;
  }
  
  return fabs(Result);
}

void Polygon2D::GetLeftRightTopBottom(int& left, int& right, int& top, int& bottom) const
// returns the indices of the point with the smallest/largest (x/y)-coordinate
{
  left=0;
  right=0;
  top=0;
  bottom=0;

  for (int i=1; i<NumPoints(); i++) {
    if (x(i) < x(left))   left   = i;
    if (x(i) > x(right))  right  = i;
    if (y(i) > y(top))    top    = i;
    if (y(i) < y(bottom)) bottom = i;
  }
}

int Polygon2D::IsConvex(double precision)
{
  int pos=0; // the number of times that we turn left
  int neg=0; // the number of times that we turn right

  for (int i=0; i<NumPoints()-1; i++) {
    double val = (points[i+1] - points[i]) ^ (points[(i+2) % NumPoints()] - points[i+1]);
    if (val > precision) pos++;
    if (val < -precision) neg++;
  }

  return (pos == 0) || (neg == 0);
}

int Polygon2D::IsClockwise() const
// PRECONDITION: the polygon is convex
// returns true if the points of the polygon are ordered clockwise
{
  int left, right, top, bottom;
  GetLeftRightTopBottom(left, right, top, bottom);
  
  return
    ( (left<right) && (left<=bottom) && (bottom<=right) ) ||
    ( (left>right) && ((left<=bottom)||(bottom<=right)) );
}

HorizontalTrapezoid2D Polygon2D::GetHorizontalTrapezoid(double ymin, double ymax, const int2& left, const int2& right) const
// returns the trapezoid bounded by the lines
//
//               y=ymin
//               y=ymax
//
// and the line segments
//
//      [ Point(left[0]), Point(left[1]) ]
//      [ Point(right[0]), Point(right[1]) ]
//
{
  double a1 = (x(left[1])-x(left[0])) / (y(left[1])-y(left[0]));
  double a0 = x(left[0]) - a1*y(left[0]);
  
  double b1 = (x(right[1])-x(right[0])) / (y(right[1])-y(right[0]));
  double b0 = x(right[0]) - b1*y(right[0]);
  
  return HorizontalTrapezoid2D(ymin, ymax, a0, a1, b0, b1);
}

VerticalTrapezoid2D Polygon2D::GetVerticalTrapezoid(double xmin, double xmax, const int2& bottom, const int2& top) const
// returns the trapezoid bounded by the lines
//
//               x=xmin
//               x=xmax
//
// and the line segments
//
//      [ Point(bottom[0]), Point(bottom[1]) ]
//      [ Point(top[0]), Point(top[1]) ]
//
{
  double a1 = (y(bottom[1])-y(bottom[0])) / (x(bottom[1])-x(bottom[0]));
  double a0 = y(bottom[0]) - a1*x(bottom[0]);
  
  double b1 = (y(top[1])-y(top[0])) / (x(top[1])-x(top[0]));
  double b0 = y(top[0]) - b1*x(top[0]);
  
  return VerticalTrapezoid2D(xmin, xmax, a0, a1, b0, b1);
}

List<HorizontalTrapezoid2D> Polygon2D::GetHorizontalTrapezoids(int SplitAtIntegers, double precision) const
// PRECONDITION: the polygon is convex
// splits the polygon in horizontal trapezoids by cutting it with
// horizontal lines through the vertices
//
// SplitAtIntegers: if true, then the polygon is also cut at horizontal
//                  lines y=yi, with yi an integer
// 
// precision: trapezoids with (height < precision) are neglected
{
  List<HorizontalTrapezoid2D> Result;

  int dummy1, dummy2, bottom, top;
  GetLeftRightTopBottom(dummy1, dummy2, top, bottom);

  int delta;
  if (IsClockwise())
    delta = 1;
  else
    delta = -1;

  int2 left(bottom, (bottom-delta+NumPoints()) % NumPoints()); 
  int2 right(bottom, (bottom+delta+NumPoints()) % NumPoints());

  double ymin = y(bottom);

  while (ymin < y(top)) {
    if (SplitAtIntegers) {
      double ymax = Min(y(left[1]), y(right[1]));
      for (int yy=ceil(ymin); yy<ymax; yy++) {
        if (yy - ymin > precision)
          Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, yy, left, right));
        ymin = yy;
      }  
    }
    
    if (y(left[1]) <= y(right[1])) {
      if (y(left[1]) - ymin > precision)
        Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, y(left[1]), left, right));
      ymin = y(left[1]);
      left[0] = left[1];
      left[1] = (left[1] - delta + NumPoints()) % NumPoints();
    }
    else {
      if (y(right[1]) - ymin > precision)
        Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, y(right[1]), left, right));
      ymin = y(right[1]);
      right[0] = right[1];
      right[1] = (right[1] + delta + NumPoints()) % NumPoints();
    }
  }

  return Result;
}

List<VerticalTrapezoid2D> Polygon2D::GetVerticalTrapezoids(int SplitAtIntegers, double precision) const
// PRECONDITION: the polygon is convex
// splits the polygon in vertical trapezoids by cutting it with
// vertical lines through the vertices
//
// SplitAtIntegers: if true, then the polygon is also cut at vertical
//                  lines x=xi, with xi an integer
// 
// precision: trapezoids with (width < precision) are neglected
{
  List<VerticalTrapezoid2D> Result;

  int left, right, dummy1, dummy2;
  GetLeftRightTopBottom(left, right, dummy1, dummy2);

  int delta;
  if (IsClockwise())
    delta = 1;
  else
    delta = -1;

  int2 bottom(left, (left+delta+NumPoints()) % NumPoints()); 
  int2 top(left, (left-delta+NumPoints()) % NumPoints());

  double xmin = x(left);

  while (xmin < x(right)) {

    if (SplitAtIntegers) {
      double xmax = Min(x(bottom[1]), x(top[1]));
      for (int xx=ceil(xmin); xx<xmax; xx++) {
        if (xx - xmin > precision)
          Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, xx, bottom, top));
        xmin = xx;
      }
    }  
    
    if (x(bottom[1]) <= x(top[1])) {
      if (x(bottom[1]) - xmin > precision)
        Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, x(bottom[1]), bottom, top));
      xmin = x(bottom[1]);
      bottom[0] = bottom[1];
      bottom[1] = (bottom[1] + delta + NumPoints()) % NumPoints();
    }
    else {
      if (x(top[1]) - xmin > precision)
        Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, x(top[1]), bottom, top));
      xmin = x(top[1]);
      top[0] = top[1];
      top[1] = (top[1] - delta + NumPoints()) % NumPoints();
    }
  }

  return Result;
}

List<VerticalTrapezoid2D> Polygon2D::GetTrapezoids(double precision) const
// PRECONDITION: the polygon is convex
// splits the polygon in vertical trapezoids such that any piece
// lies inside a square [i,i+1] x [j,j+1] with i and j integers
{
  List<VerticalTrapezoid2D> Result;
  
  List<HorizontalTrapezoid2D> trapezoids = GetHorizontalTrapezoids(1,precision);
  for (int i=0; i<trapezoids.Count; i++)
    Result += trapezoids[i].GetVerticalTrapezoids(1,precision);
    
  return Result;
}

void Polygon2D::GetTriangulation(List<Triangle2D>& triangles) const
{
  if (NumPoints()<3)
    return;

  for (int i=0; i<NumPoints()-2; i++)
    triangles.Insert(new Triangle2D(points[0], points[i+1], points[i+2]));
}

void Polygon2D::Show() const
// writes the contents to standard out
{
  for (int i=0; i<NumPoints(); i++)
    std::cout << "point " << i << ": " << Point(i) << std::endl;
  std::cout << "area: " << Area() << std::endl;
  std::cout << "center: " << Center() << std::endl;
}

std::ostream& operator<<(std::ostream& os, const Polygon2D& polygon)
{
  std::cout << polygon.NumPoints();
  for (int i=0; i<polygon.NumPoints(); i++)
    os << " " << polygon.Point(i);

  return os;
}

std::istream& operator>>(std::istream& is, Polygon2D& polygon)
{ 
  int n;
  is >> n;
  for (int i=0; i<n; i++) {
    double2 point;
    is >> point;
    polygon.Insert(point);
  }

  return is;
}






/*======================================================================
void Polygon2D::InsertXIntegers()
// inserts some vertices lying on the boundary of the polygon
{
  int left, right, top, bottom;
  GetLeftRightTopBottom(left, right, top, bottom);
  
  int i = ceil(x(left));

  while (i < x(right)) {
  	do { left = (left + 1) % NumPoints(); } while (x(left) < i); 	
  	int prev = (left - 1 + NumPoints()) % NumPoints();
  	// x(prev) < i <= x(left) 	

	while (i < x(left)) {
      double alpha = (i-x(prev))/(x(left)-x(prev));
      double yi = (1-alpha)*y(prev) + alpha*y(left);
  	  points.AtInsert(left++, new double2(i++,yi));
  	}
  }
}

void Polygon2D::InsertYIntegers()
// inserts some vertices lying on the boundary of the polygon
{
  int left, right, top, bottom;
  GetLeftRightTopBottom(left, right, top, bottom);
  
  int i = ceil(y(bottom));

  while (i < y(top)) {
  	do { bottom = (bottom + 1) % NumPoints(); } while (y(bottom) < i);
  	int prev = (bottom - 1 + NumPoints()) % NumPoints();
  	// y(prev) < i <= y(bottom) 	

	while (i < x(bottom)) {
      double alpha = (i-y(prev))/(y(bottom)-y(prev));
      double xi = (1-alpha)*x(prev) + alpha*x(left);
  	  points.AtInsert(bottom++, new double2(xi,i++));
  	}
  }
}
===========================================================================*/



