// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Mon 05-27-1996

#include "sample.h"

//-----------------------------------------------------------------------//
//                          CurvSample
//-----------------------------------------------------------------------//

double CurvSample::Curvature() const
{
  double l = Ft.Length();
  return (Ft^Ftt).Length()/(l*l*l);
}

void CurvSample::Show() const
{
  std::cout << "F(t)   = " << F << std::endl
       << "F'(t)  = " << Ft << std::endl
       << "F''(t) = " << Ftt << std::endl
       << "kappa  = " << Curvature() << std::endl;
}

std::ostream& operator<<(std::ostream& to, const CurvSample& sample)
{
  return to << sample.F << std::endl 
            << sample.Ft << std::endl 
            << sample.Ftt << std::endl;
}

std::istream& operator>>(std::istream& from, CurvSample& sample)
{
  return from >> sample.F 
              >> sample.Ft 
              >> sample.Ftt;
}

//-----------------------------------------------------------------------//
//                          FuncSample1D
//-----------------------------------------------------------------------//

void FuncSample1D::Show() const
{
  std::cout << "f(t)   = " << f << std::endl
       << "f'(t)  = " << ft << std::endl
       << "f''(t) = " << ftt << std::endl;
}

std::ostream& operator<<(std::ostream& to, const FuncSample1D& sample)
{
  return to << sample.f 
            << sample.ft 
            << sample.ftt << std::endl;
}

std::istream& operator>>(std::istream& from, FuncSample1D& sample)
{
  return from >> sample.f 
              >> sample.ft 
              >> sample.ftt;
}

//-----------------------------------------------------------------------//
//                          FuncSample2D
//-----------------------------------------------------------------------//

void FuncSample2D::Show() const
{
  std::cout << "f(u,v)   = " << f << std::endl
       << "fu(u,v)  = " << fu << std::endl
       << "fv(u,v)  = " << fv << std::endl
       << "fuu(u,v) = " << fuu << std::endl
       << "fuv(u,v) = " << fuv << std::endl
       << "fvv(u,v) = " << fvv << std::endl;
}

std::ostream& operator<<(std::ostream& to, const FuncSample2D& sample) 
{
  return to << sample.f 
            << sample.fu 
            << sample.fv 
            << sample.fuu 
            << sample.fuv 
            << sample.fvv << std::endl;
}

std::istream& operator>>(std::istream& from, FuncSample2D& sample) 
{
  return from >> sample.f 
              >> sample.fu 
              >> sample.fv 
              >> sample.fuu 
              >> sample.fuv 
              >> sample.fvv;
}

//-----------------------------------------------------------------------//
//                          SurfSample
//-----------------------------------------------------------------------//

SurfSample::SurfSample(const SurfSample& sample)
{
  SetValues(sample.F, sample.Fu, sample.Fv, sample.Fuu, sample.Fuv, sample.Fvv);
}

SurfSample::SurfSample(const double3& F0, const double3& Fu0, const double3& Fv0,
                       const double3& Fuu0, const double3& Fuv0, const double3& Fvv0)
{
  SetValues(F0, Fu0, Fv0, Fuu0, Fuv0, Fvv0);
}

void SurfSample::SetValues(const double3& F0, const double3& Fu0, const double3& Fv0,
                           const double3& Fuu0, const double3& Fuv0, const double3& Fvv0)
{                           
  F   = F0;
  Fu  = Fu0;
  Fv  = Fv0;
  Fuu = Fuu0;
  Fuv = Fuv0;
  Fvv = Fvv0;
}

double2x2 SurfSample::FirstFundamentalForm() const
{
  double g_11  = (Fu,Fu);
  double g_12  = (Fu,Fv);
  double g_21  = g_12;
  double g_22  = (Fv,Fv);
  return double2x2(g_11, g_12, g_21, g_22);
}

double2x2 SurfSample::SecondFundamentalForm() const
{
  double3 n = Normal();
  
  double t11 = (Fuu,n);
  double t12 = (Fuv,n);
  double t21 = t12;
  double t22 = (Fvv,n);
  return double2x2(t11, t12, t21, t22);
}

double3 SurfSample::Normal() const
{
  return (Fu^Fv).Unit();
}

void SurfSample::GetCurvatures(double& kappa1, double& kappa2) const
// The principal curvatures are the eigenvalues of the matrix I_inv*II.
{
  double2x2 I_inv = FirstFundamentalForm().Inverse();
  double2x2 II = SecondFundamentalForm();
  int result = (I_inv*II).GetEigenvalues(kappa1, kappa2);

  if (!result)
    std::cerr << "error in SurfSample::GetCurvatures!" << std::endl;
}

double SurfSample::TotalCurvature() const
// The total curvature is the trace of the matrix (I_inv*II)^2.
{
  double2x2 I = FirstFundamentalForm();
  double2x2 II = SecondFundamentalForm();

  if (fabs(I.Det()) < 1e-20) {
    std::cerr << "irregular point in SurfSample::TotalCurvature()!!" << std::endl;
    return 0;
  }
  
  double2x2 A = I.Inverse()*II;
  return (A*A).Trace();
}

double SurfSample::MeanCurvature() const
// The mean curvatures is the trace of the matrix I_inv*II.
{
  double2x2 I = FirstFundamentalForm();
  double2x2 II = SecondFundamentalForm();
  if (fabs(I.Det()) < 1e-20) {
    std::cerr << "irregular point in SurfSample::MeanCurvature()!!" << std::endl;
    return 0;
  }
  
  return (I.Inverse()*II).Trace()/2;
}

double SurfSample::GaussianCurvature() const
// The Gaussian curvatures is the determinant of the matrix I_inv*II.
{
  double2x2 I = FirstFundamentalForm();
  double2x2 II = SecondFundamentalForm();

  if (fabs(I.Det()) < 1e-20) {
    std::cerr << "irregular point in SurfSample::GaussianCurvature()!!" << std::endl;
    return 0;
  }
  
  return (I.Inverse()*II).Det();
}

void SurfSample::Show() const
{
  double kappa1, kappa2;
  GetCurvatures(kappa1, kappa2);
  
  std::cout << "F(u,v)   = " << F << std::endl
       << "Fu(u,v)  = " << Fu << std::endl
       << "Fv(u,v)  = " << Fv << std::endl
       << "Fuu(u,v) = " << Fuu << std::endl
       << "Fuv(u,v) = " << Fuv << std::endl
       << "Fvv(u,v) = " << Fvv << std::endl
       << "kappa1             = " << kappa1 << std::endl
       << "kappa2             = " << kappa2 << std::endl
       << "total curvature    = " << TotalCurvature() << std::endl
       << "mean curvature     = " << MeanCurvature() << std::endl
       << "gaussian curvature = " << GaussianCurvature() << std::endl;
}

std::ostream& operator<<(std::ostream& to, const SurfSample& sample) 
{
  return to << sample.F << std::endl
            << sample.Fu << std::endl
            << sample.Fv << std::endl
            << sample.Fuu << std::endl
            << sample.Fuv << std::endl
            << sample.Fvv << std::endl;
}

std::istream& operator>>(std::istream& from, SurfSample& sample)
{
  return from >> sample.F 
              >> sample.Fu 
              >> sample.Fv 
              >> sample.Fuu 
              >> sample.Fuv 
              >> sample.Fvv;
}

//-----------------------------------------------------------------------//
//                          SurfSampleExt
//-----------------------------------------------------------------------//

SurfSampleExt::SurfSampleExt(const SurfSampleExt& sample)
{
  SetValues(sample.F, sample.Fu, sample.Fv, sample.Fuu, sample.Fuv, sample.Fvv);
}

SurfSampleExt::SurfSampleExt(const double3& F0, const double3& Fu0, const double3& Fv0,
                       const double3& Fuu0, const double3& Fuv0, const double3& Fvv0)
{
  SetValues(F0, Fu0, Fv0, Fuu0, Fuv0, Fvv0);
}

void SurfSampleExt::SetValues(const double3& F0, const double3& Fu0, const double3& Fv0,
                           const double3& Fuu0, const double3& Fuv0, const double3& Fvv0)
{ 
  SurfSample::SetValues(F0, Fu0, Fv0, Fuu0, Fuv0, Fvv0);

  // compute first fundamental form
  double g_11  = (Fu,Fu);
  double g_12  = (Fu,Fv);
  double g_21  = g_12;
  double g_22  = (Fv,Fv);
  I = double2x2(g_11, g_12, g_21, g_22);

  // compute inverse of first fundamental form
  g = I.Det();
  double g11   = g_22/g;
  double g22   = g_11/g;
  double g12   = - g_12/g;
  double g21   = g12;
  I_inv = double2x2(g11, g12, g21, g22);

  // compute first order Christoffel symbols
  double G_111 = (Fu,Fuu);
  double G_121 = (Fu,Fuv);
  double G_211 = G_121;
  double G_221 = (Fu,Fvv);
  double G_112 = (Fv,Fuu);
  double G_122 = (Fv,Fuv);
  double G_212 = G_122;
  double G_222 = (Fv,Fvv);
  gamma_1 = double2x2(G_111, G_121, G_211, G_221);
  gamma_2 = double2x2(G_112, G_122, G_212, G_222);

  // compute second order Christoffel symbols
  double G111  = g11*G_111 + g12*G_112;
  double G121  = g11*G_121 + g12*G_122;
  double G211  = G121;
  double G221  = g11*G_221 + g12*G_222;
  double G112  = g21*G_111 + g22*G_112;
  double G122  = g21*G_121 + g22*G_122;
  double G212  = G122;
  double G222  = g21*G_221 + g22*G_222;
  gamma1  = double2x2(G111, G121, G211, G221);
  gamma2  = double2x2(G112, G122, G212, G222);

  // compute second fundamental form
  n = (Fu^Fv).Unit();
  double t11 = (Fuu,n);
  double t12 = (Fuv,n);
  double t21 = t12;
  double t22 = (Fvv,n);
  II = double2x2(t11, t12, t21, t22);
}

void SurfSampleExt::GetCurvatures(double& kappa1, double& kappa2) const
// The principal curvatures are the eigenvalues of the matrix I_inv*II.
{
  int result = (I_inv*II).GetEigenvalues(kappa1, kappa2);

  if (!result)
    std::cerr << "error in SurfSampleExt::GetCurvatures!" << std::endl;
}

double SurfSampleExt::TotalCurvature() const
// The total curvature is the trace of the matrix (I_inv*II)^2.
{
  double2x2 A = I_inv*II;
  return (A*A).Trace()/2;
}

double SurfSampleExt::MeanCurvature() const
// The mean curvatures is the trace of the matrix I_inv*II.
{
  return (I_inv*II).Trace()/2;
}

double SurfSampleExt::GaussianCurvature() const
// The Gaussian curvatures is the determinant of the matrix I_inv*II.
{
  return (I_inv*II).Det();
}

void SurfSampleExt::Show() const
{
  double kappa1, kappa2;
  GetCurvatures(kappa1, kappa2);
  
  std::cout << "F(u,v)   = " << F << std::endl
       << "Fu(u,v)  = " << Fu << std::endl
       << "Fv(u,v)  = " << Fv << std::endl
       << "Fuu(u,v) = " << Fuu << std::endl
       << "Fuv(u,v) = " << Fuv << std::endl
       << "Fvv(u,v) = " << Fvv << std::endl
       << "I        = " << I << std::endl
       << "g        = " << g << std::endl
       << "I_inv    = " << I_inv << std::endl
       << "II       = " << II << std::endl
       << "gamma1   = " << gamma1 << std::endl
       << "gamma2   = " << gamma2 << std::endl
       << "gamma_1  = " << gamma_1 << std::endl
       << "gamma_2  = " << gamma_2 << std::endl
       << "normal   = " << n << std::endl
       << "kappa1             = " << kappa1 << std::endl
       << "kappa2             = " << kappa2 << std::endl
       << "total curvature    = " << TotalCurvature() << std::endl
       << "mean curvature     = " << MeanCurvature() << std::endl
       << "gaussian curvature = " << GaussianCurvature() << std::endl;
}

