// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sat 06-15-1996

#include "surfmesh.h"
#include <fstream>
#include <vrml.h>

//-----------------------------------------------------------------------//
//                          SurfaceMesh                                  //
//-----------------------------------------------------------------------//

DoubleVector SurfaceMesh::GetTotalCurvatureValues() const
// returns a vector with the values of the total curvature of the vertices
{
  DoubleVector Result(NumSamples());
  
  for (int i=0; i<samples.Count; i++)
    Result(i) = samples[i].TotalCurvature();
    
  return Result;
}

DoubleVector SurfaceMesh::GetSimpleTotalCurvatureValues() const
// returns a vector with the values of the simple total curvature of the vertices
{
  DoubleVector Result(NumSamples());
  
  for (int i=0; i<samples.Count; i++)
    Result(i) = samples[i].SimpleTotalCurvature();
    
  return Result;
}

void SurfaceMesh::SaveVRML(std::ostream& to, const Function3D& colorfunc) const
{
  VRMLWriteBeginSeparator(to);
  to << "Material { diffuseColor [" << std::endl;
  for(int i=0; i<NumSamples(); i++) {
    double3 color = colorfunc(Sample(i).TotalCurvature());
    to << color.x() << " "
       << color.y() << " "
       << color.z() << "," << std::endl;
  }
  to << "] }" << std::endl;

  to << "MaterialBinding { value PER_VERTEX_INDEXED }" << std::endl << std::endl;
  to << "Normal { vector [" << std::endl;
  for(int i=0; i<NumSamples(); i++) {
    double3 normal = Sample(i).Normal();
    to << normal.x() << " "
       << normal.y() << " "
       << normal.z() << "," << std::endl;
  }
  to << "] }" << std::endl;

  to << "NormalBinding { value PER_VERTEX_INDEXED }" << std::endl << std::endl;
  to << "Coordinate3 { point [" << std::endl;
  for(int i=0; i<NumSamples(); i++)
    to << Sample(i).F.x() << " "
       << Sample(i).F.y() << " "
       << Sample(i).F.z() << "," << std::endl;
  to << "] }" << std::endl << std::endl;

  to << "IndexedFaceSet { coordIndex [" << std::endl;
  for (int i=0; i<NumTriangles(); i++) {
    int3 triangle = Triangle(i); 
    to << triangle[0] << ", "
       << triangle[1] << ", "
       << triangle[2] << ", "
       << "-1,"       << std::endl;
  }
  for (int i=0; i<NumQuadrangles(); i++) {
    int4 quadrangle = Quadrangle(i); 
    to << quadrangle[0] << ", "
       << quadrangle[1] << ", "
       << quadrangle[2] << ", "
       << quadrangle[3] << ", "
       << "-1,"       << std::endl;
  }
  to << "] }" << std::endl;
  to << "}" << std::endl;
}

void SurfaceMesh::SaveVRML(std::ostream& to, const double3& rgb) const
{
  OneColorFunction colorfunc(rgb);
  SaveVRML(to, colorfunc);
}

void SurfaceMesh::SaveVRML(std::ostream& to, const double3& rgb0, const double3& rgb1) const
{
  DoubleVector X = GetTotalCurvatureValues();
  double min    = X.Min();
  double median = X.Median();
  double max    = X.Max();

  PiecewiseFunction3D colorfunc(min, rgb0, median, (rgb0+rgb1)/2, max, rgb1);
  SaveVRML(to, colorfunc);
}

void SurfaceMesh::SaveVRML(char* FileName, const double3& rgb) const
{
  std::ofstream to(FileName);
  if (!to)
    std::cerr << "Could not open file " << FileName << std::endl;
  else {
    VRMLWriteFileHeader(to);
    SaveVRML(to, rgb);
    to << "}" << std::endl << std::endl;
  }
}
 
void SurfaceMesh::SaveVRML(char* FileName, const double3& rgb0, const double3& rgb1) const
{
  std::ofstream to(FileName);
  if (!to)
    std::cerr << "Could not open file " << FileName << std::endl;
  else {
    VRMLWriteFileHeader(to);
    SaveVRML(to, rgb0, rgb1);
    to << "}" << std::endl << std::endl;
  }
}

void SurfaceMesh::SaveVRML(char* FileName, const Function3D& colorfunc) const
{
  std::ofstream to(FileName);
  if (!to)
    std::cerr << "Could not open file " << FileName << std::endl;
  else {
    VRMLWriteFileHeader(to);
    SaveVRML(to, colorfunc);
    to << "}" << std::endl << std::endl;
  }
}

void SurfaceMesh::Show() const
// writes the contents to standard out
{
  std::cout << "SurfaceMesh::Show()" << std::endl;
  std::cout << " numsamples:     " << samples.Count << std::endl;
  std::cout << " numtriangles:   " << triangles.Count << std::endl;
  std::cout << " numquadrangles: " << quadrangles.Count << std::endl;
  
  for (int i=0; i<triangles.Count; i++)
    std::cout << "triangle " << triangles[i] << std::endl;
    
  for (int i=0; i<quadrangles.Count; i++)
    std::cout << "quadrangle " << quadrangles[i] << std::endl;
}

std::ostream& operator<<(std::ostream& to, const SurfaceMesh&) 
{
  return to << "this is a SurfaceMesh" << std::endl;
}

std::istream& operator>>(std::istream& from, SurfaceMesh&) 
{
  return from;
}

void SurfaceMesh::SaveOffsetVRML(std::ostream& to, const Function3D& colorfunc, double d) const
{
  VRMLWriteBeginSeparator(to);
  to << "Material { diffuseColor [" << std::endl;
  for(int i=0; i<NumSamples(); i++) {
    double3 color = colorfunc(Sample(i).TotalCurvature());
    to << color.x() << " "
       << color.y() << " "
       << color.z() << "," << std::endl;
  }
  to << "] }" << std::endl;

  to << "MaterialBinding { value PER_VERTEX_INDEXED }" << std::endl << std::endl;
  to << "Normal { vector [" << std::endl;
  for(int i=0; i<NumSamples(); i++) {
    double3 normal = Sample(i).Normal();
    to << normal.x() << " "
       << normal.y() << " "
       << normal.z() << "," << std::endl;
  }
  to << "] }" << std::endl;

  to << "NormalBinding { value PER_VERTEX_INDEXED }" << std::endl << std::endl;
  to << "Coordinate3 { point [" << std::endl;
  for(int i=0; i<NumSamples(); i++) {
    double3 point = Sample(i).F + d*Sample(i).Normal();
    to << point.x() << " "
       << point.y() << " "
       << point.z() << "," << std::endl;
  }
  to << "] }" << std::endl << std::endl;

  to << "IndexedFaceSet { coordIndex [" << std::endl;
  for (int i=0; i<NumTriangles(); i++) {
    int3 triangle = Triangle(i); 
    to << triangle[0] << ", "
       << triangle[1] << ", "
       << triangle[2] << ", "
       << "-1,"       << std::endl;
  }
  for (int i=0; i<NumQuadrangles(); i++) {
    int4 quadrangle = Quadrangle(i); 
    to << quadrangle[0] << ", "
       << quadrangle[1] << ", "
       << quadrangle[2] << ", "
       << quadrangle[3] << ", "
       << "-1,"       << std::endl;
  }
  to << "] }" << std::endl;
  VRMLWriteEndSeparator(to);
}

void SurfaceMesh::SaveOffsetVRML(std::ostream& to, const double3& rgb, double d) const
{
  OneColorFunction colorfunc(rgb);
  SaveOffsetVRML(to, colorfunc, d);
}

