// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-12-1996

#include <trapezd.h>

//-----------------------------------------------------------------------//
//                          Min                                          //
//-----------------------------------------------------------------------//
static inline double Min(double a, double b)
{
  return (a<b) ? a : b;
}

//-----------------------------------------------------------------------//
//                          VerticalTrapezoid2D                          //
//-----------------------------------------------------------------------//

double2 VerticalTrapezoid2D::Point(int i) const
{
  switch(i) {
    case 0: return double2(x0, a0 + a1*x0);
    case 1: return double2(x1, a0 + a1*x1);
    case 2: return double2(x1, b0 + b1*x1);
    case 3: return double2(x0, b0 + b1*x0);

    default:
      std::cerr << "Index out of range in VerticalTrapezoid2D::Point" << std::endl;
      return double2(0,0);
  }
}

VerticalTrapezoid2D::VerticalTrapezoid2D(const double2& left, double leftheight, const double2& right, double rightheight)
{
  x0 = left.x();
  x1 = right.x();
  a1 = (right.y() - left.y()) / (x1-x0);
  b1 = (right.y() + rightheight - left.y() - leftheight) / (x1-x0);
  a0 = left.y() - a1*x0;
  b0 = left.y() + leftheight - b1*x0;
}

void VerticalTrapezoid2D::Scale(double t)
// scales the trapezoid with factor t w.r.t. its center
{
  double2 center = Center();
  a0 = (1-t)*(center.y() - a1*center.x()) + t*a0;
  b0 = (1-t)*(center.y() - b1*center.x()) + t*b0;
  x0 = center.x() + t*(x0-center.x());
  x1 = center.x() + t*(x1-center.x());
}

void VerticalTrapezoid2D::Show() const
// writes the contents to standard out
{
  std::cout << "points: " << std::endl;
  for (int i=0; i<4; i++)
    std::cout << "  " << Point(i) << std::endl;
  std::cout << "area: " << Area() << std::endl;
  std::cout << "center: " << Center() << std::endl;
  std::cout << "x0, x1 = " << x0 << " " << x1 << std::endl;
}

int VerticalTrapezoid2D::GetTop() const
// returns the index of the point with the highest y-value
{
  if (Point(2).y() < Point(3).y())
    return 3;
  else
    return 2;
}

int VerticalTrapezoid2D::GetBottom() const
// returns the index of the point with the highest y-value
{
  if (Point(0).y() < Point(1).y())
    return 0;
  else
    return 1;
}

HorizontalTrapezoid2D VerticalTrapezoid2D::GetHorizontalTrapezoid(double ymin, double ymax, const int2& left, const int2& right) const
// returns the trapezoid bounded by the lines
//
//               y=ymin
//               y=ymax
//
// and the line segments
//
//      [ Point(left[0]), Point(left[1]) ]
//      [ Point(right[0]), Point(right[1]) ]
//
{
  double A1 = (x(left[1])-x(left[0])) / (y(left[1])-y(left[0]));
  double A0 = x(left[0]) - A1*y(left[0]);
  
  double B1 = (x(right[1])-x(right[0])) / (y(right[1])-y(right[0]));
  double B0 = x(right[0]) - B1*y(right[0]);
  
  return HorizontalTrapezoid2D(ymin, ymax, A0, A1, B0, B1);
}

List<HorizontalTrapezoid2D> VerticalTrapezoid2D::GetHorizontalTrapezoids(int SplitAtIntegers, double precision) const
// splits the trapezoid in horizontal trapezoids by cutting it with
// horizontal lines through the vertices
//
// SplitAtIntegers: if true, then the trapezoid is also cut at horizontal
//                  lines y=yi, with yi an integer
// 
// precision: trapezoids with (height < precision) are neglected
{
  List<HorizontalTrapezoid2D> Result;

  int top = GetTop();
  int bottom = GetBottom();

  int delta = 1;

  int2 left(bottom, (bottom-delta+NumPoints()) % NumPoints()); 
  int2 right(bottom, (bottom+delta+NumPoints()) % NumPoints());

  double ymin = y(bottom);

  while (ymin < y(top)) {

    if (SplitAtIntegers) {
      double ymax = Min(y(left[1]), y(right[1]));
      for (int yy=ceil(ymin); yy<ymax; yy++) {
        if (yy - ymin > precision)
          Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, yy, left, right));
        ymin = yy;
      }
    }
    
    if (y(left[1]) <= y(right[1])) {
      if (y(left[1]) - ymin > precision)
        Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, y(left[1]), left, right));
      ymin = y(left[1]);
      left[0] = left[1];
      left[1] = (left[1] - delta + NumPoints()) % NumPoints();
    }
    else {
      if (y(right[1]) - ymin > precision)
        Result += new HorizontalTrapezoid2D(GetHorizontalTrapezoid(ymin, y(right[1]), left, right));
      ymin = y(right[1]);
      right[0] = right[1];
      right[1] = (right[1] + delta + NumPoints()) % NumPoints();
    }
  }

  return Result;
}

//-----------------------------------------------------------------------//
//                          HorizontalTrapezoid2D                        //
//-----------------------------------------------------------------------//

double2 HorizontalTrapezoid2D::Point(int i) const
{
  switch(i) {
    case 0: return double2(a0 + a1*y0, y0);
    case 1: return double2(b0 + b1*y0, y0);
    case 2: return double2(b0 + b1*y1, y1);
    case 3: return double2(a0 + a1*y1, y1);

    default:
      std::cerr << "Index out of range in HorizontalTrapezoid2D::Point" << std::endl;
      return double2(0,0);
  }
}

HorizontalTrapezoid2D::HorizontalTrapezoid2D(const double2& bottom, double bottomwidth, const double2& top, double topwidth)
{
  y0 = bottom.y();
  y1 = top.y();
  a1 = (top.x() - bottom.x()) / (y1-y0);
  b1 = (top.x() + topwidth - bottom.x() - bottomwidth) / (y1-y0);
  a0 = bottom.x() - a1*y0;
  b0 = bottom.x() + bottomwidth - b1*y0;
}

void HorizontalTrapezoid2D::Scale(double t)
// scales the trapezoid with factor t w.r.t. its center
{
  double2 center = Center();
  a0 = (1-t)*(center.x() - a1*center.y()) + t*a0;
  b0 = (1-t)*(center.x() - b1*center.y()) + t*b0;
  y0 = center.y() + t*(y0-center.y());
  y1 = center.y() + t*(y1-center.y());
}

void HorizontalTrapezoid2D::Show() const
// writes the contents to standard out
{
  std::cout << "points: " << std::endl;
  for (int i=0; i<4; i++)
    std::cout << "  " << Point(i) << std::endl;
  std::cout << "area: " << Area() << std::endl;
  std::cout << "center: " << Center() << std::endl;
  std::cout << "y0, y1 = " << y0 << " " << y1 << std::endl;
}

int HorizontalTrapezoid2D::GetRight() const
// returns the index of the point with the highest x-value
{
  if (Point(1).x() < Point(2).x())
    return 2;
  else
    return 1;
}

int HorizontalTrapezoid2D::GetLeft() const
// returns the index of the point with the lowest x-value
{
  if (Point(0).x() < Point(3).x())
    return 0;
  else
    return 3;
}

VerticalTrapezoid2D HorizontalTrapezoid2D::GetVerticalTrapezoid(double xmin, double xmax, const int2& bottom, const int2& top) const
// returns the trapezoid bounded by the lines
//
//               x=xmin
//               x=xmax
//
// and the line segments
//
//      [ Point(bottom[0]), Point(bottom[1]) ]
//      [ Point(top[0]), Point(top[1]) ]
//
{
  double A1 = (y(bottom[1])-y(bottom[0])) / (x(bottom[1])-x(bottom[0]));
  double A0 = y(bottom[0]) - A1*x(bottom[0]);
  
  double B1 = (y(top[1])-y(top[0])) / (x(top[1])-x(top[0]));
  double B0 = y(top[0]) - B1*x(top[0]);
  
  return VerticalTrapezoid2D(xmin, xmax, A0, A1, B0, B1);
}

List<VerticalTrapezoid2D> HorizontalTrapezoid2D::GetVerticalTrapezoids(int SplitAtIntegers, double precision) const
// splits the trapezoid in vertical trapezoids by cutting it with
// vertical lines through the vertices
//
// SplitAtIntegers: if true, then the trapezoid is also cut at vertical
//                  lines x=xi, with xi an integer
// 
// precision: trapezoids with (width < precision) are neglected
{
  List<VerticalTrapezoid2D> Result;

  int left = GetLeft();
  int right = GetRight();

  int delta = 1;

  int2 bottom(left, (left+delta+NumPoints()) % NumPoints()); 
  int2 top(left, (left-delta+NumPoints()) % NumPoints());

  double xmin = x(left);

  while (xmin < x(right)) {

    if (SplitAtIntegers) {
      double xmax = Min(x(bottom[1]), x(top[1]));
      for (int xx=ceil(xmin); xx<xmax; xx++) {
        if (xx - xmin > precision)
          Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, xx, bottom, top));
        xmin = xx;
      }
    }
    
    if (x(bottom[1]) <= x(top[1])) {
      if (x(bottom[1]) - xmin > precision)
        Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, x(bottom[1]), bottom, top));
      xmin = x(bottom[1]);
      bottom[0] = bottom[1];
      bottom[1] = (bottom[1] + delta + NumPoints()) % NumPoints();
    }
    else {
      if (x(top[1]) - xmin > precision)
        Result += new VerticalTrapezoid2D(GetVerticalTrapezoid(xmin, x(top[1]), bottom, top));
      xmin = x(top[1]);
      top[0] = top[1];
      top[1] = (top[1] - delta + NumPoints()) % NumPoints();
    }
  }

  return Result;
}

/*
List<VerticalTrapezoid2D> HorizontalTrapezoid2D::GetVerticalTrapezoids(int SplitAtIntegers, double precision) const
// splits the trapezoid in vertical trapezoids by cutting it with
// vertical lines through the vertices
//
// SplitAtIntegers: if true, then the trapezoid is also cut at vertical
//                  lines x=xi, with xi an integer
// 
// precision: trapezoids with (width < precision) are neglected
{
  List<VerticalTrapezoid2D> Result;

std::cout << "HorizontalTrapezoid2D::GetVerticalTrapezoids()" << std::endl;
Show();

  int left = GetLeft();
  int right = GetRight();

std::cout << "left = " << left << " " << Point(left) << std::endl;
std::cout << "right = " << right << " " << Point(right) << std::endl;

  int delta = 1;

  int2 bottom(left, (left+delta+NumPoints()) % NumPoints()); 
  int2 top(left, (left-delta+NumPoints()) % NumPoints());

  double xmin = x(left);

  while (xmin < x(right)) {
std::cout << "xmin = " << xmin << std::endl;
std::cout << "bottom[0] " << bottom[0] << " " << Point(bottom[0]) << std::endl;
std::cout << "bottom[1] " << bottom[1] << " " << Point(bottom[1]) << std::endl;
std::cout << "top[0] " << top[0] << " " << Point(top[0]) << std::endl;
std::cout << "top[1] " << top[1] << " " << Point(top[1]) << std::endl;

    if (SplitAtIntegers) {
      double xmax = Min(x(bottom[1]), x(top[1]));
      for (int xx=ceil(xmin); xx<xmax; xx++) {
        if (xx - xmin > precision) {
          Result += GetVerticalTrapezoid(xmin, xx, bottom, top);
          std::cout << "xx = " << std::endl;
        }
        xmin = xx;
      }
    }
    
    if (x(bottom[1]) <= x(top[1])) {
      if (x(bottom[1]) - xmin > precision)
        Result += GetVerticalTrapezoid(xmin, x(bottom[1]), bottom, top);
      xmin = x(bottom[1]);
      bottom[0] = bottom[1];
      bottom[1] = (bottom[1] + delta + NumPoints()) % NumPoints();
    }
    else {
      if (x(top[1]) - xmin > precision) {
        Result += GetVerticalTrapezoid(xmin, x(top[1]), bottom, top);
        std::cout << xmin << " " << x(top[1]) << " " << bottom << top << std::endl;
      }
      xmin = x(top[1]);
      top[0] = top[1];
      top[1] = (top[1] - delta + NumPoints()) % NumPoints();
    }
  }

  return Result;
}
*/
